; $Id: dm_calc_projection.pro,v 1.27 2010/08/23 16:27:01 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dm_calc_projection
;
; PURPOSE:
;  calculate projection
;  apply masking, rebinning, and detector efficiency adjustment to data
;
; CATEGORY:
;  dcs_mslice
;
; HISTORY:
;  02/2010: added NXSPE support - Andrei Savici (saviciat@ornl.gov)   
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  August, 2010
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

pro dcs_mslice::dm_calc_projection,debug=debug
    self.error = 1  ;initialize the error flag, it will be cleared at the end of the program
    if keyword_set(debug) then begin
       print,'**Calculate projection...'
       current=systime(/sec)
    end

    det_thick = 1.0     ;DCS He3 detector thickness in cm
    det_p     = 6.0     ;DCS He3 detector pressure in bar
    is_spe    = 0b      ;type of psi defintion

    case self.ftype of
        0:  begin   ;dcs
            if (~ ptr_valid(self.detPosPtr)) then self.detPosPtr=dm_load_detpos(/dcs,adjustdcs=self.dcs2thcorr)
            end
        1:  begin   ;spe
            is_spe = 1b      ;spe type detector psi definition
            if (~ ptr_valid(self.detPosPtr)) then begin
               self->dm_load_phx,phxfile=phxfile
               if strlen(phxfile) ne 0 then self.phxfile=phxfile
            endif
            end
        2:  begin   ;nxspe
            is_spe = 1b      ;spe type detector psi definition
            if (~ ptr_valid(self.detPosPtr)) then begin
               self->dm_load_phx,phxfile=phxfile
            endif
            end
        3:  begin   ;inx
            if (~ ptr_valid(self.detPosPtr)) then begin
               self.detPosPtr = dm_load_detpos(/neat)
               ok = dialog_message('Detector information is not available, using NEAT instrument info instead.',title='Please note:',dialog_parent=self.tlb)
            endif
            end
        4:  begin   ;macs
            if (~ ptr_valid(self.detPosPtr)) then self.detPosPtr=dm_load_detpos(/macs)
            end
        else:
    endcase

    if (~ ptr_valid(self.detPosPtr)) then begin
       info = ['Detector information is not available. The projection calculation is aborted.']
       ok   = dialog_message(info,/error,dialog_parent=self.tlb)
       return
    endif

    ;check if the common input parameters are all set for single crystal and diffuse scattering type
    if self.samp_typ ne 0 then begin
       missing = 0b 
       if total(finite(self.latt_parm)) ne 6 then begin
          mesg = 'Missing lattice parameters.'
          missing = 1b
       endif
       if total([finite(self.orie_u),finite(self.orie_v),finite(self.orie_psi)]) ne 7 then begin
          if n_elements(mesg) eq 0 then $
             mesg = 'Missing crystal orientation parameters.'  $
          else $
             mesg = [mesg,'Missing crystal orientation parameters.']
          missing = 1b
       endif
       if (total([finite(self.view_u1),finite(self.view_u2)]) ne 6) or (total(finite(self.view_u3)) ne 3 and self.extravaxis_yn) then begin
          if n_elements(mesg) eq 0 then $
             mesg = 'Missing viewing axes parameters.'  $
          else $
             mesg = [mesg,'Missing viewing axes parameters.']
          missing = 1b
       endif
       if missing then begin
          ok = dialog_message(mesg,/error,dialog_parent=self.tlb)
          return
       endif
    endif

    if (~ ptr_valid(self.dataStrPtr)) then begin    ;no data loaded yet
       ;ask for Ei or Ef according to instrument geometry
       eief = 5.0
       if (finite(self.eief)) then $
          if self.eief gt 0 then eief=self.eief
       if self.samp_typ eq 2 then begin             ;diffuse scattering single E
          case self.e_bin[self.samp_typ] of
            0:  en = 0.0                                      ;use elastic energy for all time channels
            1:  en = 0.0                                      ;use elastic energy for elastic peak
            2:  en = (self.e_range[0]+self.e_range[1])/2.0    ;use the mean energy for specified energy range
            else: en = 0.0
          endcase
          askeran = 0b
          info=['E='+dm_to_string(en)+' meV']
       endif else if self.samp_typ eq 3 then begin  ;diffuse scattering multiple E
          askeran = 1b
          case self.e_bin[3] of
            0:  info = (['-Ei< E <Ei','-Ef< E <Ef'])[self.instrgeom]
            1:  begin
                info = ['E~0 meV']
                askeran = 0b
                end
            2:  begin
                info = [dm_to_string(self.e_range[0])+'< E <'+dm_to_string(self.e_range[1])]
                if strlen(self.userf_str) ne 0 then self.estartend[0:1] = self.e_range[0:1]
                end
            else:
          endcase
       endif else begin
          info = (['-Ei< E <Ei','-Ef< E <Ef'])[self.instrgeom]
          askeran = 1b
       endelse
       if self.ftype eq 0 then begin    ;for DCS convert energy(meV) to wavelength(angstrom)
          eief  = sqrt(81.8042/eief)
          title = "Incident wavelength:"
       endif else $
          title = (["Incident energy:","Final energy:"])[self.instrgeom]
       notok = 1b
       while(notok) do begin
          if askeran and strlen(self.userf_str) ne 0 then begin
             if self.ftype eq 0 then begin
                eief = dm_dialog_input(['wavelength ('+string('c5'XB)+'):','E_start:','E_end:','n_E:'],/float,$
                       default=[finite(eief[0])?fix(eief[0]*100.0+0.5,type=3)/100.0:eief[0],self.estartend],$
                       dialog_parent=self.tlb,info='E_start< E <E_end, n_E is the sampling number.',/lambda2e,title=title)
             endif else begin
                eief = dm_dialog_input([(['Ei','Ef'])[self.instrgeom]+'(meV):','E_start:','E_end:','n_E:'],/float,$
                       title=title,default=[finite(eief[0])?fix(eief[0]*1000.0+0.5,type=3)/1000.0:eief[0],self.estartend],$
                       dialog_parent=self.tlb,info='E_start< E <E_end, n_E is the sampling number.',/e2lambda)
             endelse
          endif else begin
             if self.ftype eq 0 then begin
                eief = dm_dialog_input('wavelength ('+string('c5'XB)+'):',/float,default=finite(eief)?fix(eief*100.0+0.5,type=3)/100.0:eief,$
                       dialog_parent=self.tlb,info=info,/lambda2e,title=title)
             endif else begin
                eief = dm_dialog_input((['Ei','Ef'])[self.instrgeom]+'(meV):',/float,title=title,$
                       default=finite(eief)?fix(eief*1000.0+0.5,type=3)/1000.0:eief,dialog_parent=self.tlb,info=info,/e2lambda)
             endelse
          endelse
          if (total(finite(eief)) eq n_elements(eief)) then $
             if eief[0] gt 0 then notok = 0b
       endwhile
       if self.ftype eq 0 then $    ;for DCS convert wavelength to energy
          eief[0] =  81.8042/(eief[0]^2)
       self.eief = eief[0]  ;save ei or ef information
       if askeran and strlen(self.userf_str) ne 0 then begin
          if eief[3] lt 1 then eief[3] = 81
          
          if self.instrgeom eq 0 then eief[2] = min([eief[2],eief[0]]) $
          else eief[1] = max([eief[1],-eief[0]])
          en = eief[1] + findgen(eief[3])/(eief[3]-1)*(eief[2]-eief[1]) +(2*self.instrgeom-1)*1e-6
          self.estartend = eief[1:3]
          eief = eief[0]
       endif else begin
          nnen = 81.           ;sampling number
          if self.samp_typ eq 2 then begin             ;diffuse scattering single E
             if (self.instrgeom eq 0) and (en gt eief) then begin
                ok = dialog_message('Incident energy is less than the energy transfer: E>Ei. E=0 is used.',/error)
                en = 0.0
             endif
             if (self.instrgeom eq 1) and (en lt -eief) then begin
                ok = dialog_message('Neutron energy gain is more than the final energy: E<-Ef. E=0 is used.',/error)
                en = 0.0
             endif
          endif else if self.samp_typ eq 3 then begin  ;diffuse scattering multiple E
             case self.e_bin[3] of
               0:  en = (findgen(nnen)/((nnen-1)/2.)-1)*eief+(2*self.instrgeom-1)*1e-6         ;[-Ei,...,Ei] or [-Ef,...Ef]
               1:  begin
                   ;use empirical DCS energy resolution as the energy range
                   dE = 0.02008*eief^(1.4735)
                   en = (findgen(nnen)/((nnen-1)/2.)-1)*dE
                   end
               2:  begin
                   if self.instrgeom eq 0 then begin
                      if min(self.e_range) gt eief then begin
                         ok = dialog_message('Incident energy is smaller than the specified energy range. Elastic peak is used.',/error)
                         ;use empirical DCS energy resolution as the energy range
                         dE = 0.02008*eief^(1.4735)
                         en = (findgen(nnen)/((nnen-1)/2.)-1)*dE
                      endif else begin
                         eran = [self.e_range[0],min([self.e_range[1],eief])]
                         en   = eran[0] + findgen(nnen)/(nnen-1)*(eran[1]-eran[0])-1e-6
                      endelse
                   endif else begin
                      if max(self.e_range) lt -eief then begin
                         ok = dialog_message('Neutron energy gain in the specified energy range is more than the final energy. Elastic peak is used.',/error)
                         ;use empirical DCS energy resolution as the energy range
                         dE = 0.02008*eief^(1.4735)
                         en = (findgen(nnen)/((nnen-1)/2.)-1)*dE
                      endif else begin
                         eran = [max(self.e_range[0],-eief),self.e_range[1]]
                         en   = eran[0] + findgen(nnen)/(nnen-1)*(eran[1]-eran[0])+1e-6
                      endelse
                   endelse
                   end
               else:
             endcase
          endif else begin
             en = (findgen(nnen)/((nnen-1)/2.)-1)*eief +(2*self.instrgeom-1)*1e-6         ;[-Ei,...,Ei] or [-Ef,...Ef]
          endelse
       endelse
       if self.samp_typ ge 2 then begin ;A2 range for diffuse scattering type
          notok = 1b
          if total(finite(self.a2[0:2])) eq 3 then a2=self.a2[0:2] else a2 = [0,50,2]
          while(notok) do begin
             a2  = dm_dialog_input(['Start:','End:','Step:'],title='Psi range:',default=a2,/float,dialog_parent=self.tlb)
             if (total(finite(a2)) eq 3) then $
                if a2[2] ne 0 then notok = 0b
          endwhile
          a2[2]  = abs(a2[2])
          min_a  = min(a2[0:1],max=max_a)
          self.a2[0:2] = [min_a,max_a,a2[2]]
          a2     = min_a+findgen(fix((max_a-min_a)/a2[2])+1)*a2[2]
          n_file = n_elements(a2)
       endif
       qty  = 0.
       err  = 0.
       ewid = 0.
       time = 0.
    endif else begin
       ewid = (*self.dataStrPtr).ewid
       eief = (*self.dataStrPtr).eief & self.eief = eief
       qty  = ((*self.dataStrPtr).qty)
       err  = ((*self.dataStrPtr).err)
       while (~ finite(eief)) do begin
            eief = dm_dialog_input((['Ei','Ef'])[self.instrgeom]+'(meV):',/float,default=eief,dialog_parent=self.tlb,/e2lambda)
            (*self.dataStrPtr).eief = eief
            self.eief = eief
       endwhile
       en   = (*self.dataStrPtr).e
       time = (*self.dataStrPtr).time
       if self.samp_typ ge 2 then begin                    ;diffuse scattering type
          a2 = (*self.dataStrPtr).rotation                 ;angle 2-rotation
          self.orie_offset = self.orie_psi-self.a2[3]      ;save the offset information
          a2 = a2+self.orie_offset                         ;adjust the offset
          n_file  = n_elements(a2)
          tmpmin  = min(a2,max=tmpmax)
          tmpstep = (n_file gt 1)?((tmpmax-tmpmin)/(n_file-1)):(0)
          self.a2[0:2] = [tmpmin,tmpmax,tmpstep]
       endif else if self.samp_typ eq 1 then begin
          if finite((*self.dataStrPtr).rotation) then $
             self.orie_offset = self.orie_psi-(*self.dataStrPtr).rotation
       endif
       if ptr_valid(self.bgdata) then begin
          bgqty = (*self.bgdata).qty
          bgerr = (*self.bgdata).err
          bgang = (*self.bgdata).rotation
          if self.samp_typ ge 2 then n_bgfile = n_elements(bgang)
       endif
       if ptr_valid(self.bgdata2) then begin
          bgqty2 = (*self.bgdata2).qty
          bgerr2 = (*self.bgdata2).err
          bgang2 = (*self.bgdata2).rotation
          if self.samp_typ ge 2 then n_bgfile2 = n_elements(bgang2)
          ;apply the empty can factor
          if total(finite(self.empcanfactor)) eq 2 then begin
             bgqty  = bgqty*self.empcanfactor[0]
             bgerr  = bgerr*self.empcanfactor[0]
             bgqty2 = bgqty2*self.empcanfactor[1]
             bgerr2 = bgerr2*self.empcanfactor[1]
          endif
       endif
    endelse
    det_tt   = (*self.detPosPtr).two_theta
    det_psi  = (*self.detPosPtr).psi
    nen      = n_elements(en)
    ndet     = n_elements(det_tt)
    widget_control,/hourglass
    if ptr_valid(self.dataStrPtr) then begin
       ;subtract overall background rate
       if ptr_valid(self.bgratePtr) then begin
          bgrate = (*self.bgratePtr)
          nrate  = n_elements(bgrate)
          if nrate gt 1 then begin
             dbgrate = bgrate[*,1]
             bgrate  = bgrate[*,0]
          endif else $
             dbgrate = 1.0      ;arbitrary error bar
          case self.ftype of
            0:  begin   ;dcs
                if nrate eq 1 then begin    ;overall rate
                   if ptr_valid(self.eff) then tmp_eff=(*self.eff) else tmp_eff=fltarr(ndet)+1.
                   tmp_value = bgrate/3600./(self.monrate)*temporary(tmp_eff)
                endif else $
                   tmp_value = bgrate/3600./(self.monrate)
                if self.samp_typ ge 2 then begin      ;diffuse scattering type
                   if self.samp_typ eq 2 then begin   ;single energy
                      if self.eint_yn then $
                         tmp_value = tmp_value*total(ewid) $
                      else $
                         tmp_value = tmp_value
                      for i=0L,n_file-1 do qty[*,i]=qty[*,i]-tmp_value
                      if ptr_valid(self.bgdata) then $
                         for i=0L,n_bgfile-1 do bgqty[*,i]=bgqty[*,i]-tmp_value
                      if ptr_valid(self.bgdata2) then $
                         for i=0L,n_bgfile2-1 do bgqty2[*,i]=bgqty2[*,i]-tmp_value
                   endif else begin                    ;multiple energy
                      for i=0L,n_file-1L do begin
                          for j=0L,nen-1L do $
                              qty[j,*,i] = qty[j,*,i]-tmp_value
                      endfor
                      if ptr_valid(self.bgdata) then begin
                         for i=0L,n_bgfile-1L do begin
                             for j=0L,nen-1L do $
                                 bgqty[j,*,i] = bgqty[j,*,i]-tmp_value
                         endfor
                      endif
                      if ptr_valid(self.bgdata2) then begin
                         for i=0L,n_bgfile2-1L do begin
                             for j=0L,nen-1L do $
                                 bgqty2[j,*,i] = bgqty2[j,*,i]-tmp_value
                         endfor
                      endif
                   endelse
                endif else begin
                   for i=0L,nen-1L do $
                       qty[i,*] = qty[i,*]-tmp_value
                   if ptr_valid(self.bgdata) then $
                      for i=0L,nen-1L do $
                          bgqty[i,*] = bgqty[i,*]-tmp_value
                   if ptr_valid(self.bgdata2) then $
                      for i=0L,nen-1L do $
                          bgqty2[i,*] = bgqty2[i,*]-tmp_value
                endelse
                tmp_value = 1     ;free up the memory
                end
            1:  begin   ;spe
                ;no implemented
                end
            2:  begin   ;nxspe
                ;no implemented
                end
            3:  begin   ;inx
                ;no implemented
                end
            else:
          endcase
       endif
       ;subtract detector dependent background data
       if ptr_valid(self.bgdetsPtr) then begin
          tmp_value = (*self.bgdetsPtr)
          if self.samp_typ ge 2 then begin
             if self.eint_yn then $
                tmp_value = tmp_value*total(ewid)
             if self.samp_typ eq 2 then begin
                for i=0L,n_file-1L do $
                    qty[*,i] = qty[*,i]-tmp_value
             endif else begin
                tmp_value = (fltarr(nen)+1.0)#tmp_value
                for i=0L,n_file-1L do $
                    qty[*,*,i] = qty[*,*,i]-tmp_value
             endelse
          endif else begin
             qty = qty-(fltarr(nen)+1.0)#tmp_value
          endelse
          tmp_value = 1     ;free up the memory
       endif
       ;subtract time channel background data
       if ptr_valid(self.bgtchnPtr) then begin
          tmp_emin  = min((*self.bgtchnPtr).emean,max=tmp_emax)
          tmp_nen   = n_elements((*self.bgtchnPtr).emean)
          tchan_ind = where(en ge tmp_emin and en le tmp_emax,tchan_count)
          if tchan_count gt tmp_nen then tchan_ind = tchan_ind[0:(tmp_nen-1)]   ;errorproof
          if tchan_count ne 0 then begin
             if self.samp_typ eq 3 then begin
                for i=0L,n_file-1L do $
                    for j=0L,ndet-1L do $
                        qty[tchan_ind,j,i] = qty[tchan_ind,j,i]-(*self.bgtchnPtr).qty
             endif else begin
                for i=0L,ndet-1L do $
                    qty[tchan_ind,i] = qty[tchan_ind,i]-(*self.bgtchnPtr).qty
             endelse
          endif
       endif else tchan_count = 0
       ;absorption correction
       if ptr_valid(self.absp_info) then begin
          ninfo = n_elements(*(self.absp_info))
          info  = [(*(self.absp_info))[0],(*(self.absp_info))[3:(ninfo-1)]]
          if info[0] ge 0 and info[0] lt n_elements(self.absp_npts) then begin
             npts  = (*self.absp_npts[info[0]])[*,(*(self.absp_info))[1]]
             tmp   = dm_calc_shielding(info,eief,en,det_tt,interp=(*(self.absp_info))[2],instrgeom=self.instrgeom,$
                     is_dcs=(self.ftype eq 0),npoints=npts,group_leader=self.tlb,stopped=stopped)
             if (~ stopped) then begin
                if self.samp_typ ge 2 then begin    ;diffuse scattering type
                   if self.samp_typ eq 2 then begin
                      for i=0L,n_file-1L do begin
                          qty[*,i] = qty[*,i]*tmp
                          err[*,i] = err[*,i]*tmp
                      endfor
                   endif else begin
                      for i=0L,n_file-1L do begin
                          qty[*,*,i] = qty[*,*,i]*tmp
                          err[*,*,i] = err[*,*,i]*tmp
                      endfor
                   endelse
                endif else begin
                   qty = temporary(qty)*tmp
                   err = temporary(err)*tmp
                endelse
                tmp = 1    ;free up the memory
             endif else begin  ;change the menu to absorption->None
                abspNon  = widget_info(self.tlb,find_by_uname='abspNone')
                abspAnn  = widget_info(self.tlb,find_by_uname='abspAnnulus')
                abspCyl  = widget_info(self.tlb,find_by_uname='abspCylinder')
                abspSph  = widget_info(self.tlb,find_by_uname='abspSphere')
                abspFla  = widget_info(self.tlb,find_by_uname='abspFlat')
                dm_toggle_menubut,check=abspNon,uncheck=[abspCyl,abspAnn,abspSph,abspFla]
                (*(self.absp_info))[0] = -1
             endelse
             widget_control,/hourglass
          endif
       endif
       ;subtract empty can file
       if ptr_valid(self.bgdata) then begin
          if finite(self.ssfac) then ssfac=self.ssfac $
          else ssfac = 1.0     ;self-shielding factor
          if self.samp_typ ge 2 then begin    ;diffuse scattering type
             if n_bgfile eq 1 then begin
                if self.samp_typ eq 2 then begin
                   for i=0L,n_file-1L do begin
                       qty[*,i] = qty[*,i]-bgqty[*,0]*ssfac
                       err[*,i] = sqrt(err[*,i]^2+(bgerr[*,0])^2*(ssfac*ssfac))
                   endfor
                endif else begin
                   for i=0L,n_file-1L do begin
                       qty[*,*,i] = qty[*,*,i]-bgqty[*,*,0]*ssfac
                       err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr[*,*,0])^2*(ssfac*ssfac))
                       if tchan_count ne 0 then begin   ;avoid double-subtraction
                          qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty[tchan_ind,*,0]*ssfac
                          ;error bar is not changed
                       endif
                   endfor
                endelse
             endif else begin
                isort = sort(bgang)
                bgang = bgang[isort]
                if self.samp_typ eq 2 then begin
                   for i=0L,n_file-1L do begin
                       index    = 0>(value_locate(bgang,(*self.dataStrPtr).rotation[i]))
                       qty[*,i] = qty[*,i]-bgqty[*,isort[index]]*ssfac
                       err[*,i] = sqrt(err[*,i]^2+(bgerr[*,isort[index]])^2*(ssfac*ssfac))
                   endfor
                endif else begin
                   for i=0L,n_file-1L do begin
                       index      = 0>(value_locate(bgang,(*self.dataStrPtr).rotation[i]))
                       qty[*,*,i] = qty[*,*,i]-bgqty[*,*,isort[index]]*ssfac
                       err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr[*,*,isort[index]])^2*(ssfac*ssfac))
                       if tchan_count ne 0 then begin   ;avoid double-subtraction
                          qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty[tchan_ind,*,isort[index]]*ssfac
                          ;error bar is not changed
                       endif
                   endfor
                endelse
             endelse
          endif else begin
             qty = temporary(qty)-(bgqty)*ssfac
             err = sqrt(temporary(err)^2+(bgerr)^2*(ssfac*ssfac))
             if tchan_count ne 0 then begin
                qty[tchan_ind,*] = qty[tchan_ind,*]+(bgqty[tchan_ind,*])*ssfac
                ;error bar is not changed
             endif
          endelse
          bgqty = 0 & bgerr = 0    ;free up the memory
       endif
       ;subtract empty can file set 2
       if ptr_valid(self.bgdata2) then begin
          if self.samp_typ ge 2 then begin    ;diffuse scattering type
             if n_bgfile2 eq 1 then begin
                if self.samp_typ eq 2 then begin
                   for i=0L,n_file-1L do begin
                       qty[*,i] = qty[*,i]-bgqty2[*,0]*ssfac
                       err[*,i] = sqrt(err[*,i]^2+(bgerr2[*,0])^2*(ssfac*ssfac))
                   endfor
                endif else begin
                   for i=0L,n_file-1L do begin
                       qty[*,*,i] = qty[*,*,i]-bgqty2[*,*,0]*ssfac
                       err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr2[*,*,0])^2*(ssfac*ssfac))
                       if tchan_count ne 0 then begin   ;avoid double-subtraction
                          qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty2[tchan_ind,*,0]*ssfac
                          ;error bar is not changed
                       endif
                   endfor
                endelse
             endif else begin
                isort  = sort(bgang2)
                bgang2 = bgang2[isort]
                if self.samp_typ eq 2 then begin
                   for i=0L,n_file-1L do begin
                       index    = 0>(value_locate(bgang2,(*self.dataStrPtr).rotation[i]))
                       qty[*,i] = qty[*,i]-bgqty2[*,isort[index]]*ssfac
                       err[*,i] = sqrt(err[*,i]^2+(bgerr2[*,isort[index]])^2*(ssfac*ssfac))
                   endfor
                endif else begin
                   for i=0L,n_file-1L do begin
                       index      = 0>(value_locate(bgang2,(*self.dataStrPtr).rotation[i]))
                       qty[*,*,i] = qty[*,*,i]-bgqty2[*,*,isort[index]]*ssfac
                       err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr2[*,*,isort[index]])^2*(ssfac*ssfac))
                       if tchan_count ne 0 then begin   ;avoid double-subtraction
                          qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty2[tchan_ind,*,isort[index]]*ssfac
                          ;error bar is not changed
                       endif
                   endfor
                endelse
             endelse
          endif else begin
             qty = temporary(qty)-(bgqty2)*ssfac
             err = sqrt(temporary(err)^2+(bgerr2)^2*(ssfac*ssfac))
             if tchan_count ne 0 then begin
                qty[tchan_ind,*] = qty[tchan_ind,*]+(bgqty2[tchan_ind,*])*ssfac
                ;error bar is not changed
             endif
          endelse
          bgqty2 = 0 & bgerr2 = 0    ;free up the memory
       endif
       ;efficiency adjustment
       if ptr_valid(self.eff) then begin
          if self.samp_typ ge 2 then begin    ;diffuse scattering type
             if self.samp_typ eq 2 then begin
                for i=0L,n_file-1L do begin
                    qty[*,i] = qty[*,i]/(*self.eff)
                    err[*,i] = err[*,i]/(*self.eff)
                endfor
             endif else begin
                for i=0L,n_file-1L do begin
                    for j=0L,nen-1L do begin
                        qty[j,*,i] = qty[j,*,i]/(*self.eff)
                        err[j,*,i] = err[j,*,i]/(*self.eff)
                    endfor
                endfor
             endelse
          endif else begin
             for i=0L,nen-1L do begin
                 qty[i,*] = qty[i,*]/(*self.eff)
                 err[i,*] = err[i,*]/(*self.eff)
             endfor
          endelse
       endif
       ;Debye-Waller adjustment
       if (self.intn_typ eq 4) or (self.intn_typ eq 5) then begin   
          ;ask for u^2
          oldu2 = 0. & u2_existed = 0b
          if n_tags((*self.dataStrPtr).info) ne 0 then begin
             u2names = tag_names((*self.dataStrPtr).info)
             index  = where(u2names eq 'DWU2',count)
             if count ne 0 then begin
                oldu2 = (*self.dataStrPtr).info.dwu2
                u2_existed = 1b
             endif
          endif
          notvalid = 1b
          while notvalid do begin
             u2 = dm_dialog_input('u^2:',/float,default=oldu2,dialog_parent=self.tlb,$
                    title='Debye-Waller correction: exp(u^2*Q^2) will be applied.')
             widget_control,/hourglass
             if finite(u2) then $
                if u2 ge 0 then begin
                   if u2_existed then (*self.dataStrPtr).info.dwu2=u2
                   notvalid = 0b ;valid u^2 given
                endif
          endwhile
       endif
       ;kfactor adjustment
       if (self.intn_typ eq 1) or (self.intn_typ eq 3) or (self.intn_typ eq 4) $
          or (self.intn_typ eq 5) then begin   ;detailed balance factor
          ;ask for a sample temperature
          old = 300. & t_existed = 0b
          if n_tags((*self.dataStrPtr).info) ne 0 then begin
             tnames = tag_names((*self.dataStrPtr).info)
             index  = where(tnames eq 'TEMPERATURE',count)
             if count ne 0 then begin
                old = (*self.dataStrPtr).info.temperature
                t_existed = 1b
             endif
          endif
          notvalid = 1b
          while notvalid do begin
             temp = dm_dialog_input('sample temperature:',/float,default=old,dialog_parent=self.tlb,$
                    title='Detailed blance calculation')
             widget_control,/hourglass
             if finite(temp) then $
                if temp gt 0 then begin
                   if t_existed then (*self.dataStrPtr).info.temperature=temp
                   notvalid = 0b ;valid temperature given
                endif
          endwhile
          ;calculate detailed balance factor
          kb   = 8.617342e-2    ;Boltzmann constant in mev.k^-1
          dbal = exp(-en/(2.0*kb*temp))
       endif
       case self.intn_typ of    ;same case for ((self.ftype eq 2) and (self.lamcorrnxspe eq 0)) 
          0:  tmp = (*self.kfactor)^4                 ;S(Q,omega)
          1:  tmp = (*self.kfactor)^4*dbal            ;Symmetrized S(Q,omega)
          2:  tmp = (*self.kfactor)^3                 ;d2sigma/dOmegadEf
          3:  tmp = (*self.kfactor)^4*(1-dbal^2)      ;Chi(Q,omega)
          4:  tmp = (*self.kfactor)^4*(1-dbal^2)*en   ;GDOS
          5:  tmp = (*self.kfactor)^3*(1-dbal^2)*en   ;VDOS   
          6:  tmp = 1.0                               ;no kfactor adjustment
         else:tmp = 1.0                               ;no kfactor adjustment
       endcase
       if ((self.ftype eq 2) and (self.lamcorrnxspe eq 1)) then begin ;very VERY unlikely case
          case self.intn_typ of
            0:  tmp = (*self.kfactor)^3               ;S(Q,omega)
            1:  tmp = (*self.kfactor)^3*dbal          ;Symmetrized S(Q,omega)
            2:  tmp = (*self.kfactor)^2               ;d2sigma/dOmegadEf
            3:  tmp = (*self.kfactor)^3*(1-dbal^2)    ;Chi(Q,omega)
            4:  tmp = (*self.kfactor)^3*(1-dbal^2)*en ;GDOS
            5:  tmp = (*self.kfactor)^2*(1-dbal^2)*en ;VDOS   
            6:  tmp = 1.0                             ;no kfactor adjustment
            else:tmp = 1.0                            ;no kfactor adjustment
          endcase
       endif
       if ((self.ftype eq 2) and (self.lamcorrnxspe eq 2)) then begin
          case self.intn_typ of
            0:  tmp = (*self.kfactor)                 ;S(Q,omega)
            1:  tmp = (*self.kfactor)*dbal            ;Symmetrized S(Q,omega)
            2:  tmp = 1.0                             ;d2sigma/dOmegadEf
            3:  tmp = (*self.kfactor)*(1-dbal^2)      ;Chi(Q,omega)
            4:  tmp = (*self.kfactor)*(1-dbal^2)*en   ;GDOS
            5:  tmp = (1-dbal^2)*en                   ;VDOS   
            6:  tmp = 1.0                             ;no kfactor adjustment
            else:tmp = 1.0                            ;no kfactor adjustment
          endcase
       endif
       if ((self.ftype eq 2) and (self.lamcorrnxspe eq 3)) then begin
          case self.intn_typ of
            0:  tmp = 1.0                             ;S(Q,omega)
            1:  tmp = dbal                            ;Symmetrized S(Q,omega)
            2:  tmp = 1.0/(*self.kfactor)             ;d2sigma/dOmegadEf
            3:  tmp = 1-dbal^2                        ;Chi(Q,omega)
            4:  tmp = (1-dbal^2)*en                   ;GDOS
            5:  tmp = (1-dbal^2)*en/(*self.kfactor)   ;VDOS   
            6:  tmp = 1.0                             ;no kfactor adjustment
            else:tmp = 1.0                            ;no kfactor adjustment
          endcase
       endif      
       if self.eint_yn and (self.samp_typ ne 2) then $
          tmp = temporary(tmp)*ewid
       ;energy dependent detector efficiency adjustment
       if (self.edep_yn eq 1) and (self.instrgeom eq 0) then begin
          sigma_acoeff = 0.4466*det_p/6.0  ; (from Jeremy in cm-1/lambda_f) absorption factor of He3 gas
          sigma_s      = 0.0008*det_p/6.0  ; (from Jeremy in cm-1) sigma_s+
          wl0   = sqrt(81.8042/(*self.dataStrPtr).eief) ;incident wavelength
          sumtd = sigma_acoeff*wl0*(*self.kfactor)+sigma_s
          ;deteff_inv=1.0/(1.0-exp(-det_thick*sumtd))
          deteff_inv = sumtd/(sumtd-sigma_s)/(1.0-exp(-det_thick*sumtd))    ;from Jeremy's Equation 9
          sumtd = 0
          tmp   = temporary(tmp)*temporary(deteff_inv)
       endif
       if self.samp_typ ge 2 then begin
          if (self.samp_typ eq 2) or (n_elements(tmp) eq 1) then begin      
             qty = temporary(qty)*tmp       ;tmp is a scalor here, qty is a [ndet,nfile] arrray
             err = temporary(err)*tmp
          endif else begin                  ;tmp is a [nen] array, qty is a [nen,ndet,nfile] array
             for i=0L,ndet-1L do begin
                 for j=0L,n_file-1L do begin
                     qty[*,i,j] = qty[*,i,j]*tmp
                     err[*,i,j] = err[*,i,j]*tmp
                 endfor
             endfor
          endelse
       endif else begin
          if n_elements(tmp) ne 1 then begin
             for i=0L,ndet-1L do begin      ;tmp is a [nen] array, qty is a [nen,ndet] array
                 qty[*,i] = qty[*,i]*tmp
                 err[*,i] = err[*,i]*tmp
             endfor
          endif else begin
             qty = temporary(qty)*tmp
             err = temporary(err)*tmp
          endelse
       endelse
       tmp = 0    ;free up the memory
       ;energy rebin
       if (self.e_bin[self.samp_typ] gt 1) and (self.samp_typ lt 2) then begin
          dm_grid_ebin,self.e_bin[self.samp_typ],qty,err,en,debug=debug,avgsum=(self.bin_avgsum eq 1)
          dm_grid_ebin,self.e_bin[self.samp_typ],ewid
          if (self.samp_typ eq 0) and (self.ftype eq 0) then $
             dm_grid_ebin,self.e_bin[self.samp_typ],time
       endif

       ;calculate dark angle correction factor
       if (self.samp_typ ge 2) and self.dcor_yn and ptr_valid(self.dcor_info) then begin
          tmp_qty = (*self.dcor_info).factor
          tmp_ang = (*self.dcor_info).a2
          isort   = sort(tmp_ang)
          tmp_ang = tmp_ang[isort]
          tmp_qty = tmp_qty[isort]
          tmp_fac = fltarr(n_file)
          for i=0L,n_file-1L do begin
              index      = (0L)>(value_locate(tmp_ang,(*self.dataStrPtr).rotation[i]))
              tmp_fac[i] = tmp_qty[index]
          endfor
          if self.samp_typ eq 2 then begin
             for i=0L,n_file-1L do begin
                 qty[*,i] = qty[*,i]*(tmp_fac[i])
                 err[*,i] = err[*,i]*(tmp_fac[i])
             endfor
          endif else begin
             for i=0L,n_file-1L do begin
                 qty[*,*,i] = qty[*,*,i]*(tmp_fac[i])
                 err[*,*,i] = err[*,*,i]*(tmp_fac[i])
             endfor
          endelse
       endif 
    endif

    ;mask detectors
    if ptr_valid(self.mask) then begin
       tmp = bytarr(ndet)
       tmp[*self.mask] = 1b
       ind = where(tmp eq 0,count) & tmp=0
       if count ne 0 then begin
          if ptr_valid(self.dataStrPtr) then begin
             if self.samp_typ ge 2 then begin
                if self.samp_typ eq 2 then begin
                   qty = qty[ind,*]
                   err = err[ind,*]
                endif else begin
                   qty = qty[*,ind,*]
                   err = err[*,ind,*]
                endelse
             endif else begin
                qty = qty[*,ind]
                err = err[*,ind]
             endelse
          endif
          det_tt   = det_tt[ind]
          det_psi  = det_psi[ind]
          ndet     = count
          ind = 0
       endif else begin
          ok = dialog_message('All detectors were masked. The mask has been cleared.',dialog_parent=self.tlb)
          self->clear_mask
       endelse   
    endif

    if self.samp_typ ge 1 then begin
       ;direct -> reciprocal basis vectors
       abc_r = dm_basis_r(self.latt_parm,error=error)
       if error then return  ;error encountered, no further calcuation
       ;orthogonalize and normalize the crystal orientation axes
       if norm(crossp(self.orie_u,self.orie_v)) eq 0 then begin    ;parallel
          ok = dialog_message('Please check cyrstal orientation.',/error,dialog_parent=self.tlb)
          return          ;no further calculation
       endif
       uvw      = dblarr(3,3)
       uvw[*,0] = abc_r#self.orie_u
       uvw[*,1] = abc_r#self.orie_v
       uvw[*,0] = uvw[*,0]/norm(uvw[*,0],/double)
       uvw[*,1] = uvw[*,1]-uvw[*,0]#(transpose(uvw[*,1])#uvw[*,0])
       uvw[*,1] = uvw[*,1]/norm(uvw[*,1],/double)
       for i=0,2 do uvw[i,2]=uvw[(i+1) mod 3,0]*uvw[(i+2) mod 3,1]-uvw[(i+2) mod 3,0]*uvw[(i+1) mod 3,1]
                 ;cross product, not using crossp() for higher precision
       ;rewrite reciprocal lattice parameters in the (u,v,w) frame
       abc_r = transpose(uvw)#(temporary(abc_r))
    endif

    ptr_free,self.projStrPtr,self.powd_line
    
    max_en = max(en,min=min_en)
    if self.samp_typ eq 2 then begin
       min_en = self.e_range[0]
       max_en = self.e_range[1]
    endif

    case self.samp_typ of
     0:begin  ;powder
       ndat     = n_elements(det_tt)
       ;handle the inherent masks in SPE file data
       if ptr_valid(self.dataStrPtr) and (self.ftype eq 1 or self.ftype eq 2) then begin
          ind   = where(finite(qty[0,*]),count)
          if count ne ndat then begin
             qty     = qty[*,ind] 
             err     = err[*,ind]
             det_tt  = det_tt[ind]
             det_psi = det_psi[ind]
             ndat    = count
          endif
       endif
       ;combine equivalent detectors
       if self.negative2th eq 0 then det_tt=abs(temporary(det_tt))
       ind      = sort(det_tt)
       if ptr_valid(self.dataStrPtr) then begin
          qty   = qty[*,ind]
          err   = err[*,ind]
       endif
       det_tt   = det_tt[ind]
       det_psi  = det_psi[ind]
       ind      = uniq(det_tt)
       ndet     = n_elements(ind)
       if ndet ne ndat then begin
          j = 0L
          for i=0L,ndet-1 do begin
              if ptr_valid(self.dataStrPtr) then begin
                 qtmp = 0.0
                 etmp = 0.0
              endif
              itmp = 0L
              while(det_tt[j] eq det_tt[ind[i]]) do begin
                 if ptr_valid(self.dataStrPtr) then begin
                    qtmp = temporary(qtmp)+qty[*,j]
                    etmp = temporary(etmp)+err[*,j]*err[*,j]
                    itmp = itmp+1L
                 endif
                 j = j+1L
                 if j eq ndat then break
              endwhile
              if self.bin_avgsum eq 1 then itmp=1L
              if ptr_valid(self.dataStrPtr) then begin
                 qty[*,i] = qtmp/(itmp>1)
                 err[*,i] = sqrt(etmp)/(itmp>1)
              endif
          endfor
          if ptr_valid(self.dataStrPtr) then begin
             qtmp  = 0 & etmp = 0  ;free up the memory
             qty   = qty[*,0:ndet-1]
             err   = err[*,0:ndet-1]
          endif
          det_tt   = det_tt[ind]
          det_psi  = det_psi[ind]
       endif
       ind = 0
       Q = dm_proj_spec_samp(eief,en,det_tt,det_psi,type=0,debug=debug,instrgeom=self.instrgeom)   ;Q[nen,ndet]
       if self.intn_typ eq 4 then begin ;GDOS
          factor = 0.48259649/Q^2
          qty = qty*factor*exp(u2*Q^2)
          err = err*temporary(factor)
       endif
       if self.intn_typ eq 5 then begin  ;VDOS
          tthmax = max(det_tt,min=tthmin)
          if self.instrgeom eq 0 then $
             efei = eief-en $
          else $
             efei = eief+en
          root = sqrt(eief*efei)
          q4maxmin = (eief+efei-2.0*root*cos(tthmax*!dtor))^2-(eief+efei-2.0*root*cos(tthmin*!dtor))^2
          factor = (1./q4maxmin)#sin(det_tt*!dtor)
          qty = qty*factor*exp(u2*Q^2)
          err = err*temporary(factor)
       endif
       if ptr_valid(self.powd_qvalue) and ((min_en le 0) and (max_en ge 0)) then begin  ;powder line
          maxQ = max(sqrt(eief/2.0721246)*sin(det_tt/2*!dtor)*2.,min=minQ) & yval = [-2e27,1e-20,2e27]
          for i=0L,n_elements(*self.powd_qvalue)-1 do begin
             if ((*self.powd_qvalue)[i] gt maxQ) or ((*self.powd_qvalue)[i] lt minQ) then continue
             if n_elements(powd_x) ne 0 then begin
                powd_x = [powd_x,(*self.powd_qvalue)[i],(*self.powd_qvalue)[i],(*self.powd_qvalue)[i]]
                powd_y = [powd_y,yval]
             endif else begin
                powd_x = [(*self.powd_qvalue)[i],(*self.powd_qvalue)[i],(*self.powd_qvalue)[i]]
                powd_y = yval 
             endelse
             yval[[0,2]] = -yval[[0,2]]
          endfor
          if n_elements(powd_x) ne 0 then self.powd_line=ptr_new([[powd_x],[powd_y]],/no_copy)
       endif
       end
     else:begin  ;single crystal and diffuse scattering
       if self.samp_typ eq 1 then begin
          a2 = self.orie_psi & type = 1
       endif else type=2 
       ;reconstruct the viewing axis and check orthogonality
       U1 = abc_r#self.view_u1
       U2 = abc_r#self.view_u2
       norm1 = norm(U1,/double) & norm2 = norm(U2,/double)
       if self.view_ortharbi eq 0 then begin    ;check orthogonality
          if abs(transpose(U1)#U2/norm1/norm2) gt 1e-8 then begin      ;not orthogonal
             ok = dialog_message('Viewing axes are not orthogonal.',/error,dialog_parent=self.tlb)
             return  ;no further calculation
          endif
       endif else begin
          ;make sure the viewing axes are not parallel to each other
          if norm(crossp(self.view_u1,self.view_u2)) eq 0 then begin
             ok = dialog_message('Parallel viewing axes',/error,dialog_parent=self.tlb)
             return  ;no further calculation
          endif
       endelse
       if self.extravaxis_yn then begin
          U3    = abc_r#self.view_u3
          norm3 = norm(U3,/double)
          if self.view_ortharbi eq 0 then begin    ;check orthogonality
             if max(abs([transpose(U2)#U3/norm2/norm3,transpose(U3)#U1/norm3/norm1])) gt 1e-8 then begin   ;not orthogonal
                ok = dialog_message('Viewing axes are not orthogonal.',/error,dialog_parent=self.tlb)
                return   ;no further calculation
             endif
          endif else begin
             ;make sure no viewing axes are parallel to each other
             if (norm(crossp(self.view_u1,self.view_u3)) eq 0) or (norm(crossp(self.view_u2,self.view_u3)) eq 0) then begin
                ok = dialog_message('Parallel viewing axes',/error,dialog_parent=self.tlb)
                return   ;no further calculation
             endif
          endelse
       endif else $
          U3 = [0.0,0.0,0.0]
       
       Q   = dm_proj_spec_samp(eief,en,det_tt,det_psi,psi=a2,type=type,debug=debug,is_spe=is_spe,diffetyp=(self.samp_typ eq 3),instrgeom=self.instrgeom)
                                                        ;Q[3,nen,ndet,nfile] in sample frame
       QV  = dm_proj_view(Q,en,U1,U2,U3,debug=debug)    ;project onto viewing axis
       if (~ self.extravaxis_yn) then begin
          tmp = size(QV,/dimension)
          case n_elements(tmp) of  ;throw away the third projection
               2:    QV = QV[0:1,*]
               3:    QV = QV[0:1,*,*]
               4:    QV = QV[0:1,*,*,*]
               else: QV = QV[0:1,*]
          endcase
          fold = [self.view_u1fold,self.view_u2fold]
          fcnt = [self.view_u1foldcenter,self.view_u2foldcenter]
       endif else begin
          fold = [self.view_u1fold,self.view_u2fold,self.view_u3fold]
          fcnt = [self.view_u1foldcenter,self.view_u2foldcenter,self.view_u3foldcenter]
       endelse
       
       ;folding of viewing axis
       for i=0,n_elements(fold)-1 do begin
           if fold[i] then begin
              if finite(fcnt[i]) then begin    ;valid folding center
                 xdat  = QV[i,*,*,*]    
                 if self.foldright then begin  ;always fold to the right unless no data to the right
                    index = where(xdat lt fcnt[i],count,ncomplement=count1)
                    if count1 eq 0 then begin  ;no data to the right
                       count = 0 & index = -1
                    endif
                 endif else begin
                    ;determine on which side of the folding center that the data has wider range
                    ;fold to greater range side   
                    index = where(xdat gt fcnt[i],count,complement=index1,ncomplement=count1)
                    if (count ne 0) and (count1 ne 0) then begin
                       x_max = max(xdat[index],min=x_min)
                       p_ran = x_max-x_min
                       x_max = max(xdat[index1],min=x_min)
                       n_ran = x_max-x_min  
                       if p_ran ge n_ran then begin
                          index = temporary(index1) 
                          count = count1
                       endif else begin
                          index1 = -1
                       endelse
                    endif 
                 endelse
                 if count ne 0 then begin
                    xdat[index] = (2.*fcnt[i]-xdat[index]) & index = -1
                    QV[i,*,*,*]  = temporary(xdat)
                 endif
              endif else begin  ;invalid folding center
                 case i of  
                      0: self.view_u1fold = 0b
                      1: self.view_u2fold = 0b
                      2: self.view_u3fold = 0b
                 endcase 
                 self->my_widget_control,'viewU'+dm_to_string(i,/int)+'Fold',self.projBas,set_button=0
                 self->my_widget_control,'viewU'+dm_to_string(i,/int)+'Foldcenter',self.projBas,sensitive=0
                 ok = dialog_message('u'+dm_to_string(i,/int)+' folding is disabled. Invalid folding center.',/error,dialog_parent=self.tlb)
              endelse
           endif
       endfor

       ;Al, Cu, or Stainless Steel powder line calculation
       if ((self.samp_typ eq 2) or (self.disp_perp eq 2) or (~ ptr_valid(self.dataStrPtr))) $
          and ptr_valid(self.powd_qvalue) and ((min_en le 0) and (max_en ge 0)) then begin
          maxQ = max(sqrt(eief/2.0721246)*sin(det_tt/2*!dtor)*2.,min=minQ)
          tmp  = transpose([[cos(findgen(361)*!dtor)],[sin(findgen(361)*!dtor)],[fltarr(361)]])
          for i=0L,n_elements(*self.powd_qvalue)-1 do begin
              if ((*self.powd_qvalue)[i] gt maxQ) or ((*self.powd_qvalue)[i] lt minQ) then continue
              Vpow = dm_proj_view((*self.powd_qvalue)[i]*tmp,0,U1,U2,U3)
              x = reform(Vpow[0,*]) & y = reform(Vpow[1,*])
              if n_elements(lines) ne 0 then lines=[[[lines]],[[[x],[y]]]] $
              else lines=[[x],[y]]
          endfor
          if n_elements(lines) ne 0 then self.powd_line=ptr_new(lines,/no_copy)
       endif
       end
    endcase
    
    ;user function 
    if strlen(self.userf_str) ne 0 then begin
       if self.userfwarning then begin
          if lmgr(/runtime) then usertmp = 'User Macro' else usertmp = 'User Macro/Function/Procedure'
          mesg = usertmp+' is specified.'
          self.userfwarning = 0b
       endif
       tmp = 1-execute(self.userf_str,1,1)
       if tmp then begin
          ok = dialog_message(!error_state.msg,/error,dialog_parent=self.tlb)
          return  ;error encountered
       endif
       if n_elements(err) ne n_elements(qty) then begin ;in case qty and err are not both defined in user functions
          if n_elements(err) eq 1 then begin
             err = replicate(err,size(qty,/dimension))
          endif else begin
             help,qty,err,output=tmp
             ok = dialog_message(['Incompatible array sizes of intensity and error bar.','',tmp],/error,dialog_parent=self.tlb)
             return
          endelse
       endif
    endif

    ;store projection
    if self.samp_typ eq 0 then $    ;powder
       self.projStrPtr = ptr_new({Q:temporary(Q),en:temporary(en),two_theta:temporary(det_tt),$
            qty:temporary(qty),err:temporary(err),ewid:temporary(ewid),time:temporary(time)}) $
    else $                          ;single crystal
       self.projStrPtr = ptr_new({V:temporary(QV),en:temporary(en),ewid:temporary(ewid),$
            two_theta:temporary(det_tt),qty:temporary(qty),err:temporary(err)})

    ;allow plots
    if (~ keyword_set(debug)) then begin
       self->set_display_button 
       if (~ ptr_valid(self.dataStrPtr)) then self->my_widget_control,'dispTab',self.dispBas,$
          set_tab_current=([2,3])[(self.samp_typ eq 1) or (self.samp_typ eq 2 and self.extravaxis_yn) or (self.samp_typ eq 3)]
    endif else $
       print,'**Projection calculation finished in ',systime(/sec)-current,' sec.'
    if ptr_valid(self.dataStrPtr) and (self.bin_avgsum eq 1) and (self.e_bin[self.samp_typ] gt 1) and (self.samp_typ lt 2)then begin
       if n_elements(mesg) eq 0 then mesg='The current binning method is SUM.' $
       else mesg = [mesg,'The current binning method is SUM.']
    endif   
    if n_elements(mesg) ne 0 then ok = dialog_message([mesg,'This message is only a reminder.'],dialog_parent=self.tlb)
    self.error = 0       ;clear the error flag
end