; $Id: dm_ereslconv.pro,v 1.6 2010/06/04 18:15:47 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dm_ereslconv
;
; PURPOSE:
;  This program convolutes data with DCS energy resolution
;
; CATEGORY:
;  DCS data tools
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  June, 2010
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

function dm_eres_convol,x,y,fwhm
    nx     = n_elements(x)
    dx     = x[1:nx-1] - x[0:nx-2]
    dx     = [dx[0],dx]
    ; Inflate the x,dx,y vector to an array that is NX by NX.
    dxmat  = rebin(dx,nx,nx,/sample)
    xmat   = rebin(x,nx,nx,/sample)
    ymat   = rebin(y,nx,nx,/sample)
    sigma  = fwhm/2.354
    ; Inflate the width from an NX element vector to an array
    ; that is NX by NX.
    sigmat = rebin(sigma,nx,nx,/sample)
    ; Calculate the difference matrix for the argument in the Gaussian.
    diff   = transpose(xmat)-xmat
    ; Resolution matrix
    rmat   = 1/sqrt(2.0*!dpi*sigmat^2)*exp(-0.5*(diff/sigmat)^2)
    ; Integral
    return,total(ymat*rmat*dxmat,1)
end

pro dm_ereslconv_readfile,file,data=data,error=error,dialog_parent=dialog_parent
    error = 0
    
    mesg  = widget_base(title='Loading data ...',/col,/base_align_center,group_leader=dialog_parent)
    label = widget_label(mesg,value='Loading '+file[0]+'. Please wait ...',/align_center)
    dm_center_kid,mesg,dialog_parent
    widget_control,mesg,/realize,map=1
    
    catch, myerror
    if myerror ne 0 then begin
       error = 1
       catch,/cancel
       ok = dialog_message(!error_state.msg,/error)
       if n_elements(unit) ne 0 then free_lun,unit
       if widget_info(mesg,/valid_id) then widget_control,mesg,/destroy
       return
    end
    
    openr,unit,file,/get_lun,error=error
    if error ne 0 then begin
       ok = dialog_message("Can't read file "+file+'.',dialog_parent=dialog_parent)
       widget_control,mesg,/destroy
       return
    endif
    
    data = 0 & tmp = temporary(data) ;destroy data
    tmp = '' & line = 0
    
    while (~ eof(unit)) do begin
          readf,unit,tmp & tmp = strtrim(tmp,2)
          if strlen(tmp) eq 0 then continue
          tmp1 = strsplit(tmp,' ,'+string(9b),/extract)
          tmp1 = dm_to_number(tmp1,/float,count=count)
          if total(finite(tmp1)) eq count then $
             if n_elements(data) eq 0 then data=tmp1 else data=[[data],[tmp1]] 
    endwhile
    free_lun,unit
    widget_control,mesg,/destroy
    if n_elements(data) gt 1 then data = transpose(data)
end

pro dm_ereslconv_writefile,file=file,data=data,path=path,dialog_parent=dialog_parent
    file = dm_choose_file('',/write,dialog_parent=dialog_parent,title='Save convolution result as an ASCII file.',path=path,file=file)
    if strlen(file) eq 0 then return
    openw,unit,file,/get_lun,error=openerr
    if openerr ne 0 then begin
       ok = dialog_message("Can't write in "+file[0],/error,dialog_parent=dialog_parent)
    endif else begin
       widget_control,/hourglass  ;busy signal
       for i=0L,n_elements(data[*,0])-1 do $
           printf,unit,data[i,0],' ',data[i,1]                   
       free_lun,unit
    endelse
end

pro dm_ereslconv_resetbutton,pState
    widget_control,(*pState).calcBut, sensitive=ptr_valid((*pState).xPtr)
    widget_control,(*pState).saveBut, sensitive=ptr_valid((*pState).convPtr)
    widget_control,(*pState).clearBut,sensitive=ptr_valid((*pState).xPtr)
    widget_control,(*pState).printBut,sensitive=ptr_valid((*pState).xPtr)
end

pro dm_ereslconv_event,event
    compile_opt IDL2    ;,strictarrsubs
                        ;idl2=defint32,strictarr
                        ;strictarr:   [] to index array
                        ;strictarrsubs: error when out-of-range indices,IDL5.6 or after
    WIDGET_CONTROL,/HOURGLASS
    widget_control,event.handler,get_uvalue=pState

    if (*pState).enotebook then $
       IF DAVE_SET_FOCUS(EVENT) THEN RETURN

    widget_type = widget_info(event.id,/type)
    case widget_type of
       3:    begin                                               ;text box
             widget_control,event.id,get_value=tmp_value
             tmp_value = strtrim(tmp_value,2)
             end
       8:    tmp_value = widget_info(event.id,/droplist_select)  ;droplist
       12:   tmp_value = event.index                             ;combobox
       else: tmp_value = 0
    endcase

    ;catch and ignore all errors in this program
    catch, myerror
    if myerror ne 0 then begin
       ok = dialog_message(dialog_parent=(*pState).tlb,!error_state.msg,/error)
       catch,/cancel
       return
    end
 
    case (event.id) of
       (*pState).tlb:    widget_control,event.id,scr_xsize=(*pState).geom[0],scr_ysize=(*pState).geom[1]
       (*pState).loadBut: begin
            (*pState).filesel->getproperty,path=open_path,file=open_file,sep=sep
            if n_elements(open_file) eq 0 then begin
               ok = dialog_message('Please select an input data file first.',/error,dialog_parent=(*pState).tlb)
               break
            endif
            (*pState).path     = open_path
            (*pState).filename = open_file
            
            dm_ereslconv_readfile,open_path+sep+open_file,data=data,error=error,dialog_parent=(*pState).tlb
            if error eq 0 and n_elements(data) gt 1 then begin 
               if ptr_valid((*pState).xPtr) then begin       
                  ptr_free,(*pState).xPtr,(*pState).yPtr
                  if n_elements(data[0,*]) eq 3 then $
                     (*pState).plotWin->add_plot,data[*,0],data[*,1],yerr=data[*,2],linestyle='no line',$
                        color=(*pState).cols[(*pState).colind],legend='Input',psym='circle',layer=0 $
                  else $
                     (*pState).plotWin->add_plot,data[*,0],data[*,1],linestyle='no line',$
                        color=(*pState).cols[(*pState).colind],legend='Input',psym='circle',layer=0
               endif else begin
                  if n_elements(data[0,*]) eq 3 then $
                     (*pState).plotWin->setproperty,xdat=data[*,0],ydat=data[*,1],yerr=data[*,2],linestyle='no line',$
                        color=(*pState).cols[(*pState).colind],legend='Input',psym='circle',layer=0 $
                  else $
                     (*pState).plotWin->setproperty,xdat=data[*,0],ydat=data[*,1],linestyle='no line',$
                        color=(*pState).cols[(*pState).colind],legend='Input',psym='circle',layer=0 
               endelse
               (*pState).plotWin->draw
               (*pState).colind = ((*pState).colind+1) mod 6
               (*pState).xPtr = ptr_new(data[*,0])
               (*pState).yPtr = ptr_new(data[*,1])
               dm_ereslconv_resetbutton,pState
            endif
          end
       (*pState).calcBut: begin
            if (~ finite ((*pState).wl)) then begin
               ok = dialog_message('Invalid wavelength.',/error,dialog_parent=(*pState).tlb)
               break
            endif else if (*pState).wl le 0 then begin
               ok = dialog_message('Invalid wavelength.',/error,dialog_parent=(*pState).tlb)
               break
            endif 
            if (~ finite((*pState).mspeed)) then begin
               ok = dialog_message('Invalid master speed.',/error,dialog_parent=(*pState).tlb)
               break
            endif else if (*pState).mspeed le 0 then begin
               ok = dialog_message('Invalid master speed.',/error,dialog_parent=(*pState).tlb)
               break
            endif
            
            x = *(*pState).xptr
            y = *(*pState).yptr
            
            ;select the energy range <Ei
            ei = 81.8042/((*pState).wl)^2
            index = where(x lt ei,count)
            if count eq 0 then return
    
            x = x[index]
            y = y[index]
            
            ;calculate the fwhm  
            ef   = ei-x
            wlf  = sqrt(81.8042/ef)
            fwhm = dcs_jcc_resn((*pState).wl,wlf,(*pState).reslmode,(*pState).mspeed)/1000. ;ueV->meV
    
            ;convolution
            result = dm_eres_convol(x,y,fwhm)
            
            (*pState).plotWin->add_plot,x,result,linestyle='solid',$
                      color=(*pState).cols[(*pState).colind],legend='convoluted',layer=0 
            (*pState).plotWin->draw
            (*pState).colind = ((*pState).colind+1) mod 6
            ptr_free,(*pState).convPtr
            (*pState).convPtr = ptr_new([[x],[result]])
            dm_ereslconv_resetbutton,pState
          end   
       (*pState).saveBut: begin
            dm_ereslconv_writefile,file='conv_'+(*pState).filename,data=*(*pState).convPtr,$
                path=(*pState).path,dialog_parent=(*pState).tlb
          end
       (*pState).clearBut: begin
            ptr_free,(*pState).xPtr,(*pState).yPtr,(*pState).convPtr
            (*pState).plotWin->erase
            (*pState).colind = 0
            dm_ereslconv_resetbutton,pState
          end
       (*pState).printBut: begin
            if ptr_valid((*pState).xPtr) then (*pState).plotWin->saveas,'printer'
          end
       (*pState).exitBut: begin
            widget_control,event.handler,/destroy
            return
          end
  
       (*pState).wlTxt:     (*pState).wl       = dm_to_number(tmp_value,/float)
       (*pState).reslList:  (*pState).reslmode = tmp_value+1
       (*pState).speedTxt:  (*pState).mspeed   = dm_to_number(tmp_value,/long)   

       else:
    endcase
end

pro dm_ereslconv_Exit,tlb
    widget_control,tlb,get_uvalue=pState,/no_copy
    obj_destroy,(*pState).filesel
    obj_destroy,(*pState).plotWin
    cd,(*pState).root
    ptr_free,(*pState).xPtr,(*pState).yPtr,(*pState).convPtr
    ptr_free,pState
    widget_control,tlb,/destroy
end

pro dm_ereslconv,event,workdir=workdir
    state={group_leader:                 0L, $  ;group leader
           tlb:                          0L, $  ;top level base
           geom:                    [0L,0L], $  ;geometry
           filesel:               obj_new(), $  ;file selector
           plotWin:               obj_new(), $  ;draw widget
           root:                         '', $  ;starting directory
           path:                         '', $  ;save the current directory 
           filename:                     '', $  ;save the current file name
           
           cols: ['green','red','blue','magenta','cyan','yellow'], $  ;colors
           colind:                       0s, $  ;index of current color
           
           loadBut:                      0L, $  ;load file button
           calcBut:                      0L, $  ;calculate button
           saveBut:                      0L, $  ;save result button
           clearBut:                     0L, $  ;clear button
           printBut:                     0L, $  ;print plot button
           exitBut:                      0L, $  ;exit button
           
           wlTxt:                        0L, $  ;wavelength text widget
           wl:                         1.8e, $  ;wavelength
           
           reslList:                     0L, $  ;resolution mode drop list
           reslmode:                     1s, $  ;resolution mode 1:low, 2:medium
           
           speedTxt:                     0L, $  ;master speed text widget
           mspeed:                   20000l, $  ;master speed
           
           xPtr:                  ptr_new(), $  ;pointer to x data
           yPtr:                  ptr_new(), $  ;pointer to y data
           convPtr:               ptr_new(), $  ;pointer to calculated y data
           
           title:                        '', $  ;plot title
           enotebook:                    0b  $  ;use Rob's enotebook
    }
    registerName = 'dm_ereslconv'
    if xregistered(registerName) then return  ;only allow one copy to be running at one time
    cd,current = current
    state.root = current
    if n_elements(workdir) eq 0 then workdir=current
    
    if n_elements(event) ne 0 then begin
       state.group_leader = event.top
       state.tlb  = widget_base(title='DCS Energy Resolution Convolution Utility',/row,kill_notify='dm_ereslconv_Exit',/tlb_size_event,$
         group_leader=event.top,map=0)
    endif else $
       state.tlb  = widget_base(title='DCS Energy Resolution Convolution Utility',/row,kill_notify='dm_ereslconv_Exit',/tlb_size_event,$
         map=0)
    left = widget_base(state.tlb,/col)
    state.filesel = obj_new('dm_filesel',parent=left,xsize=160,ysize=20,/frame,path=workdir)
    if !version.os_family eq 'unix' then $
       center = widget_base(state.tlb,/grid_layout,/col,/align_bottom,xpad=0,ypad=0,space=0) $
    else $
       center = widget_base(state.tlb,/grid_layout,/col,/align_bottom)

    void = widget_label(center,value='Wavelength')
    state.wlTxt    = widget_text(center,value=dm_to_string(state.wl),/editable,/all_events)
    void = widget_label(center,value='Resolution')
    state.reslList = dm_widget_droplist(center,value=['Low','Medium'],select=state.reslmode-1)
    void = widget_label(center,value='Master Speed')
    state.speedTxt = widget_text(center,value=dm_to_string(state.mspeed),/editable,/all_events)

    void = widget_label(center,value=' ')

    state.loadBut  = widget_button(center,value='Load Data')
    state.calcBut  = widget_button(center,value='Calculate')
    state.saveBut  = widget_button(center,value='Save Result')
    state.clearBut = widget_button(center,value='Clear Data')
    state.printBut = widget_button(center,value='Print')
    state.exitBut  = widget_button(center,value='Exit')
    
    if float(!version.release) ge 5.6 then begin  ;tips for some widget buttons
       widget_control,state.loadBut,tooltip='Load the (E, I) 2-column ascii data file.'
       widget_control,state.calcBut,tooltip='Calculate the convolution of the DCS energy resolution with the loaded data.'
       widget_control,state.saveBut,tooltip='Save the last convolution result in an ascii file.'  
       widget_control,state.clearBut,tooltip='Clear all the plots and data.'
       widget_control,state.printBut,tooltip='Print the plot.' 
    endif

    geom  = widget_info(state.exitBut,/geometry)
    widget_control,state.wlTxt,scr_xsize=geom.scr_xsize
    widget_control,state.reslList,xsize=geom.scr_xsize
    widget_control,state.speedTxt,scr_xsize=geom.scr_xsize

    geom  = widget_info(left,/geometry)
    ysize = geom.scr_ysize
    xsize = fix(1.35*ysize,type=3)
    state.plotWin  = obj_new('dm_plot',xsize=xsize,ysize=ysize,widgetbase=state.tlb,group_leader=state.tlb,$
        /compound,background='black',xtit='E (meV)',ytit='I (arb. unit)',legdpos=[0.05,0.85])

    dm_center_kid, state.tlb   ;centering the top level base
    widget_control,state.tlb,/realize,/map

    geom  = widget_info(state.tlb,/geometry)
    state.geom = [geom.scr_xsize,geom.scr_ysize]

    pState = ptr_new(state,/no_copy)
    widget_control,(*pState).tlb,set_uvalue=pState

    dm_ereslconv_resetbutton,pState

    if xregistered('dave',/noshow) and (n_elements(event) ne 0) then begin
       (*pState).enotebook = 1b
       RET = DAVE_SET_FOCUS((*pState).tlb)
    endif

    xmanager,registerName,(*pState).tlb,cleanup='dm_ereslconv_Exit',/no_block
end