; $Id: dm_step_bin.pro,v 1.12 2010/07/22 14:38:02 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dm_step_bin
;
; PURPOSE:
;  bin data according to given step
;
; CATEGORY:
;  general
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  July, 2010
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

; parameter:
;   in_xstep: step of binning
;   xdat:     array of xdat
; keyword:
;   checkfinite: flag for checking finite values of xdat,ydat. finite values of yerr will always be checked
;   ydat:     array of ydat, using the same binning as xdat
;   yerr:     array of yerr, using the same binning as xdat
;   uniq_val: if set, then use the given uniq values instead of the xstep to determine the bins
;   avgsum:   0:average 1:sum
;   bintr:    binning threshold, bins with fewer data points than bintr*(max number of points in a bin) will be discarded
;   conststep: 0: use the average as the resulted binning values(default) 1: constant binning intervals
;   ewid:     energy width, used when e-integrated intensity option is choosed
;             input ydat=sum(I*ewid)
;             output ydat=sum(I*ewid)/sum(ewid)
;   extlib:   0:not using external library 1:use external library
pro dm_step_bin,in_xstep,xdat,ydat=ydat,yerr=yerr,checkfinite=checkfinite,uniq_val=uniq_val,debug=debug,avgsum=avgsum,bintr=bintr,conststep=conststep,ewid=ewid,extlib=extlib
    if n_elements(avgsum) eq 0 then avgsum = 0        ;default using average in binning
    if n_elements(bintr)  eq 0 then bintr = 0.0
    if n_elements(conststep) eq 0 then conststep = 0  ;default is using average as the binned value
    if n_elements(extlib) eq 0 then extlib = 1b       ;default using external binning library if available
    if keyword_set(debug) then current=systime(/sec)
    xstep = abs(in_xstep)
    if xstep eq 0 then xstep=1.
    nxdat    = n_elements(xdat)
    ok_ydat  = (n_elements(ydat) eq nxdat)
    ok_yerr  = (n_elements(yerr) eq nxdat)
    ok_eint  = (n_elements(ewid) eq nxdat)
    ;check data format
    mesg = ''
    if (~ ok_ydat) and n_elements(ydat) gt 0 then mesg = [mesg,'xdat and ydat sizes are different.']
    if (~ ok_yerr) and n_elements(yerr) gt 0 then mesg = [mesg,'xdat and yerr sizes are different.']
    if (~ ok_eint) and n_elements(ewid) gt 0 then mesg = [mesg,'xdat and ewid sizes are different.']
    if n_elements(mesg) gt 1 then begin
       ok = dialog_message(mesg[1:n_elements(mesg)-1],/error)
       return
    endif
    ;check finite values
    indcount = 0
    if keyword_set(checkfinite) then begin
       ind = where(finite(xdat),indcount)
       if indcount eq 0 then mesg = [mesg,'xdat are all non-finite.']
       if indcount eq nxdat then begin
          ind = 0 & tmp = temporary(ind) & indcount = 0
       endif
       if ok_ydat then begin
          ind1 = where(finite(ydat),count) 
          if (count ne nxdat) and (count ne 0) then begin
             if n_elements(ind) eq 0 then begin
                ind = temporary(ind1) & indcount = count
             endif else ind = dm_common(ind,ind1,indcount,/no_copy)
          endif else begin
             if count eq 0 then mesg = [mesg,'ydat are all non-finite.']
             ind1 = 0
          endelse
       endif
    endif
    if ok_yerr then begin ;always check the finite values of yerr
       ind1 = where(finite(yerr),count) 
       if (count ne nxdat) and (count ne 0) then begin
          if n_elements(ind) eq 0 then begin
             ind = temporary(ind1) & indcount = count
          endif else ind = dm_common(ind,ind1,indcount,/no_copy)
       endif else begin
          if count eq 0 then mesg = [mesg,'yerr are all non-finite.']
          ind1 = 0
       endelse
    endif
    if n_elements(mesg) gt 1 then begin
       ok = dialog_message(mesg[1:n_elements(mesg)-1],/error)
       return
    endif
    if indcount ne 0 then begin ;remove non-finite values
       xdat = xdat[ind]
       if ok_ydat then ydat = ydat[ind]
       if ok_yerr then yerr = yerr[ind]
       if ok_eint then ewid = ewid[ind]
       nxdat = indcount & ind  = 0 
    endif
    tmp_xmax = max(xdat,min=tmp_xmin)
    xmax  = temporary(tmp_xmax)+0.5*xstep
    xmin  = temporary(tmp_xmin)-0.5*xstep
    xsize = double(xmax-xmin)/xstep 
    if xsize gt fix(xsize,type=3) then xsize=(2.>(xsize+0.5))
    if xsize eq 2 then conststep=0
    xsize = fix(xsize,type=3)
    if n_elements(uniq_val) ne 0 then begin
       index = where((uniq_val ge xmin) and (uniq_val lt xmax),count)
       if count ne 0 then uniq_val=uniq_val[index]
       index = -1
       xlookup = uniq_val[uniq(uniq_val,sort(uniq_val))]
       xlookup = [xlookup,xmax]
    endif else $
       xlookup = xmin+findgen(xsize+1)*xstep
    xsize = n_elements(xlookup)
    tmp_x = fltarr(xsize)
    if ok_ydat then tmp_y = fltarr(xsize)
    if ok_yerr then tmp_e = fltarr(xsize)
    if ok_eint then tmp_w = fltarr(xsize)
    num_x = lonarr(xsize)

    defsysv,'!DAVE_AUXILIARY_DIR',exists=exists
    if exists and extlib then begin
       filename = 'rebinnd'+'_'+!version.os+'_'+strtrim(string(!VERSION.MEMORY_BITS),2)
       shlib = (FILE_SEARCH(!DAVE_AUXILIARY_DIR+'rebin_nd'+path_sep()+filename+'*'))[0]
    endif else shlib = ''
    
    if ((strlen(shlib) gt 0) and (n_elements(uniq_val) eq 0) and ok_ydat)then begin
      if (ok_eint and ok_yerr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin1d_yew', long(nxdat),float(xdat),double(ydat),$
        double(yerr),double(ewid),long(xsize),float(xmin),float(xstep),$
        num_x, tmp_x, tmp_y,  tmp_e, tmp_w, /CDECL) 
      end       
      if (ok_eint and ~ok_yerr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin1d_yw', long(nxdat),float(xdat),double(ydat),$
        double(ewid),long(xsize),float(xmin),float(xstep),$
        num_x, tmp_x, tmp_y, tmp_w, /CDECL)      
      end       
      if (~ok_eint and ok_yerr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin1d_ye', long(nxdat),float(xdat),double(ydat),$
        double(yerr), long(xsize),float(xmin),float(xstep),$
        num_x, tmp_x, tmp_y,  tmp_e, /CDECL) 
      end       
      if (~ok_eint and ~ok_yerr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin1d_y', long(nxdat),float(xdat),double(ydat),$
        long(xsize),float(xmin),float(xstep),$
        num_x, tmp_x, tmp_y, /CDECL)      
      end    
    end else begin
      for i=0L,nxdat-1L do begin
        ix = value_locate(xlookup , xdat[i])
        tmp_x[ix] = tmp_x[ix]+xdat[i]
        if ok_ydat then tmp_y[ix] = tmp_y[ix]+ydat[i]
        if ok_yerr then tmp_e[ix] = tmp_e[ix]+yerr[i]*yerr[i]
        if ok_eint then tmp_w[ix] = tmp_w[ix]+ewid[i]
        num_x[ix] = num_x[ix]+1
      endfor
    endelse
    
    
    index = where(num_x gt 0,count,complement=zindex,ncomplement=zcount)
    if count ne 0 then begin
       if keyword_set(conststep) and n_elements(uniq_val) eq 0 then begin
          if zcount ne 0 then begin
             num_x[zindex] = 1.0
             if ok_eint then tmp_w[zindex] = 1.0
          endif
          xdat = xmin+(findgen(xsize-1)+0.5)*xstep
          num_x = num_x[0:xsize-2]
          if ok_ydat then tmp_y = tmp_y[0:xsize-2]
          if ok_yerr then tmp_e = tmp_e[0:xsize-2]
          if ok_eint then tmp_w = tmp_w[0:xsize-2]
       endif else begin
          tmp_x = tmp_x[index]
          num_x = num_x[index]
          if ok_ydat then tmp_y = tmp_y[index]
          if ok_yerr then tmp_e = tmp_e[index]
          if ok_eint then tmp_w = tmp_w[index]
          xdat  = tmp_x/num_x
       endelse
       index = -1 & zindex=-1
       if avgsum eq 1 then num_x = 1
       if ok_eint then num_x = tmp_w
       if ok_ydat then ydat  = tmp_y/num_x
       if ok_yerr then yerr  = sqrt(tmp_e)/num_x
    endif
    if (bintr gt 0.0) and (bintr lt 1.0) then begin
       min_num_x = bintr*(max(num_x)>1)
       ind = where(num_x gt min_num_x,ct1)
       if ct1 gt 0 then begin
          xdat = xdat[ind]
          if ok_ydat then ydat  = ydat[ind]
          if ok_yerr then yerr  = yerr[ind]
       endif
    endif
    if keyword_set(debug) then print,'step_bin finished in ',systime(/sec)-current,' secs.'
end