; $Id: dcs_mslice.pro,v 1.77.2.2 2010/10/04 19:10:37 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dcs_mslice
;
; PURPOSE:
;  Viewing and analyzing neutron time-of-flight data.
;
; CATEGORY:
;  dcs time-of-flight
;
; HISTORY:
;  06/2003: IDL5.6 compatible, include all functionality of matlab mslice, and windows .hlp help file
;  10/2003: IDL6.0 compatible, add diffuse scattering type, and dcs_mslice.pdf help file
;  09/2004: Make IDL dcs_mslice extendable to other TOF instruments, documentation in dcs_mslice_addftype.pdf
;  06/2005: IDL6.1 compatible, add diffuse scattering multiple energy type, rearange sample types
;  11/2005: Remove some items in the droplists, and add extra HKL range limitation
;  03/2006: Add INX file type
;  03/2007: Add grouped data fitting in powder mode, i.e. add QENS data reduction
;           modifications in background subtraction, especially the treatment of dark count files
;  07/2007: Add empty can subtraction from vanadium file in the first two sample types
;           Add automatic Bragg peak masking
;  02/2008: Compatible with inverse geometry TOF instruments
;  06/2008: Add MACS
;  11/2008: more flexible fold x-axis operation, add energy resolution convolution tool
;  05/2009: revamp viewing axes, allow more flexbility in folding and viewing
;  06/2009: add stand-alone interface for general volume data
;  07/2009: add volume plot, and change the display window to tab widgets
;  11/2009: allow two sets of empty can backgrounds, rebuild the context sensitive menu 
;  12/2009: more robust automatic masking in SPE files
;  02/2010: added NXSPE support - Andrei Savici (saviciat@ornl.gov)
;  04/2010: possible to discard bins that have few pixels mapped into them 
;           keep track of correct ki/kf powers for NXSPE files - Andrei Savici (saviciat@ornl.gov)
;  05/2010: modify the way to handle the masking in SPE and NXSPE files      
;  05/2010: add user function option 
;  09/2010: IDL 8.0 compatible
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  October, 2010
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

;files need to be included
@dm_calc_projection
@dm_choose_cutdata
@dm_choose_slicedata
@dm_choose_volumedata
@dm_choose_trajdata
@dm_det_spurion
@dm_filemenu
@dm_filetools
@dm_load_parameter
@dm_mask_detbanks
@dm_plot_cut
@dm_plot_slice
@dm_plot_volume
@dm_plot_trajectory
@dm_save_parameter
@dm_to_fit
@dcs_plotdetmask    ;for plotting masks

;kill the object
pro dcs_mslice_Exit,tlb
    widget_control,tlb,get_uvalue=self
    if obj_valid(self) then obj_destroy,self
end

;widget event handler
pro dcs_mslice_event,event
    widget_control,event.handler,get_uvalue=self
    if obj_valid(self) then self->event,event
end

;absorption correction
;varible:
;   type: -1-none 0-cylinder 1-annulus 2-sphere 3-flat plate
pro dcs_mslice::absorption,type
    abspNon = widget_info(self.tlb,find_by_uname='abspNone')
    abspCyl = widget_info(self.tlb,find_by_uname='abspCylinder')
    abspAnn = widget_info(self.tlb,find_by_uname='abspAnnulus')
    abspSph = widget_info(self.tlb,find_by_uname='abspSphere')
    abspFla = widget_info(self.tlb,find_by_uname='abspFlat')
    default = [0,0,!values.f_nan,!values.f_nan,!values.f_nan,!values.f_nan,!values.f_nan]
    if ptr_valid(self.absp_info) then begin
       ninfo = n_elements(*(self.absp_info))
       default[0:(ninfo-2)] = (*(self.absp_info))[1:(ninfo-1)]
       ptr_free,self.absp_info
    endif
    case type of
          0 :   begin   ;cylinder
                onbut  = abspCyl
                offbut = [abspNon,abspAnn,abspSph,abspFla]
                names  = ['L0(cm):','L1(cm):','R(cm):','Sampling Points:','Interpolate:']
                title  = 'Cylindrical geometry'
                end
          1 :   begin   ;annulus
                onbut  = abspAnn
                offbut = [abspNon,abspCyl,abspSph,abspFla]
                names  = ['L0(cm):','L1(cm):','Rout(cm):','Rin(cm):','Sampling Points:','Interpolate:']
                title  = 'Annular geometry'
                end
          2 :   begin   ;sphere
                onbut  = abspSph
                offbut = [abspNon,abspAnn,abspCyl,abspFla]
                names  = ['L0(cm):','L1(cm):','R(cm):','Sampling Points:','Interpolate:']
                title  = 'Spherical geometry'
                end
          3 :   begin   ;flat plate
                onbut  = abspFla
                offbut = [abspNon,abspAnn,abspCyl,abspSph]
                names  = ['L0(cm):','L1(cm):','Width(cm):','Thickness(cm):','Angle:','Sampling Points:','Interpolate:']
                title  = 'Flat plate geometry'
                end
        else:   begin   ;none
                onbut  = abspNon
                offbut = [abspAnn,abspCyl,abspSph,abspFla]
                end
    endcase
    nitem = n_elements(names)
    if nitem ne 0 then begin
       npts  = *self.absp_npts[type]
       nptst = *self.absp_nptst[type]
       if self.samp_typ ge 2 then default[1]=0 ;no need to interpolate for diffuse scattering type
       default = [default[2:(nitem-1)],default[0:1]]
       is_droplist = intarr(nitem) & is_droplist[nitem-2:nitem-1] = 1
       mesg = ['L0 is the energy independent mean free path;',$
               'L1 is the energy dependent mean free path for',$
               '2200m/s or 25.3 meV neutron.',$
               'The effective mean free path is:',$
               'L=1/{1/L0+1/(L1*sqrt[E(meV)/25.3])}' ]
       if type eq 3 then mesg = [mesg,'','The angle is from ki to the normal of the plate.',$
                                'Counterclockwise rotation is positive.']
       info = dm_dialog_input(names,default=default,/float,info=mesg,$
                is_droplist=is_droplist,droplist_content=[ptr_new(nptst),ptr_new(['No','Yes'])],$
                /return_number,title=title,xsize=140,dialog_parent=self.tlb)
       if total(finite(info)) eq nitem then begin
          self.absp_info=ptr_new([type,info[nitem-2:nitem-1],info[0:(nitem-3)]])
       endif else begin
          onbut  = abspNon
          offbut = [abspAnn,abspCyl,abspSph,abspFla]
          ok = dialog_message('Invalid inputs. Absorption correction is disabled.',dialog_parent=self.tlb,/error)
       endelse
    endif
    dm_toggle_menubut,check=onbut,uncheck=offbut
end

;adjust the viewing axis label according to hkl value
pro dcs_mslice::adjust_viewlabel,uid
    tmp = ([[self.view_u1],[self.view_u2],[self.view_u3]])[*,uid]
    if tmp[0] ne 0 then $
       label = 'H'      $
    else if tmp[1] ne 0 then $
       label = 'K'      $
    else if tmp[2] ne 0 then $
       label = 'L'      
    if n_elements(label) ne 0 then begin
       if uid eq 0 then begin
          self.view_u1lab[1] = label
          self->my_widget_control,'viewU1lab',self.projBas,set_value=label
       endif else if uid eq 1 then begin
          self.view_u2lab[1] = label
          self->my_widget_control,'viewU2lab',self.projBas,set_value=label
       endif else begin
          self.view_u3lab = label
          self->my_widget_control,'viewU3lab',self.projBas,set_value=label
       endelse
    endif
end

;change show powder line option
pro dcs_mslice::change_powderline,option,menuonly=menuoly
    self.powd_typ = 0>(option)<4
    ids    = lonarr(5)
    ids[0] = widget_info(self.tlb,find_by_uname='powderNone')
    ids[1] = widget_info(self.tlb,find_by_uname='powderAl')
    ids[2] = widget_info(self.tlb,find_by_uname='powderCu')
    ids[3] = widget_info(self.tlb,find_by_uname='powderSt')
    ids[4] = widget_info(self.tlb,find_by_uname='powderSp')
    dm_toggle_menubut,check=ids[self.powd_typ],uncheck=ids[where(indgen(5) ne self.powd_typ)]
    if keyword_set(menuonly) then return
    default = ''
    if ptr_valid(self.powd_qvalue) then for i=0,n_elements(*self.powd_qvalue)-1 do default=default+dm_to_string((*self.powd_qvalue)[i])+'  '
    ptr_free,self.powd_qvalue,self.powd_line
    fcc = sqrt([3,4,8,11,12,16,19,20,24,27,32,35,36,40,43,44,48,51,52,56,59,64,67,68,72,75,76,80,83,84,88,91,$
                96,99,100,104,107,108,115,116,120,123,128,131,132,136,139,140,144,147,148,152,155,160,163,164,$
                168,171,172,176,179,180,184,187,192,195,196,200,203,204,208,211,212,216,219,224,227,228,232,235,$
                236,243,244,248,251,256,259,260,264,267,268,272,275,276,280,283,288,291,292,296,299,300])
    case self.powd_typ of
          1: qvalue = fcc*2*!pi/4.05     ;Al a=4.05 FCC
          2: qvalue = fcc*2*!pi/3.6148   ;Cu a=3.6148  FCC
          3: qvalue = fcc*2*!pi/3.594    ;Stainless Steel a=3.594 FCC
          4: begin
             info   = 'Numbers should be separated by a space or comma.'
             qvalue = dm_dialog_input('Q ('+string('c5'XB)+'^-1):',xsize=180,dialog_parent=self.tlb,info=info,default=default)
             if (~ finite(dm_to_number(qvalue))) then begin
                ok = dialog_message('Invalid inputs. No powder line is specified.',dialog_parent=self.tlb,/error)
                self.powd_typ = 0
                dm_toggle_menubut,check=ids[self.powd_typ],uncheck=ids[where(indgen(5) ne self.powd_typ)]
             endif else begin
                qvalue = dm_to_number(strsplit(qvalue,', '+string(9b),/extract))   
             endelse
             end 
          else:
    endcase
    if n_elements(qvalue) ne 0 and self.powd_typ gt 0 then self.powd_qvalue=ptr_new(qvalue)
end

;keyword:
;   init: if not set, free some data pointers
pro dcs_mslice::change_samptyp,new_typ,init=init
    old_typ  = self.samp_typ
    self.samp_typ = new_typ
    self->my_widget_control,'darkCorr',sensitive=((self.samp_typ ge 2) and (self.ftype eq 0))
    self->my_widget_control,['vaxisMenu','extraVaxis','foldRight'],sensitive=(self.samp_typ ne 0)
    self->my_widget_control,'bgtcBut',sensitive=(self.samp_typ ne 2)
    if old_typ eq new_typ then return
    sampTyp  = widget_info(self.fileBas,find_by_uname='sampTyp')
    powdBut  = widget_info(self.fileBas,find_by_uname='powdTyp')
    crystTyp = widget_info(self.fileBas,find_by_uname='crystTyp')
    sct1But  = widget_info(self.fileBas,find_by_uname='crystTyp1')
    sct2But  = widget_info(self.fileBas,find_by_uname='crystTyp2')
    sct3But  = widget_info(self.fileBas,find_by_uname='crystTyp3')
    case self.samp_typ of
       0: dm_toggle_menubut,check=powdBut,uncheck=[sct1But,sct2But,sct3But]
       1: dm_toggle_menubut,check=sct1But,uncheck=[powdBut,sct2But,sct3But]
       2: dm_toggle_menubut,check=sct2But,uncheck=[powdBut,sct1But,sct3But]
       3: dm_toggle_menubut,check=sct3But,uncheck=[powdBut,sct1But,sct2But]
       else:
    endcase
    sptypes = ['Powder','Single Crystal','Diffuse Scattering Single E','Diffuse Scattering Multiple E']
    widget_control,sampTyp,set_value=sptypes[self.samp_typ]
    if (float(!version.release) ge 5.6) then begin
       widget_control,crystTyp,set_button=(self.samp_typ ge 1)  ;doesn't work as of IDL6.1
    endif
    if keyword_set(init) then return
    if (self.samp_typ ge 2) or (old_typ ge 2) then begin    ;if changed from/to diffuse scattering type
       ;clear data
       ptr_free,self.kfactor,self.dataStrPtr,self.powd_line
       self->my_widget_control,['addBut','clearDat'],sensitive=0          ;disallow adding data and clear data
       self->my_widget_control,'titlLab',set_value='' & self.titl_lab=''  ;clear title label
       self->my_widget_control,'legdLab',set_value='' & self.legd_lab=''  ;clear legend label
       ;clear background
       ptr_free,self.bgdata,self.bgdata2,self.bgratePtr,self.bgtchnPtr,self.bgdetsPtr
       self->my_widget_control,'bgclBut',sensitive=0
       if self.extraempcan_yn then self->my_widget_control,'emptBut2',sensitive=0
    endif
    ptr_free,self.projStrPtr
    self->reset_file
    self->reset_pdt
end

;clear mask pointer and set the menu items accordingly
pro dcs_mslice::clear_mask
    if ptr_valid(self.mask) then begin
       ptr_free,self.mask
       if self.ftype eq 0 then begin
          self.dcsdetmask[*] = 0
          self->dm_mask_detbanks
       endif else $
          self->my_widget_control,['mksvBut','mkclBut'],sensitive=0
    endif
    if ptr_valid(self.projStrPtr) then begin
       ;need to recalculate projections
       ptr_free,self.projStrPtr
       self->set_display_button
    endif
end

; adds an object to the list and shows its title to the view menu or deletes an object from the list
; and removes the corresponding item in the view menu
; parameter:
;   opcode:   0: add plotwindow , this is the default operation
;             1: make plotwindow current, object keyword must be given
;             2: keep the plotwindow, object keyword must be given
;             3: delete plotwindow, object keyword must be given
;             4: close all plot windows
; keyword:
;   type:     0:slice 1:cut 2:trajectory 3:volume
;   title:    string of the button name to be added to the view menu
;   object:   opcode=1,2--object chosen opcode=3--object to be deleted
pro dcs_mslice::dm_add_plotwindow,opcode,type=type,title=title,object=object
    if n_elements(opcode) eq 0 then opcode=0   ;default add plot window
    viewMenu = widget_info(self.tlb,find_by_uname='viewMenu')
    if widget_info(viewMenu,/valid_id) eq 0 then return
    case opcode of
       0:begin
         if (~ obj_valid(self.current)) then return
         ;create the unique id for the view menu item
         id  = self->dm_get_id(self.current)
         ;check if the current object is already in the list, if not add it to the list
         all = self.gr_list->get(/all,count=count)
         in  = 0b
         for i=0,count-1 do begin
             if all[i] eq self.current then begin
                in = 1b
                break
             endif
         endfor
         if in eq 0 then begin    ;not in the list
            self.gr_list->add,self.current            ;add the object to the list
            id = dm_widget_button(viewMenu,value='',uname=id,separator=(count eq 0)) ;add the item to the view menu
         endif else begin     ;in the list, find the menu item id
            id = widget_info(viewMenu,find_by_uname=id)
         endelse
         ;generate proper string for the menu item
         if n_elements(type) eq 0 then type=0
         case type of
            0:   value = 'Slice: '
            1:   value = 'Cut: '
            2:   value = 'Trajectory: '
            3:   value = 'Volume: '
            else:value = 'Slice: '
         endcase
         if n_elements(title) ne 0 then value=value+strtrim(title,2)
         if widget_info(id,/valid_id) eq 1 then begin
            if float(!version.release) ge 5.6 then $
               widget_control,id,set_value=value,set_button=1 $
            else $
               widget_control,id,set_value=value
         endif
         self->my_widget_control,['killAll','arrPlot'],viewMenu,/sensitive
         end
       1:begin 
         if obj_valid(self.current) and (float(!version.release) ge 5.6) then begin   ;remove the current check mark
            id = self->dm_get_id(self.current) ;locate the corresponding view menu item id
            self->my_widget_control,id,viewMenu,set_button=0
         endif
         all = self.gr_list->get(/all,count=count)
         for i=0,count-1 do begin
            if  all[i] ne object then all[i]->setproperty,keep=1 $
            else self.current=all[i]
         endfor
         if float(!version.release) ge 5.6 then begin ;add check mark to the current plot window
            id = self->dm_get_id(self.current)
            self->my_widget_control,id,viewMenu,set_button=1
         endif
         end
       2:begin          
         self.current=obj_new()
         if float(!version.release) ge 5.6 then begin ;remove check mark from the view menu
            id = self->dm_get_id(object)  ;locate the corresponding view menu item id
            self->my_widget_control,id,viewMenu,set_button=0
         endif
         end
       3:begin       
         all = self.gr_list->get(/all,count=count)
         if count eq 0 then return
         for i=0,count-1 do begin   ;destroy all the plot menu items
             id = self->dm_get_id(all[i])
             self->my_widget_control,id,viewMenu,get_value=value,/destroy  ;remove the item in the view menu
             if n_elements(info) eq 0 then info = [id,value] $  ;save the names
             else info = [[info],[[id,value]]]
         endfor      
         self.gr_list->remove,object       ;remove the object from the list
         all = self.gr_list->get(/all,count=count)
         for i=0,count-1 do begin   ;reconstruct the plot menu items
             id  = self->dm_get_id(all[i])
             all[i]->getproperty,keep=keep,title=title
             value = title ;change the menu name to the title of the plot
             id1 = where(info[0,*] eq id,existed)
             if existed ne 0 then begin
                tmp = info[1,id1[0]]
                tmp = strsplit(tmp,':',/extract,/preserve_null)
                if n_elements(tmp) ne 0 then value=tmp[0]+': '+title ;include the type string
             endif
             id = dm_widget_button(viewMenu,uname=id,value=value,separator=(i eq 0),set_button=(1-keep))
         endfor
         self->my_widget_control,['killAll','arrPlot'],viewMenu,sensitive=(count gt 0)
         end
       4:begin
         all = self.gr_list->get(/all,count=count)
         self.gr_list->remove,/all
         if count gt 0 then begin
            obj_destroy,all
             ;destroy all the plot menu items
            for i=0,count-1 do begin  
                id = self->dm_get_id(all[i])
                self->my_widget_control,id,viewMenu,get_value=value,/destroy  ;remove the item in the view menu
            endfor
            self->my_widget_control,['killAll','arrPlot'],viewMenu,sensitive=0
         endif
         end
       else:
    endcase
end

;arrange the plot windows
pro dcs_mslice::dm_arrange_plot,cascade=cascade
    oplots = self.gr_list->get(/all,count=count)
    if count eq 0 then return
    for i=0,count-1 do oplots[i]->totop
    if count eq 1 then return
    device,get_screen_size=scr_size
    if keyword_set(cascade) then begin
       oplots[count-1]->getproperty,xsize=xsize,ysize=ysize & ysize=ysize+20
       xstep = (40<((scr_size[0]-xsize)/count))
       ystep = (40<((scr_size[1]-ysize)/count))
       xoff  = 5>((scr_size[0]-xstep*count-xsize)/2)
       yoff  = 5>((scr_size[1]-ystep*count-ysize)/2)
       for i=0,count-1 do begin
           oplots[i]->getproperty,tlb=tlb
           widget_control,tlb,xoffset=((xoff+xstep*i)<(scr_size[0]-10)),$
                yoffset=((yoff+ystep*i)<(scr_size[1]-10))
       endfor
    endif else begin
       if count le 4 then ncol=2 $
       else if count le 9 then ncol=3 $
       else ncol=4
       xsize = intarr(count) & ysize = intarr(count) & wid = lonarr(count)
       for i=0,count-1 do begin
           oplots[i]->getproperty,tlb=tlb,xsize=tmpx,ysize=tmpy
           wid[i] = tlb & xsize[i] = tmpx+10 & ysize[i] = tmpy+75
       endfor
       nrow = (count+ncol-1)/ncol
       roww = intarr(nrow) & rowh = intarr(nrow) & xoff = intarr(nrow)
       for i=0,nrow-1 do begin
           roww[i] = total(xsize[ncol*i:((ncol*(i+1)-1)<(count-1))])
           rowh[i] = max(ysize[ncol*i:((ncol*(i+1)-1)<(count-1))])
           xoff[i] = 0>((scr_size[0]-roww[i])/2)
       endfor
       yoff = 0>((scr_size[1]-total(rowh))/2)
       for i=0,count-1 do begin
           row = i/ncol
           col = (i mod ncol)
           if xoff[row] eq 0 then xoffset=scr_size[0]/ncol*col $
           else begin
              xoffset = xoff[row]
              if col ne 0 then $
                 xoffset = (xoffset+total(xsize[row*ncol:row*ncol+col-1]))<(scr_size[0]-20)
           endelse
           if yoff eq 0 then yoffset=scr_size[1]/nrow*row $
           else begin
              yoffset = yoff
              if row ne 0 then $
                 yoffset = (yoffset+total(rowh[0:row-1]))<(scr_size[1]-20)
           endelse
           widget_control,wid[i],xoffset=xoffset,yoffset=yoffset
       endfor
    endelse
end

;create a unique id for an object
function dcs_mslice::dm_get_id,object
    help,object,output=id & id=strsplit(id,' ',/extract) &  id=id[n_elements(id)-1]
    tmp = strsplit(id,'()',/extract) & if n_elements(tmp) eq 3 then id=tmp[0]+tmp[2]
    return,id
end

;object event handler
pro dcs_mslice::event,event
    compile_opt IDL2    ;,strictarrsubs
                        ;idl2=defint32,strictarr
                        ;strictarr:   [] to index array
                        ;strictarrsubs: error when out-of-range indices,IDL5.6 or after
    ;catch and ignore all errors in this program
    catch, myerror
    if myerror ne 0 then begin
       catch,/cancel
       errmesg = !error_state.msg 
       if obj_valid(self) then begin
          if (self.samp_typ eq 3) and (!error_state.name eq 'IDL_M_CNTGETMEM') then $
             errmesg = [errmesg,'','Please reduce the energy range, or use single E option.']
          if (lmgr(/runtime) eq 0) and (!version.os_family eq 'Windows') then begin
             if strpos(errmesg[0],'rebinnd') ne -1 then begin
                errmesg = 'Please copy msvcr100'+(['','_64'])[!VERSION.MEMORY_BITS eq 64]+'.dll from \dave\programs\auxiliary\rebin_nd\ to your Windows system directory.'
                errmesg = [errmesg,'','Or you can disable the external binning library in the menu Option->Binning Method->Use External Binning Library->No.']
             endif
          endif  
       endif
       ok = dialog_message(errmesg,/error)
       if obj_valid(self) then self.error=1
       return
    end
    widget_type = widget_info(event.id,/type)
    case widget_type of
       1:tmp_value = event.select                               ;button
       3:begin                                                  ;text box
         widget_control,event.id,get_value=tmp_value
         tmp_value = strtrim(tmp_value,2)
         end
       8:tmp_value = widget_info(event.id,/droplist_select)     ;drop list
       12:begin                                                 ;combobox
         tmp_value = event.index
         widget_control,event.id,get_value=tmp_list
         if tmp_list[tmp_value] ne event.str then begin         ;to avoid events from set_value in unix operating system
            return
         endif
         end
       else:
    endcase
    if n_elements(tmp_value) eq 1 then tmp_value=tmp_value[0]

    resetu1 = 0b & resetu2 = 0b & resetu3 = 0b & recalc_proj = 0b

    uname = widget_info(event.id,/uname)
    type  = tag_names(event,/structure)
    case strlowcase(type) of
        'dm_plot_make_current': uname = type
        'dm_plot_keep':         uname = type
        'dm_plot_kill':         uname = type
        'widget_context':       uname = 'status'
        'widget_kbrd_focus':    if widget_info(event.id,/name) eq 'TEXT' then uname = 'text_focus'
        else:
    endcase
    tmp = strmid(strlowcase(uname),0,8)
    ind_sel = dm_to_number(strmid(strlowcase(uname),0,1,/reverse),/int)
    case tmp of
        'cutsel_t': uname = 'cutsel_t'
        'cutselfr': uname = 'cutsel_from'
        'cutselto': uname = 'cutsel_to'
        'dispsel_': uname = 'dispsel_t'
        'dispself': uname = 'dispsel_from'
        'dispselt': uname = 'dispsel_to'
        'volsel_t': uname = 'volsel_t'
        'volselfr': uname = 'volsel_from'
        'volselto': uname = 'volsel_to'
        else:
    endcase
    if strmid(strlowcase(uname),0,4) eq 'user' then begin
       userf_info = 'Available variables:'
       if !version.os_family eq 'Windows' then begin
          sep  = string(9b)+': ' 
          tail = '             '
       endif else begin 
          sep  = ': '
          tail = ''
       endelse
       if self.samp_typ eq 0 then begin ;powder
          userf_info = userf_info+'qty, err, Q, en, det_tt.          '
          if ptr_valid(self.projStrPtr) then begin
             if n_elements((*self.projStrPtr).qty) ne 1 then begin
             help,(*self.projStrPtr).qty,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'qty'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             help,(*self.projStrPtr).err,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'err'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             endif
             help,(*self.projStrPtr).Q,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ')
             userf_info = [userf_info,'Q'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             help,(*self.projStrPtr).en,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ')
             userf_info = [userf_info,'en'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             help,(*self.projStrPtr).two_theta,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'det_tt'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
          endif 
       endif else begin                 ;single crystal
          userf_info = userf_info+'qty, err, Q, Qv, en, det_tt.          '
          if ptr_valid(self.projStrPtr) then begin
             if n_elements((*self.projStrPtr).qty) ne 1 then begin
             help,(*self.projStrPtr).qty,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'qty'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             help,(*self.projStrPtr).err,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'err'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             endif
             help,(*self.projStrPtr).V,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') & pos1 = strpos(strlowcase(tmp),'array[')
             userf_info = [userf_info,'Q'+sep+strmid(tmp,pos+1,pos1+5-pos)+'3'+strmid(tmp,pos1+7,strlen(tmp)-7-pos1)+tail]
             userf_info = [userf_info,'Qv'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             help,(*self.projStrPtr).en,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'en'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
             help,(*self.projStrPtr).two_theta,output=tmp & tmp = strtrim(tmp[0],2) & pos = strpos(tmp,' ') 
             userf_info = [userf_info,'det_tt'+sep+strmid(tmp,pos+1,strlen(tmp)-1-pos)+tail]
          endif
       endelse
    endif
    if uname ne 'text_focus' then WIDGET_CONTROL,/HOURGLASS
    case strlowcase(uname) of
        'dm_plot_make_current': self->dm_add_plotwindow,1,object=event.object   ;this event is sent from dm_plot window
        'dm_plot_keep':         self->dm_add_plotwindow,2,object=event.object   ;this event is sent from dm_plot window
        'dm_plot_kill':         self->dm_add_plotwindow,3,object=event.object   ;this event is sent from dm_plot window
        'text_focus': begin
            if event.enter then begin
               if !version.os_family eq 'Windows' then begin ;only for Windows OP
                  widget_control,event.id,get_value=tmp
                  widget_control,event.id,set_text_select=[0,strlen(tmp[0])] 
               endif
            endif else $
               widget_control,event.id,set_text_select=[0,0]
          end
        'tlb':begin
            if float(!version.release) ge 5.5 then $
               widget_control,self.tlb,xsize=self.geom[0],ysize=self.geom[1] $
            else $
               widget_control,self.tlb,scr_xsize=self.geom[0],scr_ysize=self.geom[1]
          end
        'savevolumetxt': self->save_volume,/txt
        'saveslicedave': self->save_slice,/dave
        'saveslicespe':  self->save_slice,/spe
        'savesliceiexy': self->save_slice,/iexy
        'saveslicexyie': self->save_slice,/xyie
        'sendslicedave': begin
            self->save_slice,dPtr=dPtr
            (self.DAVETool)->AddDavePtrToDataManager, dPtr, 'slice from mslice'  
            heap_free,dPtr  
          end              
        'savecutxye':    self->save_cut,/xye
        'savecutdave':   self->save_cut,/dave
        'sendcutdave': begin
            self->save_cut,dPtr=dPtr    
            (self.DAVETool)->AddDavePtrToDataManager, dPtr, 'cut from mslice'
            heap_free,dPtr
          end              
        'saveprojtxt':   self->save_projection,/txt
        'hist_0':        self->set_filehistory,choose=0
        'hist_1':        self->set_filehistory,choose=1
        'hist_2':        self->set_filehistory,choose=2
        'hist_3':        self->set_filehistory,choose=3
        'hist_4':        self->set_filehistory,choose=4
        'hist_5':        self->set_filehistory,choose=5
        'hist_6':        self->set_filehistory,choose=6
        'hist_7':        self->set_filehistory,choose=7
        'hist_8':        self->set_filehistory,choose=8
        'hist_9':        self->set_filehistory,choose=9
        'clearhist': begin
            ptr_free,self.file_hist
            self.n_history = 0
            self->set_filehistory
           end
        'exitbut':begin
            widget_control,self.tlb,/destroy
            return
          end
        'ldpmbut':begin  ;load parameter file
            dcsdetmask = self.dcsdetmask
            path = self.dirs[1]
            parmfile = dm_choose_file('prm',dialog_parent=self.tlb,/read,path=path)
            if strlen(parmfile) ne 0 then self->dm_load_parameter,parmfile
            if file_test(path,/directory) then self.dirs[1]=path
            if (self.ftype eq 0) and (total(abs(fix(dcsdetmask)-fix(self.dcsdetmask))) ne 0) then $
               ok = dialog_message('The detector masks have been changed.',/info,dialog_parent=self.tlb)
          end
        'svpmbut':begin ;save parameters into a file
            path = self.dirs[1]
            parmfile = dm_choose_file('prm',dialog_parent=self.tlb,/write,path=path)
            if strlen(parmfile) ne 0 then self->dm_save_parameter,parmfile
            if file_test(path,/directory) then self.dirs[1]=path
          end
        'knpmbut':begin  ;change smoothing kernel
            old = dm_to_string(float(self.disp_sm_kernel))
            old = '[['+old[0]+' ,'+old[1]+' ,'+old[2]+'],['+old[3]+' ,'+old[4]+' ,'+old[5]+'],['+old[6]+' ,'+old[7]+' ,'+old[8]+']]'
            new = dm_dialog_input('kernel matrix:',title='Change smoothing kernel',default=old,xsize=250,dialog_parent=self.tlb)
            new = dm_to_number(strsplit(new,'[], '+string(9b),/extract),/float)
            if n_elements(new) ne 9 then begin
               ok = dialog_message('Please input a 3x3 matrix.',/error,dialog_parent=self.tlb)
               break
            endif
            if max(abs(new[indgen(3,3)]-new[transpose(indgen(3,3))])) ne 0.0 then begin
               ok = dialog_message('Please input a symmetric matrix.',/error,dialog_parent=self.tlb)
               break
            endif
            if total(new[0:8]) ne 0 then self.disp_sm_kernel=new[indgen(3,3)]
          end
        'bnthbut':begin  ;change binning threshold
            notok = 1b
            old   = self.binthreshold
            info  = ['Bins with fewer number of data points than (filling factor)*(NMAX)',$
                     'will be discarded, where NMAX is the maximum number of points in a bin.']
            if self.samp_typ ne -1 then info=[info,'This factor will be reset to 0 every time a file is loaded.']
            while notok do begin
              new = dm_dialog_input('filling factor:',title='Change binning threshold',dialog_parent=self.tlb,default=old,$
                   info=info,/float)
              if finite(new) then begin
                 if (new ge 0) and (new lt 1) then $
                     notok = 0b $
                 else begin
                     old  = new
                     info = '0.0<= filling factor <1.0'
                 endelse
              endif
            endwhile
            self.binthreshold = new           
          end
        'ssfcbut':begin  ;change self-shielding factor
            old = self.ssfac
            new = dm_dialog_input('self-shielding factor:',title='Change self-shielding factor',default=old,$
               /float,info='I=I_samp-ssfac*I_can',dialog_parent=self.tlb)
            if finite(new) then $
               if new ne self.ssfac then begin
                  self.ssfac = new
                  if ptr_valid(self.projStrPtr) and ptr_valid(self.bgdata) then recalc_proj = 1b    ;need to recalculate projections
               endif
          end
        'emfcbut':begin  ;change 2-set empty can factors
            notok = 1b
            while(notok) do begin
                tmp = dm_dialog_input(['empty can 1 factor','empty can 2 factor'],/float,default=self.empcanfactor,$
                      info=['I=I_samp-ssfac*(factor1*I_can1+factor2*I_can2)','ssfac: self-shielding factor'],dialog_parent=self.tlb)
                if total(finite(tmp)) eq 2 then begin
                   self.empcanfactor = tmp
                   notok = 0b
                endif
            endwhile
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end  
        'ddirbut':begin    ;set data directory
            old = self.dirs[0]
            new = dialog_pickfile(/directory,path=old,dialog_parent=self.tlb)
            if (strlen(new) ne 0) then begin
               self.fileSel->set_path,new
               self.dirs[0] = new
               defsysv,'!dave_defaults',exists=exists
               if exists then (*!dave_defaults).datDir = new
            endif
          end
        'wdirbut':begin  ;set working directory
            old = self.dirs[1]
            new = dialog_pickfile(/directory,path=old,dialog_parent=self.tlb)
            if (strlen(new) ne 0) then begin
               self.dirs[1] = new
               defsysv,'!dave_defaults',exists=exists
               if exists then (*!dave_defaults).workDir = new
            endif
          end
        'bgovbut':begin  ;overall background rate
            value = !values.f_nan & nvalue = 1
            if ptr_valid(self.bgratePtr) then begin
               value  = mean((*self.bgratePtr)[*,0])
               nvalue = n_elements(*self.bgratePtr)   ;dark count file existed
            endif
            tmp_value = dm_dialog_input(dialog_parent=self.tlb,'background rate:',info='in (cts/hr/det)',/float,$
               default=value)
            if finite(tmp_value) then begin
               if tmp_value ge 0 then begin
                  if nvalue gt 1 then begin
                     ok = dialog_message('Do you want to replace the existing dark count file data?',/question,/default_no,dialog_parent=self.tlb)
                     if strupcase(ok) eq 'NO' then break
                  endif
                  ptr_free,self.bgratePtr
                  self.bgratePtr = ptr_new(tmp_value)
                  self->my_widget_control,'bgclBut',/sensitive
                  if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
               endif else $
                  ok = dialog_message('Negative number is entered. The overall background rate has not been changed.',dialog_parent=self.tlb)
            endif else $
               ok = dialog_message('Invalid input. The overall background rate has not been changed.',dialog_parent=self.tlb)
          end
        'bgclbut':begin
            ptr_free,self.bgdata,self.bgdata2,self.bgratePtr,self.bgtchnPtr,self.bgdetsPtr
            widget_control,event.id,sensitive=0
            if self.extraempcan_yn then self->my_widget_control,['emfcBut','emptBut2'],sensitive=0
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end
        'vdclbut':begin
            widget_control,event.id,sensitive=0
            ptr_free,self.eff
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end
        'maskfiller':begin   
            self.mask_filler = dm_dialog_input('mask value:',/float,default=self.mask_filler,$
                       info=['The data with intensity less than or equal to this value will be masked.',$
                       'Clear the value to disable it.'],dialog_parent=self.tlb)
            if finite(self.mask_filler) then begin
               if ptr_valid(self.dataStrPtr) then $
                  index = where((*self.dataStrPtr).data[*,0] gt self.mask_filler,count) $
               else $
                  index = where((*self.dataPtr).data[*,0] gt self.mask_filler,count)
               if count ne 0 then begin
                  if (~ ptr_valid(self.dataStrPtr)) then self.dataStrPtr = ptr_new((*self.dataPtr))
                  (*self.dataPtr) = {data:(*self.dataStrPtr).data[index,*],label:(*self.dataStrPtr).label,$
                                     unit:(*self.dataStrPtr).unit,instrument:(*self.dataStrPtr).instrument}
               endif else begin
                  ok = dialog_message('All data have been masked by the given value. Mask is disabled.',dialog_parent=self.tlb)
                  self.mask_filler = !values.f_nan
               endelse
            endif 
            if (~ finite(self.mask_filler)) and ptr_valid(self.dataStrPtr) then begin
               (*self.dataPtr) = (*self.dataStrPtr)
               ptr_free,self.dataStrPtr
            endif
            widget_control,event.id,set_button=finite(self.mask_filler)   
          end
        'mkngbut': self->dm_mask_detbanks, 'negative'   ;mask negative angle detectors
        'mkulbut': self->dm_mask_detbanks, 'upperlower' ;mask upper and lower banks
        'mkcnbut': self->dm_mask_detbanks, 'center'     ;mask central bank
        'mklobut': self->dm_mask_detbanks, 'lower'      ;mask lower bank
        'mkupbut': self->dm_mask_detbanks, 'upper'      ;mask upper bank
        'mkstbut':begin ;set mask
            default=''
            ;clean the mask, if the number of masked detectors is less than 20, save it to the default
            if ptr_valid(self.mask) then begin
               exist  = 1
               n_mask = n_elements(*self.mask)
               if n_mask le 20 then begin
                  default = strcompress(string(*self.mask,format='('+string(n_mask)+'(i5,","))'),/remove_all)
                  if strmid(default,0,/reverse_offset) eq ',' then $
                     default = strmid(default,0,strlen(default)-1)
               endif
            endif else exist=0
            self->clear_mask    ;clear the mask
            tmp = dm_dialog_input(dialog_parent=self.tlb,'Det num:',title='Masking detectors',default=default,$
               info=['Detector number starts from 0 to 912 for DCS.','eg:370,373 or 370:374c'],xsize=200)
            if strlen(tmp) eq 0 then begin
               if exist then ok=dialog_message(['No detector is specified.',$
                  'Previously set detector mask has been cleared.'],dialog_parent=self.tlb)
               break
            endif else begin
               tmpstrs = strsplit(tmp,',',/extract,count=count1)
               for ii=0,count1-1 do begin
                   status = ''
                   index  = strpos(tmpstrs[ii],'c',/reverse_search)
                   if index[0] ne -1 then status='c'
                   index  = strpos(tmpstrs[ii],'C',/reverse_search)
                   if index[0] ne -1 then status='c'
                   index  = strpos(tmpstrs[ii],'u',/reverse_search)
                   if index[0] ne -1 then status='u'
                   index  = strpos(tmpstrs[ii],'U',/reverse_search)
                   if index[0] ne -1 then status='u'
                   index  = strpos(tmpstrs[ii],'l',/reverse_search)
                   if index[0] ne -1 then status='l'
                   index  = strpos(tmpstrs[ii],'L',/reverse_search)
                   if index[0] ne -1 then status='l'

                   ind0  = strsplit(tmpstrs[ii],'[a-zA-Z '+string(9b)+',:-]',/regex)
                   ind1  = strsplit(tmpstrs[ii],':-')
                   index = -1L
                   for i=1,n_elements(ind1)-1 do begin
                       ind2 = where(ind0 eq ind1[i],count)
                       if count ne 0 then index=[index,ind2]
                   endfor
                   n_index = n_elements(index)
                   if n_index ne 1 then index=index[1:n_index-1]-1
                   tmp = strsplit(tmpstrs[ii],'[a-zA-Z '+string(9b)+',:-]',/regex,/extract)
                   tmpmask = -1L
                   for i=0,n_elements(tmp)-1 do begin
                       tmp1 = dm_to_number(tmp[i],/float)
                       if finite(tmp1) then begin
                          ind0 = where(index eq i,count)
                          if count ne 0 then begin
                             i = i+1
                             tmp2 = dm_to_number(tmp[i],/float)
                             if finite(tmp2) then tmpmask=[tmpmask,tmp1+findgen(abs(tmp2-tmp1)+1)] $
                             else tmpmask = [tmpmask,tmp1]
                          endif else tmpmask = [tmpmask,tmp1]
                       endif
                   endfor
                   index = where(tmpmask ge 0,count)
                   if count ne 0 then begin
                      tmpmask = fix(tmpmask[index]+0.1,type=3)
                      case status of
                        'c': tmpmask=dm_common(tmpmask,dm_mask_detbanks(0),/no_copy)
                        'l': tmpmask=dm_common(tmpmask,dm_mask_detbanks(-1),/no_copy)
                        'u': tmpmask=dm_common(tmpmask,dm_mask_detbanks(1),/no_copy)
                        else:
                      endcase
                      if n_elements(mask) eq 0 then $
                         mask = tmpmask $
                      else $
                         mask = [mask,tmpmask]
                   endif
               endfor
            endelse
            if n_elements(mask) eq 0 then count=0 $
            else ind = where(finite(mask) eq 1,count)
            if count ne 0 then begin
               mask = mask[ind]
               ind  = where((mask lt 913) and (mask ge 0),count)
               if count ne 0 then begin
                  self.mask = ptr_new(mask[ind])
                  self->my_widget_control,['mksvBut','mkclBut'],/sensitive
               endif else begin
                  if exist then ok=dialog_message(['No valid detector is specified.',$
                     'Previously set detector mask has been cleared.'],dialog_parent=self.tlb) $
                  else ok = dialog_message('No valid detector mask is specified.',dialog_parent=self.tlb)
               endelse
               ind = -1
            endif
          end
        'mkpldnum':begin ;plot and set mask by detector number
            ;must have data
            if (~ ptr_valid(self.dataStrPtr)) then begin
               ok = dialog_message('Please load data files first.',dialog_parent=self.tlb)
               widget_control,event.id,sensitive=0
               break
            endif
            case self.samp_typ of
                 0: detsum = total((*self.dataStrPtr).qty,1)
                 1: detsum = total((*self.dataStrPtr).qty,1)
                 2: detsum = (*self.dataStrPtr).qty[*,0]
                 3: detsum = total((*self.dataStrPtr).qty[*,*,0],1)
            endcase
            ndet  = n_elements(detsum)
            blist = intarr(ndet)
            if ptr_valid(self.mask) then begin
               ind = where(*self.mask lt ndet,count)
               if count gt 0 then $
                  blist[0:count-1] = (*self.mask)[ind]+1
            endif
            self->plot_mask,detsum,blist
          end
        'mkplangl':begin ;plot and set mask by angle, only available for DCS
            ;must have data
            if (~ ptr_valid(self.dataStrPtr)) then begin
               ok = dialog_message('Please load data files first.',dialog_parent=self.tlb)
               widget_control,event.id,sensitive=0
               break
            endif
            case self.samp_typ of
                 0: detsum = total((*self.dataStrPtr).qty,1)
                 1: detsum = total((*self.dataStrPtr).qty,1)
                 2: detsum = (*self.dataStrPtr).qty[*,0]
                 3: detsum = total((*self.dataStrPtr).qty[*,*,0],1)
            endcase
            ndet  = n_elements(detsum)
            blist = intarr(ndet)
            if ptr_valid(self.mask) then begin
               ind = where(*self.mask lt ndet,count)
               if count gt 0 then $
                  blist[0:count-1] = (*self.mask)[ind]+1
            endif
            self->plot_mask,detsum,blist,/angle
          end
        'mkbpbut':begin ;automatically mask bragg peaks
            if (~ ptr_valid(self.dataStrPtr)) then begin
               ok = dialog_message('Please load data files first.',dialog_parent=self.tlb)
               widget_control,event.id,sensitive=0
               break
            endif
            if ptr_valid(self.mask) then begin
               ok = dialog_message('This operation will remove the existing masks. Do you want to continue?',$
                    /question,dialog_parent=self.tlb)
               if strupcase(ok) eq 'NO' then break
               self->clear_mask     ;clear the mask
            endif
            case self.samp_typ of
                 0: detsum = total((*self.dataStrPtr).qty,1)
                 1: detsum = total((*self.dataStrPtr).qty,1)
                 2: detsum = (*self.dataStrPtr).qty[*,0]
                 3: detsum = total((*self.dataStrPtr).qty[*,*,0],1)
            endcase
            ndet  = n_elements(detsum)

            ;smooth data
            x = indgen(ndet) & y = smooth(detsum,4) & yerr = sqrt(abs(y))

            ;search for peaks
            ;method A: 1 find the highest peak, fit it with a Gaussian, width w0
            ;          2 replace the peak with background value
            ;          3 repeat 1 & 2, if w<0.4w0 or w>2.5w0, not a valid peak
            ;          4 false peak if peak intensity is less than 2.5% of the strongest peak
            ;          5 stop when two consecutive false peaks
            false_peak = 0
            fit_fwhm  = !values.f_nan
            fit_max   = !values.f_nan
            while (false_peak lt 2) do begin   ;search for peaks
                ;find the maximum
                if finite(fit_fwhm) then begin
                   ymax = max(y,ind_ymax,min=ymin)
                   fwhm = fit_fwhm
                endif else begin
                   i0 = 20<(ndet/20)
                   ymax = max(y[i0:ndet-1],ind_ymax,min=ymin)
                   fwhm = 10<(ndet/20)
                   ind_ymax = ind_ymax+i0
                endelse
                yave = mean(y)
                ;find peak boundaries
                ip_0 = where(y[ind_ymax:ndet-1] le (ymax-1./3*(ymax-yave)),count)
                if count ne 0 then begin
                   ip_1 = (ind_ymax+ip_0[0]*3)<(ndet-1) ;peak boundary
                   ip_0 = 0>(ind_ymax-ip_0[0]*3)
                endif else begin
                   ip_0 = where(y[0:ind_ymax] ge (ymax-1./3*(ymax-yave)),count)
                   if count eq 0 then begin
                      ip_0 = 0
                      ip_1 = ndet-1
                   endif else begin
                      ip_1 = (ind_ymax+(ind_ymax-ip_0[count-1])*3)<(ndet-1)
                      ip_0 = 0>(ind_ymax-(ind_ymax-ip_0[count-1])*3)
                   endelse
                endelse
                if (ip_1-ip_0 eq ndet-1) or (ip_1-ip_0 lt 4) then break
                ;fit the peak with a Gaussian
                estimates = [ymax,x[ind_ymax],fwhm/sqrt(8.*alog(2)),y[ip_1]]
                result = gaussfit(x[ip_0:ip_1],y[ip_0:ip_1],params,estimates=estimates,$
                         measure_errors=yerr[ip_0:ip_1],nterms=n_elements(estimates))
                fwhm   = abs(params[2])*sqrt(8.*alog(2))
                if (~ finite(fit_max))  then fit_max = params[0]
                if (~ finite(fit_fwhm)) then fit_fwhm = fwhm
                ip_0   = (0>(fix(params[1]-fwhm))<(ndet-1))
                ip_1   = (0>(fix(params[1]+fwhm))<(ndet-1))
                if (ip_1-ip_0 eq ndet-1) or (ip_1-ip_0 lt 4) then break
                if (fwhm gt 0.4*fit_fwhm) and (fwhm lt 2.5*fit_fwhm) then begin
                   if params[0] gt 0.025*fit_max then begin
                      false_peak = 0   ;clear the false peak flag
                      if n_elements(mask) eq 0 then $
                         mask = ip_0+indgen(ip_1-ip_0+1)+1 $
                      else $
                         mask = [mask,ip_0+indgen(ip_1-ip_0+1)+1]
                   endif else $
                      false_peak = false_peak+1
                endif else $
                   false_peak = false_peak+1
                ;replace the peak with the background
                y[ip_0:ip_1] = params[3]
            endwhile
            x = 0 & y = 0 & yerr = 0
            if n_elements(mask) eq 0 then $
               mask = [0] $
            else $
               mask = mask[uniq(mask,sort(mask))]
            if n_elements(mask) eq ndet then mask = [0]
            self->plot_mask,detsum,mask
          end
        'mkrdbut':begin ;read mask file
            path = self.dirs[1]
            file = dm_choose_file('',/read,dialog_parent=self.tlb,title='Read mask file:',path=path)
            if strlen(file) eq 0 then return
            if file_test(path,/directory) then self.dirs[1] = path
            self->clear_mask    ;clear the mask
            openr,unit,file,/get_lun
            tmp = ''
            while(~ eof(unit)) do begin
                readf,unit,tmp
                tmp = strtrim(tmp,2)
                if strlen(tmp) ne 0 then begin
                   tmp1 = strsplit(tmp,' ,'+string(9b),/extract,count=count1)
                   if finite(dm_to_number(tmp1[0])) then begin  ;make sure it's not a comment line
                      for i=0L,count1-1L do begin
                          tmp2 = strsplit(tmp1[i],'-:',/extract,count=count2)
                          tmp2 = dm_to_number(tmp2,/long)
                          if count2 eq 2 then $
                             tmp2 = min(tmp2)+lindgen(abs(tmp2[1]-tmp2[0])+1)
                          if n_elements(mask) eq 0 then mask=tmp2 else mask=[mask,tmp2]
                      endfor
                   endif
                endif
            endwhile
            free_lun,unit
            if n_elements(mask) ne 0 then begin
               ok = dm_dialog_input('det number starts from',default=0,droplist_content=ptr_new(['0','1']),$
                    is_droplist=[1],/return_number,dialog_parent=self.tlb)
               self.mask = ptr_new(mask-ok)
               self->my_widget_control,['mksvBut','mkclBut'],/sensitive
            endif
          end
        'mksvbut':begin ;save mask in a file
            if ptr_valid(self.mask) then begin
               path = self.dirs[1]
               file = dm_choose_file('msk',/write,dialog_parent=self.tlb,title='Save mask in an ASCII file.',path=path)
               if strlen(file) eq 0 then break
               if file_test(path,/directory) then self.dirs[1] = path
               openw,unit,file,/get_lun,error=openerr
               if openerr ne 0 then begin
                  ok = dialog_message("Can't write in "+file[0],/error,dialog_parent=self.tlb)
               endif else begin
                  widget_control,/hourglass  ;busy signal
                  printf,unit,*(self.mask)
                  free_lun,unit
               endelse
            endif
          end
        'mkclbut': self->clear_mask         ;clear the mask
        'directgeom':if self.instrgeom ne 0 then $
            self->set_instrgeom,geom=0
        'inversegeom':if self.instrgeom ne 1 then $
            self->set_instrgeom,geom=1
        'intntyp0':if self.intn_typ ne 0 then self->set_intntype,0
        'intntyp1':if self.intn_typ ne 1 then self->set_intntype,1
        'intntyp2':if self.intn_typ ne 2 then self->set_intntype,2
        'intntyp3':if self.intn_typ ne 3 then self->set_intntype,3
        'intntyp4':if self.intn_typ ne 4 then self->set_intntype,4
        'intntyp5':if self.intn_typ ne 5 then self->set_intntype,5
        'intntyp6':if self.intn_typ ne 6 then self->set_intntype,6
        'intneint': begin
            self.eint_yn = ~self.eint_yn
            dm_set_button,event.id,self.eint_yn,onstring='E-Integrated Intensity Enabled',$
                offstring='E-Integrated Intensity Disabled'
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end
        'binavg':if self.bin_avgsum ne 0 then begin
            self.bin_avgsum = 0
            binSum = widget_info(self.tlb,find_by_uname='binSum')
            dm_toggle_menubut,check=event.id,uncheck=binSum
          endif
        'binsum':if self.bin_avgsum ne 1 then begin
            self.bin_avgsum = 1
            binAvg = widget_info(self.tlb,find_by_uname='binAvg')
            dm_toggle_menubut,check=event.id,uncheck=binAvg
          endif
        'bincsyes': if self.bin_conststep ne 1 then begin
            self.bin_conststep = 1
            bincsNo  = widget_info(self.tlb,find_by_uname='bincsNo')
            dm_toggle_menubut,check=event.id,uncheck=bincsNo
          endif
        'bincsno': if self.bin_conststep ne 0 then begin
            self.bin_conststep = 0
            bincsYes  = widget_info(self.tlb,find_by_uname='bincsYes')
            dm_toggle_menubut,check=event.id,uncheck=bincsYes
          endif
        'binexyes': if self.bin_extlib ne 1 then begin
            self.bin_extlib = 1
            binexNo  = widget_info(self.tlb,find_by_uname='binexNo')
            dm_toggle_menubut,check=event.id,uncheck=binexNo
          endif
        'binexno': if self.bin_extlib ne 0 then begin
            self.bin_extlib = 0
            binexYes  = widget_info(self.tlb,find_by_uname='binexYes')
            dm_toggle_menubut,check=event.id,uncheck=binexYes
          endif
        'monsum':if self.mon_sumfit ne 0 then begin
            self.mon_sumfit = 0
            monFit = widget_info(self.tlb,find_by_uname='monFit')
            dm_toggle_menubut,check=event.id,uncheck=monFit
          endif
        'monfit':if self.mon_sumfit ne 1 then begin
            self.mon_sumfit= 1
            monSum = widget_info(self.tlb,find_by_uname='monSum')
            dm_toggle_menubut,check=event.id,uncheck=monSum
          endif
        'monfc0':if self.dcs_mon ne 0 then begin
            self.dcs_mon = 0
            monBM1  = widget_info(self.tlb,find_by_uname='monBM1')
            monTIME = widget_info(self.tlb,find_by_uname='monTIME')
            dm_toggle_menubut,check=event.id,uncheck=[monBM1,monTIME]
            self->my_widget_control,'bgovBut',/sensitive
          endif
        'monbm1':if self.dcs_mon ne 1 then begin
            self.dcs_mon = 1
            monFC0  = widget_info(self.tlb,find_by_uname='monFC0')
            monTIME = widget_info(self.tlb,find_by_uname='monTIME')
            dm_toggle_menubut,check=event.id,uncheck=[monFC0,monTIME]
            self->my_widget_control,'bgovBut',/sensitive
          endif
        'montime':if self.dcs_mon ne 2 then begin
            self.dcs_mon = 2
            monFC0  = widget_info(self.tlb,find_by_uname='monFC0')
            monBM1  = widget_info(self.tlb,find_by_uname='monBM1')
            dm_toggle_menubut,check=event.id,uncheck=[monFC0,monBM1]
          endif
        'abspnone':     self->absorption,-1
        'abspcylinder': self->absorption,0
        'abspannulus':  self->absorption,1
        'abspsphere':   self->absorption,2
        'abspflat':     self->absorption,3
        'vaxisarbi': if self.view_ortharbi ne 1 then begin
            self.view_ortharbi = 1
            vOrth = widget_info(self.tlb,find_by_uname='vaxisOrtho')
            dm_toggle_menubut,check=event.id,uncheck=vOrth
            self->my_widget_control,'viewAxis',set_value='  Viewing Axes:'
          endif
        'vaxisortho':if self.view_ortharbi ne 0 then begin
            self.view_ortharbi = 0
            vArbi = widget_info(self.tlb,find_by_uname='vaxisArbi')
            dm_toggle_menubut,check=event.id,uncheck=vArbi
            self->my_widget_control,'viewAxis',set_value='  Orthogonal Viewing Axes:'
          endif
        'powdernone': self->change_powderline,0
        'powderal':   self->change_powderline,1
        'powdercu':   self->change_powderline,2
        'powderst':   self->change_powderline,3
        'powdersp':   self->change_powderline,4
        'epadnone': self->eadjust,0
        'epadauto': self->eadjust,1
        'epadspec': self->eadjust,2
        'epadcheckfirst':begin
            self.eadjust_checkfirst = ~self.eadjust_checkfirst  
            dm_set_button,event.id,self.eadjust_checkfirst,onstring='Check the First File Only',$
              offstring='Check All Files'
          end
        'pbgblack': if self.pbgcol_bw ne 0 then begin
            self.pbgcol_bw = 0
            bgWhite  = widget_info(self.tlb,find_by_uname='pbgWhite')
            dm_toggle_menubut,check=event.id,uncheck=bgwhite
            self.colors[0] = 'white'
            if total(self.coldefine) eq 0 then self.coldefine = [255b,255b,255b]
            dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutColor'),select=self.cut_color,set_value=self.colors
          endif    
        'pbgwhite': if self.pbgcol_bw ne 1 then begin
            self.pbgcol_bw = 1
            bgBlack  = widget_info(self.tlb,find_by_uname='pbgBlack')
            dm_toggle_menubut,check=event.id,uncheck=bgBlack
            self.colors[0] = 'black'
            if total(255-self.coldefine) eq 0 then self.coldefine = [0b,0b,0b]
            dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutColor'),select=self.cut_color,set_value=self.colors
          endif    
        'userfunction': begin
             self.userfwarning = 0b ;clear the warning flag
             if lmgr(/runtime) then begin
                name  = 'user macro:' 
                title = 'Specify user macro'
                examp = ['Example: index = where(qty lt 0,count)',$
                         '              if count gt 0 then qty[index]=!values.f_nan'] 
             endif else begin
                name  = 'user macro/call:'
                title = 'Specify user macro or calls'
                examp = ['Example 1: index = where(qty lt 0,count)',$
                         '                if count gt 0 then qty[index]=!values.f_nan',$
                         'Example 2: qty=my_function(qty,err,en=en,Q=Q)',$
                         'Example 3: my_procedure,qty,err,en,det_tt=det_tt']
             endelse
             default = self.userf_str
             ;replace '&' with string('A4'xb)
             while((index=strpos(default,'&')) ge 0) do strput,default,string('A4'xb),index
             str = dm_dialog_input(name,title=title,default=default,dialog_parent=self.tlb,$
                     info=[examp,userf_info,'Clear the string to disable this option.'],xsize=30,multiline=4)
             ;replace possible string('A4'xb) with '&'
             while((index=strpos(str,string('A4'xb))) ge 0) do strput,str,'&',index
             ;remove possible '$&'
             while((index=strpos(str,'$&')) ge 0) do strput,str,'  ',index
             changed = (self.userf_str ne str)
             self.userf_str = str
             widget_control,event.id, set_button=(strlen(str) ne 0)
             if ptr_valid(self.projStrPtr) and changed then recalc_proj = 1b    ;need to recalculate projections
          end
        'extravaxis':begin
            self.extravaxis_yn = ~self.extravaxis_yn
            dm_set_button,event.id,self.extravaxis_yn,onstring='Allow Extra Viewing Axis',$
               offstring='Disallow Extra Viewing Axis'
            if self.extravaxis_yn then ptr_free,self.projStrPtr  ;need to recalculate projection
            self.volsel_z = self.volsel_z+(self.extravaxis_yn)*2-1
            self->reset_pdt
          end
        'foldright':begin
            self.foldright = ~self.foldright
            dm_set_button,event.id,self.foldright,onstring='Fold to the Plus Side',$
               offstring='Fold to the Side with Larger Range'
          end  
        'darkcorr':begin
            self.dcor_yn = ~self.dcor_yn
            ptr_free,self.dcor_info
            if self.dcor_yn then begin
               self.filesel->getproperty,path=open_path,file=open_file
               if n_elements(open_file) eq 0 then begin
                  ok = dialog_message(dialog_parent=self.tlb,'Please select dark angle correction files first.',/error)
                  self.dcor_yn = 0b
               endif else begin
                  default = self.dcor_eqran
                  eqran   = dm_dialog_input(['E_start (meV)','E_end (meV)','2theta_start',$
                            '2theta_end'],title='Specify E, 2theta range:',xsize=80,$
                            default=default,info=['The total intensity in the specified range will',$
                            'be used to calculate the dark angle correction.'],dialog_parent=self.tlb)
                  if total(finite(eqran)) ne 4 then $
                     self.dcor_yn = 0b $
                  else if (eqran[0] ge eqran[1]) or (eqran[2] ge eqran[3]) then $
                     self.dcor_yn = 0b
                  if self.dcor_yn eq 0 then $
                     ok = dialog_message('Invalid input. Dark angle correction is disabled.',dialog_parent=self.tlb) $
                  else begin
                     self.dcor_eqran = eqran
                     det_tt = (*self.detPosPtr).two_theta
                     index  = where((abs(det_tt) ge self.dcor_eqran[2]) and (abs(det_tt) le self.dcor_eqran[3]),count)
                     if count eq 0 then begin
                        ok  = dialog_message(['The specified 2theta angle is beyond the detector range.',$
                              'Dark angle correction is skipped.'],dialog_parent=self.tlb)
                        self.dcor_yn = 0b
                     endif else begin
                        dm_load_dcsdata,open_path,open_file,qty=tmp_qty,t_chan=2,e_range=self.dcor_eqran[0:1],$
                              ang=ang,parent=self.tlb,/raw,title='Load Dark Angle Files',eadjust = self.eadjust,$
                              epadjust=self.eadjust_spec,epadcheckfirst=self.eadjust_checkfirst,montype=self.dcs_mon
                        tmp_qty = tmp_qty[index,*]
                        tmp_qty = total(tmp_qty,1)
                        tmp_qty = max(tmp_qty)/tmp_qty
                        self.dcor_info = ptr_new({factor:tmp_qty,a2:reform(ang[0,*])})
                     endelse
                  endelse
               endelse
            endif
            dm_set_button,event.id,self.dcor_yn,onstring='Dark Angle Correction Enabled',$
               offstring='Dark Angle Correction Disabled'
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end
        'edepeff':begin
            self.edep_yn = ~self.edep_yn
            dm_set_button,event.id,self.edep_yn,onstring='E-Dependent Detector Efficiency Correction Enabled',$
                    offstring='E-Dependent Detector Efficiency Correction Disabled'
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end
        'errbar':begin
            self.errbar = ~self.errbar
            dm_set_button,event.id,1-self.errbar,onstring='Error Bar Disabled',offstring='Error Bar Enabled'
          end
        'viewspurion': self->dm_det_spurion
        'useitool': begin
            self.useitool = ~self.useitool
            dm_set_button,event.id,self.useitool,onstring='Use Classical Plotting',offstring='Use iTools Plotting'
          end
        'killall':    self->dm_add_plotwindow,4
        'arrcascade': self->dm_arrange_plot,/cascade
        'arrtable':   self->dm_arrange_plot
        'hpctbut':if strlen(self.helpfile[0]) ne 0 then begin
               online_help,book=self.helpfile[0],/full_path
          endif
        'hpftbut':if strlen(self.helpfile[1]) ne 0 then begin
               online_help,book=self.helpfile[1],/full_path
          endif
        'selfbut':if ptr_valid(self.dataPtr) then $
                     ok = dialog_message(*self.info,/information,dialog_parent=self.tlb,title='About Mslice') $
                  else $
                     ok = dialog_message(*self.info,/information,dialog_parent=self.tlb,title=$
                          'About Mslice for '+(*self.ftypename)[self.ftype])
        'cleardat':  if event.select then begin
            self.monrate = 0d
            ptr_free,self.kfactor,self.dataStrPtr,self.projStrPtr
            self->my_widget_control,['addBut','mkplBut','mkbpBut'],sensitive=0  ;disallow adding data & plotting mask
            widget_control,event.id,sensitive=0                                 ;disallow clear data button
            self->my_widget_control,'titlLab',set_value='' & self.titl_lab=''   ;clear title label
            self->my_widget_control,'legdLab',set_value='' & self.legd_lab=''   ;clear legend label
            self->set_display_button
          end
        'intnlab':  self.intn_lab = tmp_value   ;intensity label
        'titllab':  self.titl_lab = tmp_value   ;title label
        'legdlab':  self.legd_lab = tmp_value   ;legend label
        'powdtyp':  self->change_samptyp,0
        'crysttyp1':self->change_samptyp,1
        'crysttyp2':self->change_samptyp,2
        'crysttyp3':self->change_samptyp,3
        'ebin':  begin    ;ebin size /energy range
            changed  = 0b
            old_ebin = [self.e_bin[self.samp_typ],self.e_bin[4]]
            old_eran = self.e_range 
            if self.samp_typ ge 2 then begin   ;diffuse scattering
               self.e_bin[self.samp_typ] = tmp_value
               if tmp_value eq 2 then begin     ;specify energy range
                  default = dm_to_string(self.e_range)
                  items   = ['from:','to:']
                  info    = ['Energy range is in units of meV.','Data files need to be reloaded.']
                  if self.samp_typ eq 3 then begin
                     default = [default,dm_to_string(self.e_bin[4])]
                     items   = [items,'E rebin size:']
                  endif
                  new = dm_dialog_input(items,title='Energy range',default=default,/float,$
                        dialog_parent=self.tlb,info=info)
                  self.e_bin[self.samp_typ] = 1
                  self.e_range[*] = !values.f_nan
                  if total(finite(new[0:1])) eq 2 then $
                     if new[1] gt new[0] then begin
                        self.e_range = new[0:1]
                        self.e_bin[self.samp_typ] = 2
                     endif
                  if self.e_bin[self.samp_typ] eq 1 then begin
                     ok = dialog_message('Invalid energy range. Elastic peak option is used.',dialog_parent=self.tlb)
                     dm_set_droplist,event.id,select=1
                  endif
                  if self.samp_typ eq 3 then begin  ;rebin size
                     new = fix(new[2])
                     if new le 0 then begin
                        new = 1
                        ok = dialog_message('Invalid rebin size. 1 is used instead.',dialog_parent=self.tlb)
                     endif
                     self.e_bin[4] = new
                  end
               endif else if self.samp_typ eq 3 then begin  ;need to ask for rebin size for multiple E type
                  new = dm_to_string(self.e_bin[4])
                  new = dm_dialog_input('E rebin size:', title='Rebin size',default=new,/int,dialog_parent=self.tlb)
                  if new le 0 then begin
                     new = 1
                     ok = dialog_message('Invalid rebin size. 1 is used instead.',dialog_parent=self.tlb)
                  endif
                  self.e_bin[4] = new   
               endif
            endif else begin    ;single crystal or powder
               self.e_bin[self.samp_typ] = (*self.rebinsize)[tmp_value]
            endelse
            if self.e_bin[self.samp_typ] ne old_ebin[0] then changed = 1b else begin
               if self.samp_typ ge 2 then begin
                  if self.e_bin[self.samp_typ] eq 2 then begin
                     if (total(finite(self.e_range)) ne 2) or (total(finite(old_eran)) ne 2) then changed=1b else begin
                        if total(abs(self.e_range-old_eran)) gt 1e-7 then changed = 1b
                     endelse
                  endif
                  if self.samp_typ eq 3 then begin
                     if self.e_bin[4] ne old_ebin[1] then changed = 1b
                  endif
               endif
            endelse
            if ptr_valid(self.bgdata) and changed then begin
               ptr_free,self.bgdata,self.bgdata2
               if self.extraempcan_yn then self->my_widget_control,'emptBut2',sensitive=0
               self->my_widget_control,'bgclBut',$
                     sensitive=(ptr_valid(self.bgratePtr) or ptr_valid(self.bgtchnPtr) or ptr_valid(self.bgdetsPtr))
               ok = dialog_message('Please reload empty can files.',dialog_parent=self.tlb)
            endif
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end
        'latta': self.latt_parm[0] = dm_to_number(tmp_value,/float)   ;a
        'lattb': self.latt_parm[1] = dm_to_number(tmp_value,/float)   ;b
        'lattc': self.latt_parm[2] = dm_to_number(tmp_value,/float)   ;c
        'lattaa':self.latt_parm[3] = dm_to_number(tmp_value,/float)   ;alpha
        'lattbb':self.latt_parm[4] = dm_to_number(tmp_value,/float)   ;beta
        'lattcc':self.latt_parm[5] = dm_to_number(tmp_value,/float)   ;gama
        'orieux':self.orie_u[0]    = dm_to_number(tmp_value,/float)   ;ux
        'orieuy':self.orie_u[1]    = dm_to_number(tmp_value,/float)   ;uy
        'orieuz':self.orie_u[2]    = dm_to_number(tmp_value,/float)   ;uz
        'orievx':self.orie_v[0]    = dm_to_number(tmp_value,/float)   ;vx
        'orievy':self.orie_v[1]    = dm_to_number(tmp_value,/float)   ;vy
        'orievz':self.orie_v[2]    = dm_to_number(tmp_value,/float)   ;vz
        'oriepsi': begin  ;psi
            tmp_value = dm_to_number(tmp_value,/float)
            if self.orie_psi ne tmp_value then begin
               self.orie_psi    = tmp_value   
               self.orie_offset = !values.f_nan
            endif
          end
        'viewu1typ':if self.view_u1typ ne tmp_value then begin
            self.view_u1typ    = tmp_value
            self.view_u1lab[0] = self.view_label[tmp_value]
            self->my_widget_control,'viewU1lab',set_value=self.view_u1lab[0]
            resetu1 = 1 & islabel = 1
          endif
        'viewu1h': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu1 = finite(tmp)?(finite(self.view_u1[0])?((self.view_u1[0] eq tmp)?0:1):(1)):(0)
            self.view_u1[0] = tmp
          end
        'viewu1k': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu1 = finite(tmp)?(finite(self.view_u1[1])?((self.view_u1[1] eq tmp)?0:1):(1)):(0)
            self.view_u1[1] = tmp
          end
        'viewu1l': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu1 = finite(tmp)?(finite(self.view_u1[2])?((self.view_u1[2] eq tmp)?0:1):(1)):(0)
            self.view_u1[2] = tmp
          end
        'viewu1lab':if self.view_u1lab[(self.samp_typ ge 1)] ne tmp_value then begin
            self.view_u1lab[(self.samp_typ ge 1)] = (strlen(tmp_value) eq 0)?' ':tmp_value
            if self.samp_typ eq 0 then $
               self.view_label[self.view_u1typ] = self.view_u1lab[0]
            resetu1 = 1 & islabel = 1
          endif
        'viewu1fold':  begin
            self.view_u1fold = ~self.view_u1fold
            self->my_widget_control,'viewU1Foldcenter',self.projBas,sensitive=self.view_u1fold
            self->my_widget_control,'viewFoldcenter',self.projBas,sensitive=(self.view_u1fold or self.view_u2fold)
            if self.extravaxis_yn and self.view_u3fold then self->my_widget_control,'viewFoldcenter',self.projBas,/sensitive  
          end 
        'viewu1foldcenter': self.view_u1foldcenter = dm_to_number(tmp_value,/float)
        'viewu2typ':if self.view_u2typ ne tmp_value then begin
            self.view_u2typ    = tmp_value
            self.view_u2lab[0] = self.view_label[tmp_value]
            self->my_widget_control,'viewU2lab',set_value=self.view_u2lab[0]
            resetu2 = 1 & islabel = 1
          endif
        'viewu2h': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu2 = finite(tmp)?(finite(self.view_u2[0])?((self.view_u2[0] eq tmp)?0:1):(1)):(0)
            self.view_u2[0] = tmp
          end
        'viewu2k': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu2 = finite(tmp)?(finite(self.view_u2[1])?((self.view_u2[1] eq tmp)?0:1):(1)):(0)
            self.view_u2[1] = tmp
          end
        'viewu2l': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu2 = finite(tmp)?(finite(self.view_u2[2])?((self.view_u2[2] eq tmp)?0:1):(1)):(0)
            self.view_u2[2] = tmp
          end
        'viewu2lab':if self.view_u2lab[(self.samp_typ ge 1)] ne tmp_value then begin
            self.view_u2lab[(self.samp_typ ge 1)] = (strlen(tmp_value) eq 0)?' ':tmp_value
            if self.samp_typ eq 0 then $
               self.view_label[self.view_u2typ] = self.view_u2lab[0]
            resetu2 = 1 & islabel = 1
          endif
        'viewu2fold':  begin
            self.view_u2fold = ~self.view_u2fold
            self->my_widget_control,'viewU2Foldcenter',self.projBas,sensitive=self.view_u2fold
            self->my_widget_control,'viewFoldcenter',self.projBas,sensitive=(self.view_u1fold or self.view_u2fold)
            if self.extravaxis_yn and self.view_u3fold then self->my_widget_control,'viewFoldcenter',self.projBas,/sensitive
          end 
        'viewu2foldcenter': self.view_u2foldcenter = dm_to_number(tmp_value,/float)
        'viewu3h': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu3 = finite(tmp)?(finite(self.view_u3[0])?((self.view_u3[0] eq tmp)?0:1):(1)):(0)
            self.view_u3[0] = tmp
          end
        'viewu3k': begin
            tmp     = dm_to_number(tmp_value,/float)
            resetu3 = finite(tmp)?(finite(self.view_u3[1])?((self.view_u3[1] eq tmp)?0:1):(1)):(0)
            self.view_u3[1] = tmp
          end
        'viewu3l': begin
            tmp     =  dm_to_number(tmp_value,/float)
            resetu3 = finite(tmp)?(finite(self.view_u3[2])?((self.view_u3[2] eq tmp)?0:1):(1)):(0)
            self.view_u3[2] = tmp
          end
        'viewu3lab':if self.view_u3lab ne tmp_value then begin
            self.view_u3lab = (strlen(tmp_value) eq 0)?' ':tmp_value
            resetu3 = 1 & islabel = 1
          endif
        'viewu3fold':  begin
            self.view_u3fold = ~self.view_u3fold
            self->my_widget_control,'viewU3Foldcenter',self.projBas,sensitive=self.view_u3fold
            self->my_widget_control,'viewFoldcenter',self.projBas,sensitive=(self.view_u1fold or self.view_u2fold or self.view_u3fold)
          end 
        'viewu3foldcenter': self.view_u3foldcenter = dm_to_number(tmp_value,/float)  
        'swapviewaxis1': if event.select then begin
            if self.samp_typ eq 0 then begin
               temp = self.view_u1typ
               self.view_u1typ = self.view_u2typ
               self.view_u2typ = temp
               temp = self.view_u1lab[0]
               self.view_u1lab[0] = self.view_u2lab[0]
               self.view_u2lab[0] = temp
            endif else begin
               temp = self.view_u1
               self.view_u1 = self.view_u2
               self.view_u2 = temp
               temp = self.view_u1lab[1]
               self.view_u1lab[1] = self.view_u2lab[1]
               self.view_u2lab[1] = temp
               temp = self.view_u1fold
               self.view_u1fold = self.view_u2fold
               self.view_u2fold = temp
               temp = self.view_u1foldcenter
               self.view_u1foldcenter = self.view_u2foldcenter
               self.view_u2foldcenter = temp
            endelse
            if (self.samp_typ eq 0) or (self.samp_typ eq 2 and self.extravaxis_yn eq 0) then begin
               temp = self.disp_hran     
               self.disp_hran = self.disp_vran
               self.disp_vran = temp
               temp = self.disp_hstep
               self.disp_hstep = self.disp_vstep
               self.disp_vstep = temp
            endif
            resetu1 = 1 & resetu2 = 1 & islabel = 1
            self->reset_projection,/reset
            if ptr_valid(self.projStrPtr) and (self.samp_typ gt 0) then recalc_proj = 1b    ;need to recalculate projections
          end
        'swapviewaxis2': if event.select then begin
            temp = self.view_u2
            self.view_u2 = self.view_u3
            self.view_u3 = temp
            temp = self.view_u2lab[1]
            self.view_u2lab[1] = self.view_u3lab
            self.view_u3lab  = temp
            temp = self.view_u2fold
            self.view_u2fold = self.view_u3fold
            self.view_u3fold = temp
            temp = self.view_u2foldcenter
            self.view_u2foldcenter = self.view_u3foldcenter
            self.view_u3foldcenter = temp
            resetu2 = 1 & resetu3 = 1 & islabel = 1
            self->reset_projection,/reset
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
          end  
        'calcprj':    if event.select then self->dm_calc_projection    ;calculate projection
        'volselx':    self.volsel_x     = tmp_value
        'volxfrom':   self.vol_xran[0]  = dm_to_number(tmp_value,/float)
        'volxto':     self.vol_xran[1]  = dm_to_number(tmp_value,/float)
        'volxstep':   self.vol_xstep    = dm_to_number(tmp_value,/float)
        'volsely':    self.volsel_y     = tmp_value
        'volyfrom':   self.vol_yran[0]  = dm_to_number(tmp_value,/float)
        'volyto':     self.vol_yran[1]  = dm_to_number(tmp_value,/float)
        'volystep':   self.vol_ystep    = dm_to_number(tmp_value,/float)
        'volselz':    self.volsel_z     = tmp_value
        'volzfrom':   self.vol_zran[0]  = dm_to_number(tmp_value,/float)
        'volzto':     self.vol_zran[1]  = dm_to_number(tmp_value,/float)
        'volzstep':   self.vol_zstep    = dm_to_number(tmp_value,/float)
        'volsel_t':   if self.samp_typ eq -1 then $
                         (*self.volsel_t)[ind_sel] = tmp_value $
                      else $
                         self.volsel_t1 = tmp_value
        'volsel_from':if self.samp_typ eq -1 then $
                         (*self.volsel_tft)[ind_sel,0] = dm_to_number(tmp_value,/float) $
                      else $
                         self.volsel_tft1[0] = dm_to_number(tmp_value,/float)
        'volsel_to':  if self.samp_typ eq -1 then $
                         (*self.volsel_tft)[ind_sel,1] = dm_to_number(tmp_value,/float) $
                      else $
                         self.volsel_tft1[1] = dm_to_number(tmp_value,/float)
        'volifrom':   self.vol_iran[0]  = dm_to_number(tmp_value,/float)
        'volito':     self.vol_iran[1]  = dm_to_number(tmp_value,/float)
        'volsmooth':  self.vol_smooth   = dm_to_number(tmp_value,/int)
        'volinterp':  self.vol_interp   = tmp_value
        'volcolor':   self.vol_color    = tmp_value
        'volplot':    if event.select then self->dm_plot_volume             ;plot volume
        'dispperp': if self.disp_perp ne tmp_value then begin
            self.disp_perp = tmp_value
            self->reset_display,/reset
          endif
        'disppfrom':  self.disp_pran[0] = dm_to_number(tmp_value,/float)
        'disppto':    self.disp_pran[1] = dm_to_number(tmp_value,/float)
        'disptfrom':  self.disp_tran[0] = dm_to_number(tmp_value,/float)
        'disptto':    self.disp_tran[1] = dm_to_number(tmp_value,/float)
        'disphfrom':  self.disp_hran[0] = dm_to_number(tmp_value,/float)
        'disphto':    self.disp_hran[1] = dm_to_number(tmp_value,/float)
        'disphstep':  self.disp_hstep   = dm_to_number(tmp_value,/float)
        'dispvfrom':  self.disp_vran[0] = dm_to_number(tmp_value,/float)
        'dispvto':    self.disp_vran[1] = dm_to_number(tmp_value,/float)
        'dispvstep':  self.disp_vstep   = dm_to_number(tmp_value,/float)
        'dispifrom':  self.disp_iran[0] = dm_to_number(tmp_value,/float)
        'dispito':    self.disp_iran[1] = dm_to_number(tmp_value,/float)
        'dispsmooth': self.disp_smooth  = dm_to_number(tmp_value,/int)
        'dispshading':self.disp_shading = tmp_value
        'dispcolor':  self.disp_color   = tmp_value
        'dispfitslice': if event.select then self->dm_to_fit,/slice
        'dispplotslice':if event.select then self->dm_plot_slice             ;plot slice
        'dispsurfslice':if event.select then self->dm_plot_slice,/surfplot   ;surf slice
        'dispselh':     self.dispsel_h = tmp_value
        'dispselv':     self.dispsel_v = tmp_value
        'dispsel_t':    (*self.dispsel_t)[ind_sel] = tmp_value
        'dispsel_from': (*self.dispsel_tft)[ind_sel,0] = dm_to_number(tmp_value,/float)
        'dispsel_to':   (*self.dispsel_tft)[ind_sel,1] = dm_to_number(tmp_value,/float)
        'cutsel_t':     (*self.cutsel_t)[ind_sel] = tmp_value
        'cutsel_from':  (*self.cutsel_tft)[ind_sel,0] = dm_to_number(tmp_value,/float)  
        'cutsel_to':    (*self.cutsel_tft)[ind_sel,1] = dm_to_number(tmp_value,/float)  
        'cutx':  if self.cut_x[0>(self.samp_typ)] ne tmp_value then begin
            self.cut_x[0>(self.samp_typ)] = tmp_value
            if self.samp_typ ge 0 then self->reset_display,/reset
          endif
        'cutfrom':   self.cut_ran[0]   = dm_to_number(tmp_value,/float)
        'cutto':     self.cut_ran[1]   = dm_to_number(tmp_value,/float)
        'cutstep':   self.cut_step     = dm_to_number(tmp_value,/float)
        'cutt1from': self.cut_t1ran[0] = dm_to_number(tmp_value,/float)
        'cutt1to':   self.cut_t1ran[1] = dm_to_number(tmp_value,/float)
        'cutt2from': self.cut_t2ran[0] = dm_to_number(tmp_value,/float)
        'cutt2to':   self.cut_t2ran[1] = dm_to_number(tmp_value,/float)
        'cutt3from': self.cut_t3ran[0] = dm_to_number(tmp_value,/float)
        'cutt3to':   self.cut_t3ran[1] = dm_to_number(tmp_value,/float)
        'cuty':      self.cut_y[0>(self.samp_typ)] = tmp_value
        'cutyfrom':  self.cut_yran[0]  = dm_to_number(tmp_value,/float)
        'cutyto':    self.cut_yran[1]  = dm_to_number(tmp_value,/float)
        'cutsymbol': self.cut_symbol   = tmp_value
        'cutcolor':  begin
                     self.cut_color    = tmp_value
                     if self.colors[self.cut_color] eq 'user define' then begin
                        define = dm_dialog_input(['r:','g:','b:'],title='Define color',default=self.coldefine,dialog_parent=self.tlb,xsize=80,/coltable)
                        define = 0>(define)<255
                        if total(define-[255b,255b,255b]*self.pbgcol_bw) eq 0 then begin
                           ok = dialog_message('The specified color is the same as the background color. The '+(['white','black'])[self.pbgcol_bw]+' color is used instead.',$
                                dialog_parent=self.tlb,/info)
                           define = 255-define
                           self.cut_color = 0
                           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutColor'),select=self.cut_color
                        endif
                        self.coldefine = define
                     endif
                     end
        'cutline':   self.cut_line     = tmp_value
        'cutfit':    if event.select then self->dm_to_fit
        'cutplot':   if event.select then self->dm_plot_cut
        'cutoplot':  if event.select then self->dm_plot_cut,/overplot
        'trajx':     self.traj_x[self.samp_typ] = tmp_value
        'trajxfrom': self.traj_xran[0] = dm_to_number(tmp_value,/float)
        'trajxto':   self.traj_xran[1] = dm_to_number(tmp_value,/float)
        'trajy':     self.traj_y[self.samp_typ] = tmp_value
        'trajyfrom': self.traj_yran[0] = dm_to_number(tmp_value,/float)
        'trajyto':   self.traj_yran[1] = dm_to_number(tmp_value,/float)
        'trajz1':    self.traj_z1[self.samp_typ] = tmp_value
        'trajz1from':self.traj_z1ran[0] = dm_to_number(tmp_value,/float)
        'trajz1to':  self.traj_z1ran[1] = dm_to_number(tmp_value,/float)
        'trajz2':    self.traj_z2[self.samp_typ] = tmp_value
        'trajz2from':self.traj_z2ran[0] = dm_to_number(tmp_value,/float)
        'trajz2to':  self.traj_z2ran[1] = dm_to_number(tmp_value,/float)
        'plottra':   if event.select then self->dm_plot_trajectory               ;plot trajectory
        'oplottra':  if event.select then self->dm_plot_trajectory,/overplot     ;overplot trajectory
        'status':  begin  ;context menu
            menuid = widget_info(event.id,find_by_uname='contextMenu')
            status  = 'status_'+['geom','int','userf','bin','bincs','extb','binth','montyp','bgrate','bgdet','bgtch','bgcan','van',$
                      'mask','lamcor','eint','ssfac','absp','eadj','edep','fpls','dcor','bitf','neg2th','2thc','2emp']
            nstatus = n_elements(status)
            if !version.os_family eq 'Windows' then begin
               for i=0,nstatus-1 do $
                   self->my_widget_control,status[i],menuid,/destroy
               for i=0,nstatus-1 do begin
                   if status[i] eq 'status_bitf' then $
                      void = dm_widget_button(menuid,_extra=self->getstatus(status[i],/init),uname=status[i],/separator) $
                   else $
                      void = dm_widget_button(menuid,_extra=self->getstatus(status[i],/init),uname=status[i])
               endfor
            endif else begin
               for i=0,nstatus-1 do $
                   self->my_widget_control,status[i],menuid,_extra=self->getstatus(status[i])
            endelse
            self->my_widget_control,'status_dcor',menuid,sensitive=((self.samp_typ ge 2) and (self.ftype eq 0))
            self->my_widget_control,'status_'+['montyp','bitf','2thc'],menuid,sensitive=(self.ftype eq 0)
            self->my_widget_control,'status_lamcor',menuid,sensitive=(self.ftype eq 2)
            self->my_widget_control,'status_fpls',menuid,sensitive=(self.samp_typ gt 0)
            self->my_widget_control,'status_neg2th',menuid,sensitive=(self.samp_typ eq 0)
            widget_displaycontextmenu,event.id,event.x,event.y,menuid
          end
        'status_geom':   ok = self->getstatus(strlowcase(uname),/help)
        'status_userf':  ok = self->getstatus(strlowcase(uname),/help)
        'status_montyp': ok = self->getstatus(strlowcase(uname),/help)
        'status_bgrate': ok = self->getstatus(strlowcase(uname),/help)
        'status_bgtch':  ok = self->getstatus(strlowcase(uname),/help)
        'status_bgdet':  ok = self->getstatus(strlowcase(uname),/help)
        'status_bgcan':  ok = self->getstatus(strlowcase(uname),/help)
        'status_van':    ok = self->getstatus(strlowcase(uname),/help)
        'status_mask':   ok = self->getstatus(strlowcase(uname),/help)
        'status_int':    ok = self->getstatus(strlowcase(uname),/help)
        'status_ssfac':  ok = self->getstatus(strlowcase(uname),/help)
        'status_eint':   ok = self->getstatus(strlowcase(uname),/help)
        'status_lamcor': ok = self->getstatus(strlowcase(uname),/help)
        'status_bin':    ok = self->getstatus(strlowcase(uname),/help)
        'status_bincs':  ok = self->getstatus(strlowcase(uname),/help)
        'status_extb':   ok = self->getstatus(strlowcase(uname),/help)
        'status_binth':  ok = self->getstatus(strlowcase(uname),/help)
        'status_absp':   ok = self->getstatus(strlowcase(uname),/help)
        'status_eadj':   ok = self->getstatus(strlowcase(uname),/help)
        'status_edep':   ok = self->getstatus(strlowcase(uname),/help)
        'status_dcor':   ok = self->getstatus(strlowcase(uname),/help)
        'status_fpls':   ok = self->getstatus(strlowcase(uname),/help)
        'status_bitf':   begin
            self.bitflip = ~self.bitflip
            ok = dialog_message('Bitflip error checking is '+(['off.','on.'])[self.bitflip],dialog_parent=self.tlb,/info)
          end
        'status_neg2th': begin
            self.negative2th = ~self.negative2th
            ;need to recalculate projections if negative 2theta exists
            if ptr_valid(self.detPosPtr) then begin
               if (min((*self.detPosPtr).two_theta) lt 0) and ptr_valid(self.projStrPtr) then recalc_proj = 1b
            endif
            ok = dialog_message((['Absolute 2theta angle value is used.','Negative 2theta angle value is preserved.'])[self.negative2th],dialog_parent=self.tlb,/info)
          end
        'status_2thc': begin
            self.dcs2thcorr = ~self.dcs2thcorr
            ptr_free,self.detPosPtr
            self.detPosPtr = dm_load_detpos(/dcs,adjustdcs=self.dcs2thcorr) 
            if ptr_valid(self.projStrPtr) then recalc_proj = 1b    ;need to recalculate projections
            ok = dialog_message('DCS 2theta correction is '+(['disabled.','enabled.'])[self.dcs2thcorr],dialog_parent=self.tlb,/info)
          end
        'status_2emp': begin
            self.extraempcan_yn = ~self.extraempcan_yn
            ptr_free,self.bgdata2
            self->reset_menu,/backgroundmenuonly  ;reset background menu
            ok = dialog_message((['One set of empty can files are allowed.','Two sets of empty can files are allowed.'])[self.extraempcan_yn],$
                  dialog_parent=self.tlb,/info)
          end  
        else:begin
            passed = self->dm_filetools(uname,event=event)  ;check for file type and file tools events
            if (~ passed) then begin            ;view specified plot window
               all = self.gr_list->get(/all,count=count)
               for i=0,count-1 do begin
                   id = self->dm_get_id(all[i])
                   if id eq uname then begin
                      all[i]->totop
                      break
                   endif
               endfor
            endif
          end
    endcase
    if resetu1 then begin
       if (~ keyword_set(islabel)) then self->adjust_viewlabel,0
       if self.samp_typ gt 0 then self.traj_type[0] = dm_set_viewlabel(self.view_u1lab[1],self.view_u1)
    endif
    if resetu2 then begin
       if (~ keyword_set(islabel)) then self->adjust_viewlabel,1
       if self.samp_typ gt 0 then self.traj_type[1] = dm_set_viewlabel(self.view_u2lab[1],self.view_u2)
    endif
    if resetu3 then begin
       if (~ keyword_set(islabel)) then self->adjust_viewlabel,2
       self.traj_type[2] = dm_set_viewlabel(self.view_u3lab,self.view_u3)
    endif
    if resetu1 or resetu2 or resetu3 then begin
       self->reset_display,/reset
    endif
    if recalc_proj then begin
       ptr_free,self.projStrPtr
       self->set_display_button
    endif
    if strlowcase(uname) ne 'calcprj' then self.error=0    ;reset the error flag
end

pro dcs_mslice::eadjust,option
    old = self.eadjust
    self.eadjust = 0>(option)<2
    if self.eadjust eq 2 then begin
       new  = dm_dialog_input('specify the apparent elastic peak position:',title='Specify elastic peak position',$
              default=self.eadjust_spec,/float,info='in meV',dialog_parent=self.tlb)
       if finite(new) then $
          self.eadjust_spec = new $
       else $
          self.eadjust = old
    endif
    if old eq self.eadjust then return
    epadNone = widget_info(self.tlb,find_by_uname='epadNone')
    epadAuto = widget_info(self.tlb,find_by_uname='epadAuto')
    epadSpec = widget_info(self.tlb,find_by_uname='epadSpec')
    case self.eadjust of 
            0:   dm_toggle_menubut,check=epadNone,uncheck=[epadAuto,epadSpec]
            1:   dm_toggle_menubut,check=epadAuto,uncheck=[epadNone,epadSpec]
            2:   dm_toggle_menubut,check=epadSpec,uncheck=[epadAuto,epadNone]
            else:
    endcase
    self->my_widget_control,'epadCheckfirst',sensitive=(self.eadjust ne 0)
    ;need to reload data
    self.monrate = 0d
    ptr_free,self.kfactor,self.dataStrPtr,self.projStrPtr
    self->my_widget_control,'addBut',sensitive=0     ;disallow adding data
    self->my_widget_control,'clearDat',sensitive=0   ;disallow clear data button
    self->my_widget_control,'titlLab',set_value='' & self.titl_lab=''   ;clear title label
    self->my_widget_control,'legdLab',set_value='' & self.legd_lab=''   ;clear legend label
    self->set_display_button
end

function dcs_mslice::getproperty,tlb=tlb,error=error
    catch, myerror
    if myerror ne 0 then begin
       catch,/cancel
       ok = dialog_message(!error_state.msg,/error)
       error = 1
       return,0
    end
    error = 0
    if keyword_set(tlb) then return,self.tlb
end

pro dcs_mslice::getproperty,tlb=tlb,error=error
    ;catch and ignore all errors in this program
    catch, myerror
    if myerror ne 0 then begin
       catch,/cancel
       ok = dialog_message(!error_state.msg,/error)
       error = 1
       return
    end
    error = 0
    if arg_present(tlb) then tlb=self.tlb
end

;get the status string or keywords for context sensitive menu
function dcs_mslice::getstatus,name,help=help,init=init
    status   = ''
    if keyword_set(help) then begin
       case strlowcase(name) of
           'status_userf': status = ['Go to "Option->'+(lmgr(/runtime)?'User Macro':'User Macro/Function/Procedure')+'" to change this setting.']
           'status_ssfac': status = ['Go to "Parameter->Change Self-shielding Factor..." to change this setting.']
           'status_geom':  status = ['Go to "Option->Instrument Geometry" to change this setting.']
           'status_montyp':status = ['Go to "Option->Monitor Count->DCS Monitor Choice" to change this setting.']
           'status_bgrate':status = ['Go to "Background->Set Overall Background Rate..." to change this value',' or',$
                                     'go to "Background->Load Dark Count File(s)" to change this value',' or',$
                                     'go to "Background->Clear Background" to clear all backgrounds.']
           'status_bgtch': status = ['Go to "Background->Load Time Channel Background File(s)" to change this value',' or',$
                                     'go to "Background->Clear Background" to clear all backgrounds.']
           'status_bgdet': status = ['Go to "Background->Load Detector Background File(s)" to change this value',' or',$
                                     'go to "Background->Clear Background" to clear all backgrounds.']
           'status_bgcan': status = ['Go to "Background->Load Empty Can File(s)" to change this value',' or',$
                                     'go to "Background->Clear Background" to clear all backgrounds.']
           'status_van':   status = ['Go to "Background->Load Vanadium File(s)" to change this value',' or',$
                                     'go to "Background->Clear Vanadium" to clear it.']
           'status_mask':  status = ['Go to Mask menu to change it',' or',$
                                     'go to "Mask->Clear Mask" to clear the existing mask.']
           'status_int':   status = ['Go to "Option->View Intensity As" to change this setting.']
           'status_eint':  status = ['Go to "Option->View Intensity As" to change this setting.']
           'status_lamcor':status = (['Constant time bins with no ki/kf correction.',$
                                      'Constant time bins with ki/kf correction.',$
                                      'Constant energy bins with no ki/kf correction.',$
                                      'Constant energy bins with ki/kf correction.'])[self.lamcorrnxspe]
           'status_bin':   status = ['Go to "Option->Binning Method" to change this setting.']
           'status_bincs': status = ['Go to "Option->Binning Method->Constant Intervals" to change this setting.']
           'status_extb':  status = ['Go to "Option->Binning Method->Use External Binning Library" to change this setting.']
           'status_binth': status = ['Go to "Parameter->Change Binning Threshold..." to change this setting.']
           'status_eadj':  status = ['Go to "Option->Elastic Peak Position Adjustment" to change this setting.']
           'status_edep':  status = ['Go to "Option->E-Dependent Detector Efficiency Correction" to change this setting.']
           'status_dcor':  status = ['Go to "Option->Dark Angle Correction" to change this setting.']
           'status_absp':  status = ['Go to "Option->Apsorption Correction" to change this setting.']
           'status_fpls':  status = ['Go to "Option->Fold to the Plus Side" to change this setting.']
           else:
       endcase
       ok = dialog_message(dialog_parent=self.tlb,status,/info)
       return,status
    endif
    if !version.os_family eq 'Windows' then sep=string(9b)+': ' else sep=': '
    case strlowcase(name) of
        'status_userf': status = (lmgr(/runtime)?'user macro':'user macro/function/procedure')+sep+((strlen(self.userf_str) ne 0)?'yes':'none')
        'status_ssfac': status = 'self-shielding factor'+sep+dm_to_string(self.ssfac)
        'status_geom':  status = 'instrument geometry'+sep+(['direct','inverse'])[self.instrgeom]
        'status_montyp':status = 'monitor type'+sep+(['FC0','BM1','Duration'])[self.dcs_mon]
        'status_bgrate':begin
                        status = 'dark count'+sep
                        if ptr_valid(self.bgratePtr) then begin
                           nvalue = n_elements(*self.bgratePtr)
                           if nvalue eq 1 then $
                              status = status+dm_to_string(*self.bgratePtr)+' cts/hr/det' $
                           else $
                              status = status+'File'
                        endif else $
                           status = status+'none'
                        end
        'status_bgtch': status = 'time channel background'+sep+(ptr_valid(self.bgtchnPtr)?'yes':'none')
        'status_bgdet': status = 'detector background'+sep+(ptr_valid(self.bgdetsPtr)?'yes':'none')
        'status_bgcan': status = 'empty can'+sep+(ptr_valid(self.bgdata)?(ptr_valid(self.bgdata2)?'2 sets':'yes'):'none')
        'status_van':   status = 'detector efficiency'+sep+(ptr_valid(self.eff)?'yes':'none')
        'status_mask':  begin
                        status = 'detector mask'+sep+(ptr_valid(self.mask)?'yes':'none')
                        if (self.ftype eq 0) and (total(self.dcsdetmask) ne 0) then begin
                           status = 'dectector mask'+sep
                           if self.dcsdetmask[0] then status = status+'neg,'
                           if self.dcsdetmask[4] then status = status+'upper,lower,' else begin
                              if self.dcsdetmask[1] then status = status+'lower,'
                              if self.dcsdetmask[3] then status = status+'upper'
                           endelse
                           if self.dcsdetmask[2] then status = status+'central,'
                           if strmid(status,0,1,/reverse) eq ',' then status=strmid(status,0,strlen(status)-1)
                        endif 
                        end
        'status_int':   status = 'intensity'+sep+self.intn_typ_str[self.intn_typ]
        'status_eint':  status = 'E-integrated intensity'+sep+(['no','yes'])[self.eint_yn]
        'status_lamcor':status = 'NXSPE ki/kf correction'+sep+dm_to_string(self.lamcorrnxspe)
        'status_bin':   status = 'binning'+sep+(['average','sum'])[self.bin_avgsum]
        'status_bincs': status = 'constant binning interval'+sep+(['no','yes'])[self.bin_conststep]
        'status_extb':  status = 'use external binning library'+sep+(['no','yes'])[self.bin_extlib]
        'status_binth': status = 'binning threshold'+sep+dm_to_string(self.binthreshold) 
        'status_eadj':  begin
                        status = 'elastic peak adjustment'+sep+(['no','automatic','specified:'+dm_to_string(self.eadjust_spec)])[self.eadjust]
                        if self.eadjust ne 0 then $         
                           status = status+'('+(['Check All','Check 1st Only'])[self.eadjust_checkfirst]+')'
                        end
        'status_edep':  status = 'E-dep efficiency correction'+sep+(['no','yes'])[self.edep_yn]
        'status_dcor':  status = 'dark angle correction'+sep+(['no','yes'])[self.dcor_yn]
        'status_absp':  begin
                        status = 'absorption correction'+sep
                        if ptr_valid(self.absp_info) then begin
                           if (*self.absp_info)[0] ge 0 and (*self.absp_info)[0] lt n_elements(self.absp_npts) then $
                              status = status+(['cylinder','annulus','sphere','flat plate'])[(*self.absp_info)[0]] $
                           else $
                              status = status+'none'
                        endif else $
                           status = status+'none'
                        end
        'status_fpls':  status = 'fold to the plus side'+sep+(['no','yes'])[self.foldright] 
        'status_bitf':  status = 'Check Bitflip Error'+sep+(['no','yes'])[self.bitflip]               
        'status_neg2th':status = 'Preserve Negative 2theta Angle Value'+sep+(['no ','yes'])[self.negative2th]
        'status_2thc':  status = 'DCS 2theta Correction'+sep+(['no','yes'])[self.dcs2thcorr]
        'status_2emp':  status = 'Allow Two Sets of Empty Can Files'+sep+(['no ','yes'])[self.extraempcan_yn]
        else:
    endcase
    ;change the string to keywords
    if keyword_set(init) then $
       result = {value:status} $
    else $
       result = {set_value:status}
    if float(!version.release) ge 5.6 then begin
       tmp = strsplit(status,':',/extract)
       tmp = strtrim(tmp[n_elements(tmp)-1],2)
       set_button = ((tmp ne 'no') and (tmp ne 'none'))
       if strlowcase(name) eq 'status_ssfac'  then set_button = (self.ssfac ne 1.0)
       if strlowcase(name) eq 'status_binth'  then set_button = (self.binthreshold ne 0.0)
       if strlowcase(name) eq 'status_lamcor' then set_button = (self.lamcorrnxspe ne 0) 
       result = create_struct(result,'set_button',set_button)
    endif
    return,result
end

;get the device symbol font in different operating systems
function dcs_mslice::getsymbolfont,fsize
    fonts=['symbol*',$
           '-adobe-symbol-medium-r-normal--14-140-75-75-p-85-adobe-fontspecific',$
           '-adobe-symbol-medium-r-normal--14-100-100-100-p-85-adobe-fontspecific']
    fonts[0] = fonts[0]+dm_to_string(fsize+2,/int)
    catch, myerror
    if myerror ne 0 then begin
       return,fonts[0]
    end
    result = fonts[0]
    switch !version.os_family of
        'Windows':break
        'unix':
        'MacOS':begin
                spawn,'xlsfonts',result,errresult,exit_status=error
                if error ne 0 then begin
                   old_win = !d.window
                   device,get_fontnames=result,set_font='*'    ;will pop up a window, reason for using xlsfonts first
                   if !d.window ne old_win then wdelete, !d.window
                endif
                nlist = n_elements(result)
                tmp   = bytarr(nlist)
                for i=0,nlist-1 do begin
                    tmp[i] = ((strpos(strlowcase(result[i]),'symbol'))<0)+1
                endfor
                index = where(tmp,count)
                if count ne 0 then f_list=result[index]
                nlist = n_elements(f_list)
                if nlist ne 0 then begin
                   tmp = intarr(nlist)
                   for i=0,nlist-1 do tmp[i]=dm_to_number((strsplit(f_list[i],'[A-Za-z\-]',/regex,/extract))[0],/int)
                   void = min(abs(tmp-fsize+2),index)
                   fonts[1] = f_list[index[0]]
                endif
                result = fonts[1]
                break
                end
        else:
    endswitch
    return,result
end

;a widget_control use uname instead of id as identifier
pro dcs_mslice::my_widget_control,uname,base,_ref_extra=extra
    if n_elements(base) eq 0 then base=self.tlb
    if n_elements(extra) ne 0 then begin
       for i=0,n_elements(uname)-1 do begin
           wid = widget_info(base,find_by_uname=uname[i])
           if wid ne 0 then widget_control,wid,_extra=extra
       endfor
    endif
end

;widget_tex with tab_mode=1,/editable,/all_events,/kbrd_focus_events,ysize=1
function dcs_mslice::my_widget_text,base,value=value,xsize=xsize,scr_xsize=scr_xsize,uname=uname
    return,widget_text(base,value=value,xsize=xsize,scr_xsize=scr_xsize,uname=uname,tab_mode=1,/editable,/all_events,/kbrd_focus_events,ysize=1)
end

;check if projection data is available
;returns 0 if not available
function dcs_mslice::pd_valid
    pd_valid = 0b
    if ptr_valid(self.projStrPtr) then begin
       if n_elements((*self.projStrPtr).qty) gt 1 then pd_valid=1b
    endif
    if (~ ptr_valid(self.dataPtr)) and ((~ pd_valid) or self.error) then begin
       ok = dialog_message((['No data.','Please load data and calculate projections first.'])[self.samp_typ ge 0],dialog_parent=self.tlb)
       return,0b
    endif
    return,1b
end


;plot and set mask
;parameter:
;   detsum: [ndet], sum of the intensity at each detector
;   blist:  [ndet], non 0 elements are the index numbers of the masked detectors
;keyword:
;   angle:  if set, plot vs angle instead of vs det number
pro dcs_mslice::plot_mask,detsum,blist,angle=angle
    pblist = ptr_new(blist,/no_copy)
    ;help string
    helptext = [    $
	  "Use the mouse left button to zoom, the right button to mask/unmask. ", $
		"If ''XZOOM'' is displayed on a button in the top row, zooming only ", $
		"occurs in x. If ''XYZOOM'' is displayed zooming occurs in x and y. ", $
		"Click on the ''XZOOM/XYZOOM'' button to toggle between these states. ", $
		"If ''MASK'' is displayed on the other button in the top row, the right ", $
		"button is used to mask one or more detectors. If ''UNMASK'' is displayed ", $
		"the right button unmasks detectors. ", $
		"Click on the ''MASK/UNMASK'' button to toggle between these states. ", $
		"The sliders may also be used to define the displayed detector range. ", $
		"The most recent zoom may be ''undone'' using the ''Undo zoom'' button. ", $
		"To return to the previous user interface click the ''PROCEED'' button."]
    if keyword_set(angle) then $
       dcs_plotdetmask_2th,self.tlb,'',detsum,pblist,helptext $
    else $
       dcs_plotdetmask,self.tlb,'',detsum,pblist,helptext
    ind = where(*pblist gt 0,count)
    self->clear_mask    ;clear the mask
    if count ne 0 then begin
       self.mask = ptr_new((*pblist)[ind]-1)
       self->my_widget_control,['mksvBut','mkclBut'],/sensitive
    endif
    ptr_free,pblist
end

;reset droplist values according to the labels
pro dcs_mslice::reset_droplist
    if self.samp_typ gt 0 then begin
       self.traj_type[0] = dm_set_viewlabel(self.view_u1lab[1],self.view_u1)
       self.traj_type[1] = dm_set_viewlabel(self.view_u2lab[1],self.view_u2)
       self.traj_type[2] = dm_set_viewlabel(self.view_u3lab,self.view_u3)
    endif
end

;(re)create all the data file widgets
pro dcs_mslice::reset_file
    if (~ obj_valid(self.fileSel)) then begin ;create the widgets
       self.fileBas = widget_base(self.tlb,/row,/frame,ypad=1,xpad=0,space=0,map=0)
       widget_control,self.fileBas,update=0
       if !version.os_family eq 'unix' then ysize=11 else ysize=9
       self.fileSel = obj_new('dm_filesel',parent=self.fileBas, filter='.'+(*self.ftypeext)[self.ftype],$
           ysize=ysize,ypad=0,group_leader=self.tlb,path=self.dirs[0])
       fileCol   = widget_base(self.fileBas,/col,/align_bottom,ypad=0,xpad=0,space=0,/context_events)
       sampleRow = widget_base(fileCol,/row,ypad=0,xpad=0,space=0)
       samplab   = widget_label(sampleRow,value='sha ding',/align_center)
       geom      = widget_info(samplab,/geometry)
       if strlen(self.symfont) eq 0 then self.symfont=self->getsymbolfont(2*(fix(geom.scr_ysize)/2))
       self.xsize[1] = geom.xsize
       widget_control,samplab,set_value='Sample Type'
       sampTyp   = dm_widget_button(sampleRow,value='Powder',/menu,xsize=168,uname='sampTyp')
       if float(!version.release) ge 5.6 then $
          widget_control,sampTyp,tooltip='Press to select a sample type.'
       powdBut   = dm_widget_button(sampTyp,value='Powder',uname='powdTyp')
       crystTyp  = widget_button(sampTyp,value='Single Crystal',uname='crystTyp',/menu)
       sct1But   = dm_widget_button(crystTyp,value='Single Crystal',uname='crystTyp1')
       sct2But   = dm_widget_button(crystTyp,value='Diffuse Scattering Single E',uname='crystTyp2')
       sct3But   = dm_widget_button(crystTyp,value='Diffuse Scattering Multiple E',uname='crystTyp3')
       ebinRow   = widget_base(fileCol,/row,ypad=0,xpad=0,space=0)
       ebinLab   = widget_label(ebinRow,value='E rebin size',/align_center,uname='ebinLab')
       eBin      = dm_widget_droplist(ebinRow,xsize=170,uname='eBin')
       labelRow  = widget_base(fileCol,/row,ypad=0,xpad=0,space=0)
       intn_lab  = widget_label(labelRow,value='Intensity Label',/align_center)
       intnLab   = self->my_widget_text(labelRow,value='',uname='intnLab')
       labelRow  = widget_base(fileCol,/row,ypad=0,xpad=0,space=0)
       titl_lab  = widget_label(labelRow,value='Title Label',/align_center)
       titlLab   = self->my_widget_text(labelRow,value='',uname='titlLab')
       labelRow  = widget_base(fileCol,/row,ypad=0,xpad=0,space=0)
       legd_lab  = widget_label(labelRow,value='Legend Label',/align_center)
       legdLab   = self->my_widget_text(labelRow,value='',uname='legdLab')
       loadRow   = widget_base(fileCol,/row,ypad=0,xpad=0,space=0,/align_right)
       loadBut   = widget_button(loadRow,value='Load Data',uname='loadBut',xsize=self.xsize[4],/pushbutton_events)
       addBut    = widget_button(loadRow,value='Add Data',uname='addBut',sensitive=0,xsize=self.xsize[4],/pushbutton_events)
       geom      = widget_info(intn_lab,/geometry)
       xsize     = geom.xsize+40 
       widget_control,samplab,xsize=xsize
       widget_control,intn_lab,xsize=xsize
       widget_control,titl_lab,xsize=xsize
       widget_control,legd_lab,xsize=xsize
       widget_control,ebinLab,xsize=xsize
       geom      = widget_info(loadRow,/geometry)
       widget_control,intnLab,scr_xsize=geom.xsize 
       widget_control,titlLab,scr_xsize=geom.xsize
       widget_control,legdLab,scr_xsize=geom.xsize
       geom      = widget_info(self.fileBas,/geometry)
       self.xsize[0] = geom.scr_xsize
       row       = widget_base(fileCol,/row,xpad=0,ypad=0,/align_right,space=0)
       clearDat  = widget_button(row,value='Clear Data',uname='clearDat',sensitive=0,xsize=self.xsize[4]*2,/pushbutton_events)
       widget_control,self.fileBas,/update,/map
       self.xsize[4] = 67+5*(!version.os_family eq 'Windows')
       ;context sensitive menu
       context = widget_base(fileCol,/CONTEXT_MENU,UNAME='contextMenu')
       if !version.os_family eq 'Windows' then void = widget_button(context,value='',sensitive=0) ;place holder
       status  = 'status_'+['geom','int','userf','bin','bincs','extb','binth','montyp','bgrate','bgdet','bgtch','bgcan','van',$
                 'mask','lamcor','eint','ssfac','absp','eadj','edep','fpls','dcor','bitf','neg2th','2thc','2emp']
       for i=0, n_elements(status)-1 do begin
           if status[i] eq 'status_bitf' then $
              void = dm_widget_button(context,_extra=self->getstatus(status[i],/init),uname=status[i],/separator) $
           else $
              void = dm_widget_button(context,_extra=self->getstatus(status[i],/init),uname=status[i])
       endfor
    endif
    ;reset all values
    tmp = self.samp_typ & self.samp_typ = -1
    self->change_samptyp,tmp,/init
    if self.samp_typ ge 2 then begin    ;diffuse scattering
       self->my_widget_control,'ebinLab',set_value='Energy Range'
       self->my_widget_control,'eBin',set_value=['All','Elastic Peak','Specify']
       tmp = self.e_bin[self.samp_typ]
       if tmp gt 2 then begin
          tmp = 1
          self.e_bin[self.samp_typ] = 1    ;default elastic peak
       end
    endif else begin
       self->my_widget_control,'ebinLab',set_value='E Rebin Size'
       self->my_widget_control,'eBin',set_value=dm_to_string(*(self.rebinsize))
       tmp = where(*(self.rebinsize) eq self.e_bin[self.samp_typ],count)
       if count eq 0 then self.e_bin[self.samp_typ]=1
       tmp = (0>(tmp[0]))
    endelse
    dm_set_droplist,widget_info(self.tlb,find_by_uname='eBin'),select=tmp
    self->my_widget_control,'intnLab',set_value=self.intn_lab
    self->my_widget_control,'titlLab',set_value=self.titl_lab
    self->my_widget_control,'legdLab',set_value=self.legd_lab
end

;(re)create and set values of all the widgets in self.projBas
;keyword:
;   reset:    if set, only set the existing widget values, otherwise regenerate all the widgets
pro dcs_mslice::reset_projection,reset=reset
    if (~ keyword_set(reset)) then begin ;(re)generate all the widgets
       if widget_info(self.projBas,/valid_id) eq 0 then begin
          self.projBas = widget_base(self.tlb,/frame,xpad=0,ypad=0,map=0,space=0,xsize=self.xsize[0])
       endif
       widget_control,self.projBas,update=0
       extra = widget_base(self.projBas,/col,ypad=1,xpad=0)
       switch self.samp_typ of
         0:begin ;powder
           row    = widget_base(extra,/row,ypad=0,xpad=0,space=0)
           left   = widget_base(row,/col,xpad=0,ypad=0)
           label  = widget_label(left,value='  Viewing Axes:',/align_left,uname='viewAxis')
           row    = widget_base(left,/row,ypad=0,xpad=0,space=0)
           label  = widget_label(row,value='',xsize=self.xsize[1]+self.xsize[2])
           label2 = widget_label(row,value='Label',/align_center)
           row    = widget_base(left,/row,ypad=0,xpad=0)
           label  = widget_label(row,value='u1',xsize=self.xsize[1],/align_center)
           viewU1typ = dm_widget_droplist(row,value=self.view_type,xsize=self.xsize[2],uname='viewU1typ')
           viewU1lab = self->my_widget_text(row,value='',xsize=self.xsize[3],uname='viewU1lab')
           geom   = widget_info(viewU1lab,/geometry)
           widget_control,label2,xsize=geom.scr_xsize
           row    = widget_base(left,/row,ypad=0,xpad=0)
           label  = widget_label(row,value='u2',xsize=self.xsize[1],/align_center)
           viewU2typ = dm_widget_droplist(row,value=self.view_type,xsize=self.xsize[2],uname='viewU2typ')
           viewU2lab = self->my_widget_text(row,value='',xsize=self.xsize[3],uname='viewU2lab')
           break
           end
         1:         ;single crystal
         2:         ;diffuse scattering single E
         3:begin    ;diffuse scattering multiple E
           toprow  = widget_base(extra,/row,ypad=0,xpad=0,space=0)
           left    = widget_base(toprow,/col,ypad=0,xpad=0,tab_mode=1)
           row     = widget_base(left,/row,ypad=0,xpad=0,space=0)
           label   = widget_label(row,value='  Unit Cell Lattice Parameters:')
           row     = widget_base(left,/row,ypad=0,xpad=0)
           label   = widget_label(row,value='a ('+string('c5'XB)+')',xsize=self.xsize[1],/align_center)
           lattA   = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='lattA')
           label   = widget_label(row,value='b ('+string('c5'XB)+')',xsize=self.xsize[1],/align_center)
           lattB   = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='lattB')
           label   = widget_label(row,value='c ('+string('c5'XB)+')',xsize=self.xsize[1],/align_center)
           lattC   = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='lattC')
           row     = widget_base(left,/row,ypad=0,xpad=0)
           label   = widget_label(row,value='a ('+string('B0'XB)+')',xsize=self.xsize[1],/align_center,font=self.symfont)
           lattAA  = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='lattAA')
           label   = widget_label(row,value='b ('+string('B0'XB)+')',xsize=self.xsize[1],/align_center,font=self.symfont)
           lattBB  = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='lattBB')
           label   = widget_label(row,value='g ('+string('B0'XB)+')',xsize=self.xsize[1],/align_center,font=self.symfont)
           lattCC  = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='lattCC')
           geom    = widget_info(left,/geometry)
           xsize   = self.xsize[0]-geom.xsize
           right   = widget_base(toprow,/col,ypad=0,xpad=0,xsize=xsize,tab_mode=1)
           xsize   = fix(xsize/6.,type=3)
           row     = widget_base(right,/row,ypad=0,xpad=0,space=0)
           label   = widget_label(row,value='  Crystal Orientation:')
           row     = widget_base(right,/row,ypad=0,xpad=0,space=0)
           label   = widget_label(row,value='ux',xsize=xsize,/align_center)
           orieUx  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='orieUx')
           label   = widget_label(row,value='uy',xsize=xsize,/align_center)
           orieUy  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='orieUy')
           label   = widget_label(row,value='uz',xsize=xsize,/align_center)
           orieUz  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='orieUz')
           row     = widget_base(right,/row,ypad=0,xpad=0,space=0)
           label   = widget_label(row,value='vx',xsize=xsize,/align_center)
           orieVx  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='orieVx')
           label   = widget_label(row,value='vy',xsize=xsize,/align_center)
           orieVy  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='orieVy')
           label   = widget_label(row,value='vz',xsize=xsize,/align_center)
           orieVz  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='orieVz')
           row     = widget_base(right,/row,ypad=0,xpad=0,space=0)
           label   = widget_label(row,value='Psi',/align_center,xsize=xsize)
           oriePsi = self->my_widget_text(row,value='',scr_xsize=xsize,uname='oriePsi')
           oriepsiLab = widget_label(row,value='   angle of ki^u',uname='oriepsiLab',/dynamic_resize)   ; positive if (ki x u)||Qz (up)
           row     = widget_base(extra,/row,ypad=0,xpad=0,space=0)
           left    = widget_base(row,/col,xpad=0,ypad=0)
           label   = widget_label(left,value='  Orthogonal Viewing Axes:',/align_left,uname='viewAxis')
           row     = widget_base(left,/row,ypad=0,xpad=0)
           label   = widget_label(row,value='',scr_xsize=self.xsize[1],/align_center)
           label   = widget_label(row,value='a*',scr_xsize=self.xsize[1],/align_center)
           label   = widget_label(row,value='b*',scr_xsize=self.xsize[1],/align_center)
           label   = widget_label(row,value='c*',scR_xsize=self.xsize[1],/align_center)
           label1  = widget_label(row,value='Label',scr_xsize=self.xsize[1]+1,/align_center)
           fold    = widget_label(row,value='Fold',scr_xsize=self.xsize[1]-1,/align_center)
           label2  = widget_label(row,value='Center',scr_xsize=self.xsize[1],/align_center,uname='viewFoldcenter')
           row     = widget_base(left,/row,ypad=0,xpad=0)
           label   = widget_label(row,value='u1',scr_xsize=self.xsize[1],/align_center)
           viewU1h = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU1h')
           viewU1k = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU1k')
           viewU1l = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU1l')
           viewU1lab = self->my_widget_text(row,value='',scr_xsize=self.xsize[1]+1,uname='viewU1lab')
           tmp       = widget_label(row,value='')
           exclbas   = widget_base(row,/row,xpad=0,ypad=0,space=0,/nonexclusive)
           viewU1fd  = widget_button(exclbas,value='',uname='viewU1Fold',/align_center)
           geom      = widget_info(viewU1fd,/geom)
           if !version.os_family eq 'Windows' then $
              widget_control,tmp,scr_xsize=0>(self.xsize[1]-geom.scr_xsize-1) $
           else $
              widget_control,fold,scr_xsize=(self.xsize[1]+geom.scr_xsize)/2
           viewU1fcn = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU1Foldcenter')
           row     = widget_base(left,/row,ypad=0,xpad=0)
           label   = widget_label(row,value='u2',scr_xsize=self.xsize[1],/align_center)
           viewU2h = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU2h')
           viewU2k = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU2k')
           viewU2l = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU2l')
           viewU2lab = self->my_widget_text(row,value='',scr_xsize=self.xsize[1]+1,uname='viewU2lab')
           tmp       = widget_label(row,value='')
           exclbas   = widget_base(row,/row,xpad=0,ypad=0,space=0,/nonexclusive)
           viewU2fd  = widget_button(exclbas,value='',uname='viewU2Fold',/align_center)
           if !version.os_family eq 'Windows' then $
              widget_control,tmp,scr_xsize=0>(self.xsize[1]-geom.scr_xsize-1)
           viewU2fcn = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU2Foldcenter')
           if self.extravaxis_yn then begin
              swap    = widget_button(row,/pushbutton_events,value='swap',uname='swapViewaxis1')
              row     = widget_base(left,/row,ypad=0,xpad=0)
              label   = widget_label(row,value='u3',scr_xsize=self.xsize[1],/align_center)
              viewU3h = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU3h')
              viewU3k = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU3k')
              viewU3l = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU3l')
              viewU3lab = self->my_widget_text(row,value='',scr_xsize=self.xsize[1]+1,uname='viewU3lab')
              tmp       = widget_label(row,value='')
              exclbas   = widget_base(row,/row,xpad=0,ypad=0,space=0,/nonexclusive)
              viewU3fd  = widget_button(exclbas,value='',uname='viewU3Fold',/align_center)
              if !version.os_family eq 'Windows' then $
                 widget_control,tmp,scr_xsize=0>(self.xsize[1]-geom.scr_xsize-1)
              viewU3fcn = self->my_widget_text(row,value='',scr_xsize=self.xsize[1],uname='viewU3Foldcenter')
           endif
           break
           end
         else:
       endswitch
       swap  = widget_button(row,/pushbutton_events,value='swap',uname='swapViewaxis'+(['1','2'])[self.extravaxis_yn*(self.samp_typ ne 0)])
       right = widget_base(row,/col,xpad=0,ypad=0)
       geom  = widget_info(left,/geom)
       widget_control,right,scr_xsize=self.xsize[0]-geom.xsize-1,/context_events
       row   = widget_base(right,/row,xpad=0,ypad=0,/align_right,space=0)
       calcPrj = widget_button(row,value='Calculate Projections',uname='calcPrj',/pushbutton_events,xsize=self.xsize[4]*2)
       if float(!version.release) ge 5.6 then begin  ;tips for some widget buttons
          if n_elements(viewU1fd) ne 0 then $
             widget_control,viewU1fd,tooltip='Specify the folding center in the text box to the right.'
          if n_elements(viewU2fd) ne 0 then $
             widget_control,viewU2fd,tooltip='Specify the folding center in the text box to the right.'
          if n_elements(viewU3fd) ne 0 then $
             widget_control,viewU3fd,tooltip='Specify the folding center in the text box to the right.'
          self->my_widget_control,'swapViewaxis1',self.projBas,tooltip='Swap viewing axis u1 and u2.'
          if self.extravaxis_yn and (self.samp_typ gt 0) then $
          self->my_widget_control,'swapViewaxis2',self.projBas,tooltip='Swap viewing axis u2 and u3.'
          widget_control,calcPrj,tooltip='Calculate the projection before any plots can be made.'
       endif
       ;context sensitive menu
       context = widget_base(right,/CONTEXT_MENU,UNAME='contextMenu')
       if !version.os_family eq 'Windows' then void = widget_button(context,value='',sensitive=0) ;place holder
       status  = 'status_'+['geom','int','userf','bin','bincs','extb','binth','montyp','bgrate','bgdet','bgtch','bgcan','van',$
                 'mask','lamcor','eint','ssfac','absp','eadj','edep','fpls','dcor','bitf','neg2th','2thc','2emp']
       for i=0, n_elements(status)-1 do begin
           if status[i] eq 'status_bitf' then $
              void = dm_widget_button(context,_extra=self->getstatus(status[i],/init),uname=status[i],/separator) $
           else $
              void = dm_widget_button(context,_extra=self->getstatus(status[i],/init),uname=status[i])
       endfor
    endif
    ;now set all the values
    switch self.samp_typ of
      0:begin   ;powder
        value = self.view_type[0:3+(self.ftype eq 0?1:0)]
        if (self.ftype gt 0) and (self.view_u1typ eq 4) then begin
           self.view_u1typ = 1
           self.view_u1lab[0] = self.view_label[self.view_u1typ]
        endif
        if (self.ftype gt 0) and (self.view_u2typ eq 4) then begin
           self.view_u2typ = 0
           self.view_u2lab[0] = self.view_label[self.view_u2typ]
        endif
        dm_set_droplist,widget_info(self.projBas,find_by_uname='viewU1typ'),set_value=value,select=self.view_u1typ
        dm_set_droplist,widget_info(self.projBas,find_by_uname='viewU2typ'),set_value=value,select=self.view_u2typ
        break
        end
      1:        ;single crystal
      2:        ;diffuse scattering single energy
      3:begin   ;diffuse scattering multiple energy
        self->my_widget_control,'viewAxis',self.projBas,set_value=(self.view_ortharbi?'  Viewing Axes:':'  Orthogonal Viewing Axes:')
        self->my_widget_control,'lattA',self.projBas,set_value=dm_to_string(float(self.latt_parm[0]))
        self->my_widget_control,'lattB',self.projBas,set_value=dm_to_string(float(self.latt_parm[1]))
        self->my_widget_control,'lattC',self.projBas,set_value=dm_to_string(float(self.latt_parm[2]))
        self->my_widget_control,'lattAA',self.projBas,set_value=dm_to_string(float(self.latt_parm[3]))
        self->my_widget_control,'lattBB',self.projBas,set_value=dm_to_string(float(self.latt_parm[4]))
        self->my_widget_control,'lattCC',self.projBas,set_value=dm_to_string(float(self.latt_parm[5]))
        self->my_widget_control,'orieUx',self.projBas,set_value=dm_to_string(self.orie_u[0])
        self->my_widget_control,'orieUy',self.projBas,set_value=dm_to_string(self.orie_u[1])
        self->my_widget_control,'orieUz',self.projBas,set_value=dm_to_string(self.orie_u[2])
        self->my_widget_control,'orieVx',self.projBas,set_value=dm_to_string(self.orie_v[0])
        self->my_widget_control,'orieVy',self.projBas,set_value=dm_to_string(self.orie_v[1])
        self->my_widget_control,'orieVz',self.projBas,set_value=dm_to_string(self.orie_v[2])
        self->my_widget_control,'oriePsi',self.projBas,set_value=dm_to_string(self.orie_psi)
        self->my_widget_control,'viewU1h',self.projBas,set_value=dm_to_string(self.view_u1[0])
        self->my_widget_control,'viewU1k',self.projBas,set_value=dm_to_string(self.view_u1[1])
        self->my_widget_control,'viewU1l',self.projBas,set_value=dm_to_string(self.view_u1[2])
        self->my_widget_control,'viewU2h',self.projBas,set_value=dm_to_string(self.view_u2[0])
        self->my_widget_control,'viewU2k',self.projBas,set_value=dm_to_string(self.view_u2[1])
        self->my_widget_control,'viewU2l',self.projBas,set_value=dm_to_string(self.view_u2[2]) 
        self->my_widget_control,'viewU1Fold',self.projBas,set_button=self.view_u1fold
        self->my_widget_control,'viewU1Foldcenter',self.projBas,set_value=dm_to_string(self.view_u1foldcenter),sensitive=self.view_u1fold
        self->my_widget_control,'viewU2Fold',self.projBas,set_button=self.view_u2fold
        self->my_widget_control,'viewU2Foldcenter',self.projBas,set_value=dm_to_string(self.view_u2foldcenter),sensitive=self.view_u2fold
        self->my_widget_control,'viewFoldcenter',self.projBas,sensitive=(self.view_u1fold or self.view_u2fold)
        if self.extravaxis_yn then begin
           self->my_widget_control,'viewU3h',self.projBas,set_value=dm_to_string(self.view_u3[0])
           self->my_widget_control,'viewU3k',self.projBas,set_value=dm_to_string(self.view_u3[1])
           self->my_widget_control,'viewU3l',self.projBas,set_value=dm_to_string(self.view_u3[2])
           self->my_widget_control,'viewU3lab',self.projBas,set_value=self.view_u3lab
           self->my_widget_control,'viewU3Fold',self.projBas,set_button=self.view_u3fold
           self->my_widget_control,'viewU3Foldcenter',self.projBas,set_value=dm_to_string(self.view_u3foldcenter),sensitive=self.view_u3fold
           self->my_widget_control,'viewFoldcenter',self.projBas,sensitive=(self.view_u1fold or self.view_u2fold or self.view_u3fold)
        endif 
        self->my_widget_control,'oriepsiLab',self.projBas,set_value=((self.samp_typ eq 1)?'   ':'   first ')+'angle of ki^u'      ; positive if (ki x u)||Qz (up)
        break
        end
      else:
    endswitch
    self->my_widget_control,'viewU1lab',self.projBas,set_value=self.view_u1lab[(self.samp_typ ge 1)]
    self->my_widget_control,'viewU2lab',self.projBas,set_value=self.view_u2lab[(self.samp_typ ge 1)]
end

;reset the data content for self.samp_typ=-1
pro dcs_mslice::reset_data,data=data,author=author,label=label,instrument=instrument,is_uniq=is_uniq,unit=unit,no_copy=no_copy
    if n_elements(author) eq 1 then begin
       date = strsplit((*self.info)[3],' ',/extract) & date = date[n_elements(date)-1]
       (*self.info)[3] = 'Yiming Qiu(yiming.qiu@nist.gov), '+author+'   '+date 
    endif
    if self.samp_typ ne -1 then return
    reset   = 1b
    n_item  = n_elements((*self.dataPtr).label)
    n_label = n_elements(label)
    n_unit  = n_elements(unit)
    n_uniq  = n_elements(is_uniq)
    n_instr = n_elements(instrument) 
    if n_elements(data) ne 0 then begin
       o_item = n_item
       n_item = n_elements(data[0,*])
       if n_item lt 4 then begin
          ok = dialog_message('The data must be [ndat,nitem] with nitem>3.',dialog_parent=self.tlb)
          return
       endif
       instrument1 = (*self.dataPtr).instrument
       label1      = (*self.dataPtr).label
       unit1       = (*self.dataPtr).unit
       is_uniq1    = bytarr(n_item)
       if n_item ne o_item then begin
          reset    = 0b   ;need to reconstruct display
          label1   = ['I','dI','x'+dm_to_string(indgen(n_item-2)+1)]
          unit1    = strarr(n_item)
       endif
       ptr_free,self.volsel_t,self.volsel_tft,self.dispsel_t,self.dispsel_tft,self.cutsel_t,self.cutsel_tft
       if n_item gt 5 then begin
          self.volsel_t   = ptr_new(indgen(n_item-5)+3)
          self.volsel_tft = ptr_new(fltarr(n_item-5,2))
          (*self.volsel_tft)[*] = !values.f_nan 
       endif
       if n_item gt 4 then begin
          self.dispsel_t   = ptr_new(indgen(n_item-4)+2)
          self.dispsel_tft = ptr_new(fltarr(n_item-4,2))
          (*self.dispsel_tft)[*] = !values.f_nan 
       endif
       if n_item gt 3 then begin
          self.cutsel_t    = ptr_new(indgen(n_item-3)+1)
          self.cutsel_tft  = ptr_new(fltarr(n_item-3,2))
          (*self.cutsel_tft)[*] = !values.f_nan 
       endif
       (*self.dataPtr) = create_struct('data',keyword_set(no_copy)?temporary(data):data,$
                      'label',label1,'unit',unit1,'is_uniq',is_uniq1,'instrument',instrument1)
       if finite(self.mask_filler) then begin
          ptr_free,self.dataStrPtr
          index = where((*self.dataPtr).data[*,0] gt self.mask_filler,count)
          if count ne 0 then begin
             ok = dialog_message('The existing mask value of '+dm_to_string(self.mask_filler)+' has been used.',dialog_parent=self.tlb,/info)
             self.dataStrPtr = ptr_new((*self.dataPtr))
             (*self.dataPtr) = {data:(*self.dataStrPtr).data[index,*],label:(*self.dataStrPtr).label,$
                                unit:(*self.dataStrPtr).unit,is_uniq:(*self.dataStrPtr).is_uniq,instrument:(*self.dataStrPtr).instrument}
          endif else begin
             self.mask_filler = !values.f_nan
             widget_control,widget_info(self.tlb,find_by_uname='maskFiller'),set_button=0b
          endelse
       endif
    endif
       
    if n_label ne 0 then begin
       (*self.dataPtr).label[0:(n_item<n_label)-1] = label[0:(n_item<n_label)-1]
    endif
    
    if n_unit ne 0 then begin
       (*self.dataPtr).unit[0:(n_item<n_unit)-1] = unit[0:(n_item<n_unit)-1]
    endif

    if n_uniq ne 0 then begin
       (*self.dataPtr).is_uniq[0:(n_item<n_uniq)-1] = is_uniq[0:(n_item<n_uniq)-1]
    endif

    if n_instr ne 0 then begin
       (*self.info)[0] = '--Mslice'+(strlen(instrument[0]) eq 0?'':' for ')+instrument[0]+' --'       
       widget_control,self.tlb,tlb_set_title=(*self.info)[0]
       (*self.dataPtr).instrument = instrument[0]
    endif
    
    if (~ keyword_set(reset)) then begin
       geom_old = widget_info(self.dispBas,/geometry)
       widget_control,self.tlb,update=0
       widget_control,self.dispBas,/destroy
    endif
    self->reset_display,reset=reset
    self->my_widget_control,'savevolumeMenu',sensitive=(n_itme ge 5)
    if (~ keyword_set(reset)) then begin
       if float(!version.release) ge 5.5 then begin
          self.geom[1] = self.geom[1]+geom.ysize-geom_old.ysize
          widget_control,self.tlb,ysize=self.geom[1]
       endif else begin
          self.geom[1] = self.geom[1]+geom.scr_ysize-geom_old.scr_ysize
          widget_control,self.tlb,scr_ysize=self.geom[1]
       endelse
       widget_control,self.dispBas,/update,/map
       widget_control,self.tlb,/update
    endif
end

;(re)create and set values of all the widgets in self.dispBas
;keyword:
;   reset:    if set, only set the existing widget values, otherwise regenerate all the widgets
pro dcs_mslice::reset_display,reset=reset
    z2_yn = self.extravaxis_yn and (self.samp_typ eq 1 or self.samp_typ eq 3)  ;flag for z2 axis in trajectory
    z1_yn = (self.samp_typ ne 2) or self.extravaxis_yn                         ;flag for z1 axis in trajectory
    if (~ keyword_set(reset)) then begin   ;(re)generate all the widgets
       if widget_info(self.dispBas,/valid_id) eq 0 then begin
          if self.samp_typ eq -1 then $ ;non-tab window
             self.dispBas = widget_base(self.tlb,/frame,xpad=0,ypad=0,map=0,space=0) $
          else $
             self.dispBas = widget_base(self.tlb,/frame,xpad=0,ypad=0,map=0,space=0,xsize=self.xsize[0])
       endif
       widget_control,self.dispBas,update=0
       xsize = fix((self.xsize[0]-3.*self.xsize[1]-self.xsize[2])/4.,type=3)
       self.xsize[4] = xsize
       if self.samp_typ eq -1 then begin ;non-tab window
          extra = widget_base(self.dispBas,/col,xpad=0,ypad=1) 
          nitem = n_elements((*self.dataPtr).label)
          sel_list = (*self.dataPtr).label[2:nitem-1]
          row   = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label = widget_label(row,value='sha ding',/align_center) 
          geom  = widget_info(label,/geometry)
          self.xsize[1] = geom.xsize
          if nitem ge 5 then begin  
             widget_control,label,set_value='  Volume:'
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='x',xsize=self.xsize[1],/align_center)
             volselx   = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='volselX',select=self.volsel_x)
             label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             volxfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volxfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volxto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volxto')
             label     = widget_label(row,value='step',xsize=xsize,/align_center)
             volxstep  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volxstep')   
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='y',xsize=self.xsize[1],/align_center)
             volsely   = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='volselY',select=self.volsel_y)
             label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             volyfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volyfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volyto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volyto')
             label     = widget_label(row,value='step',xsize=xsize,/align_center)
             volystep  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volystep')   
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='z',xsize=self.xsize[1],/align_center)
             volselz   = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='volselZ',select=self.volsel_z)
             label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             volzfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volzfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volzto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volzto')
             label     = widget_label(row,value='step',xsize=xsize,/align_center)
             volzstep  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volzstep')   
             for i=0,nitem-6 do begin
                 row      = widget_base(extra,/row,xpad=0,ypad=0,space=0)
                 label    = widget_label(row,value='',xsize=self.xsize[1],/align_center)
                 voltsel  = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='volsel_T'+dm_to_string(i),select=(*self.volsel_t)[i])
                 label    = widget_label(row,value='thick',xsize=self.xsize[1],/align_center)
                 voltfrom = self->my_widget_text(row,value=dm_to_string((*self.volsel_tft)[i,0]),scr_xsize=xsize,uname='volselfrom_T'+dm_to_string(i))
                 label    = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
                 voltto   = self->my_widget_text(row,value=dm_to_string((*self.volsel_tft)[i,1]),scr_xsize=xsize,uname='volselto_T'+dm_to_string(i))
             endfor
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label     = widget_label(row,value='intensity range',xsize=self.xsize[1]+self.xsize[2],/align_center)
             volIfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volIfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volIto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volIto')
             label     = widget_label(row,value='color    map ',xsize=xsize*1.5,/align_center)
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label     = widget_label(row,value='smoothing level',xsize=self.xsize[1]+self.xsize[2],/align_center)
             volSmooth = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volSmooth')
             label     = widget_label(row,value='interp.',xsize=self.xsize[1],/align_center)
             volInterp = dm_widget_droplist(row,value=['Nearest Neighb','Trilinear'],xsize=xsize,uname='volInterp')
             volColor  = dm_widget_droplist(row,value=self.dispcmap,xsize=xsize*2,uname='volColor')
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=3*self.xsize[1]+self.xsize[2]+xsize*2,/align_center)
             volPlot   = widget_button(row,/pushbutton_events,value='Plot Volume',xsize=xsize*2,uname='volPlot')
             row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='  Slice Plane:',/align_center) 
          endif else $
             widget_control,label,set_value='  Slice:'
          row          = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label        = widget_label(row,value='x',xsize=self.xsize[1],/align_center)
          dispselH     = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='dispselH',select=self.dispsel_h)
          label        = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
          dispHfrom    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispHfrom')
          label        = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          dispHto      = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispHto')
          label        = widget_label(row,value='step',xsize=xsize,/align_center)
          dispHstep    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispHstep')   
          row          = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label        = widget_label(row,value='y',xsize=self.xsize[1],/align_center)
          dispselV     = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='dispselV',select=self.dispsel_v)
          label        = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
          dispVfrom    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispVfrom')
          label        = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          dispVto      = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispVto')
          label        = widget_label(row,value='step',xsize=xsize,/align_center)
          dispVstep    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispVstep')
          for i=0,nitem-5 do begin
              row      = widget_base(extra,/row,xpad=0,ypad=0,space=0)
              label    = widget_label(row,value='',xsize=self.xsize[1],/align_center)
              disptsel = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='dispsel_T'+dm_to_string(i),select=(*self.dispsel_t)[i])
              label    = widget_label(row,value='thick',xsize=self.xsize[1],/align_center)
              disptfrom= self->my_widget_text(row,value=dm_to_string((*self.dispsel_tft)[i,0]),scr_xsize=xsize,uname='dispselfrom_T'+dm_to_string(i))
              label    = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
              disptto  = self->my_widget_text(row,value=dm_to_string((*self.dispsel_tft)[i,1]),scr_xsize=xsize,uname='dispselto_T'+dm_to_string(i))
          endfor
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='intensity range',xsize=self.xsize[1]+self.xsize[2],/align_center)
          dispIfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispIfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          dispIto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispIto')
          label     = widget_label(row,value='color    map ',xsize=xsize*1.5,/align_center)
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='smoothing level',xsize=self.xsize[1]+self.xsize[2],/align_center)
          dispSmooth = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispSmooth')
          label     = widget_label(row,value='shading',xsize=self.xsize[1],/align_center)
          dispShading = dm_widget_droplist(row,value=['Flat','Gouraud'],xsize=xsize,uname='dispShading')
          dispColor = dm_widget_droplist(row,value=self.dispcmap,xsize=xsize*2,uname='dispColor')
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=3*self.xsize[1]+self.xsize[2]+xsize,/align_center)
          dispFitSlice  = widget_button(row,/pushbutton_events,value='Fit',xsize=xsize,uname='dispFitSlice')
          dispPlotSlice = widget_button(row,/pushbutton_events,value='Plot Slice',xsize=xsize,uname='dispPlotSlice')
          dispSurfSlice = widget_button(row,/pushbutton_events,value='Surf Slice',xsize=xsize,uname='dispSurfSlice')
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0,/align_left)
          label     = widget_label(row,value='  Cut:',xsize=self.xsize[1],/align_center)
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='along',xsize=self.xsize[1],/align_center)
          cutX      = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='cutX',select=self.cut_x[0])
          label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
          cutFrom   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutFrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          cutTo     = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutTo')
          label     = widget_label(row,value='step',xsize=xsize,/align_center)
          cutStep   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutStep')   
          for i=0,nitem-4 do begin
              row      = widget_base(extra,/row,xpad=0,ypad=0,space=0)
              label    = widget_label(row,value='',xsize=self.xsize[1],/align_center)
              cuttsel  = dm_widget_droplist(row,value=sel_list,xsize=self.xsize[2],uname='cutsel_T'+dm_to_string(i),select=(*self.cutsel_t)[i])
              label    = widget_label(row,value='thick',xsize=self.xsize[1],/align_center)
              cuttfrom = self->my_widget_text(row,value=dm_to_string((*self.cutsel_tft)[i,0]),scr_xsize=xsize,uname='cutselfrom_T'+dm_to_string(i))
              label    = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
              cuttto   = self->my_widget_text(row,value=dm_to_string((*self.cutsel_tft)[i,1]),scr_xsize=xsize,uname='cutselto_T'+dm_to_string(i))
          endfor
          label     = widget_label(row,value='symbol',xsize=xsize,/align_center)
          cutColor  = dm_widget_droplist(row,value=self.colors,xsize=xsize,uname='cutColor')
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='y',xsize=self.xsize[1],/align_center)
          cutY      = dm_widget_droplist(row,value=['Intensity',sel_list],xsize=self.xsize[2],uname='cutY',select=self.cut_y[0]) ;volume data
          label     = widget_label(row,value='range',xsize=self.xsize[1],/align_center)
          cutYfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutYfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          cutYto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutYto')
          cutSymbol = dm_widget_droplist(row,value=self.symbols,xsize=xsize,uname='cutSymbol')
          cutLine   = dm_widget_droplist(row,value=self.lines,xsize=xsize,uname='cutLine')
          row       = widget_base(extra,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=3*self.xsize[1]+self.xsize[2]+xsize,/align_center)
          cutFit    = widget_button(row,/pushbutton_events,value='Fit',xsize=xsize,uname='cutFit')
          cutPlot   = widget_button(row,/pushbutton_events,value='Plot Cut',xsize=xsize,uname='cutPlot')
          cutOplot  = widget_button(row,/pushbutton_events,value='Oplot Cut',xsize=xsize,uname='cutOplot')
       endif else begin ;tab window
          ;adjust button size
          calcPrj     = widget_info(self.projBas,find_by_uname='calcPrj')
          if calcPrj ne 0 then $
             widget_control,calcPrj,scr_xsize=self.xsize[4]*2
          extra  = widget_tab(self.dispBas,tab_mode=0,uname='dispTab')
          sliTab = widget_base(extra,/col,title='Slice',uname='sliceTab',xpad=0,ypad=10)
          cutTab = widget_base(extra,/col,title='Cut',uname='cutTab',xpad=0,ypad=10)
          tmp = widget_info(sliTab,/geometry) 
          oldxsize1 = self.xsize[1]
          self.xsize[1] = self.xsize[1]-tmp.space
          ;volume
          if (self.samp_typ eq 1) or (self.samp_typ eq 2 and self.extravaxis_yn) or (self.samp_typ eq 3) then begin
             volTab = widget_base(extra,/col,title='Volume',uname='volTab',xpad=0,ypad=10)
             row       = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='x',xsize=self.xsize[1],/align_center)
             volselx   = dm_widget_droplist(row,xsize=self.xsize[2],uname='volselX',select=self.volsel_x)
             label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             volxfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volxfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volxto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volxto')
             label     = widget_label(row,value='step',xsize=xsize,/align_center)
             volxstep  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volxstep')   
             row       = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='y',xsize=self.xsize[1],/align_center)
             volsely   = dm_widget_droplist(row,xsize=self.xsize[2],uname='volselY',select=self.volsel_y)
             label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             volyfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volyfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volyto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volyto')
             label     = widget_label(row,value='step',xsize=xsize,/align_center)
             volystep  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volystep')   
             row       = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='z',xsize=self.xsize[1],/align_center)
             volselz   = dm_widget_droplist(row,xsize=self.xsize[2],uname='volselZ',select=self.volsel_z)
             label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             volzfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volzfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volzto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volzto')
             label     = widget_label(row,value='step',xsize=xsize,/align_center)
             volzstep  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volzstep')   
             nitem     = [0,3,2,3]+self.extravaxis_yn
             for i=0,nitem[self.samp_typ]-4 do begin
                 row      = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
                 label    = widget_label(row,value='',xsize=self.xsize[1],/align_center)
                 voltsel  = dm_widget_droplist(row,xsize=self.xsize[2],uname='volsel_T'+dm_to_string(i))
                 label    = widget_label(row,value='thick',xsize=self.xsize[1],/align_center)
                 voltfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volselfrom_T'+dm_to_string(i))
                 label    = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
                 voltto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volselto_T'+dm_to_string(i))
             endfor
             row       = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label     = widget_label(row,value='intensity range',xsize=self.xsize[1]+self.xsize[2],/align_center)
             volIfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volIfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             volIto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volIto')
             label     = widget_label(row,value='color    map ',xsize=xsize*1.5,/align_center)
             row       = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label     = widget_label(row,value='smoothing level',xsize=self.xsize[1]+self.xsize[2],/align_center)
             volSmooth = self->my_widget_text(row,value='',scr_xsize=xsize,uname='volSmooth')
             label     = widget_label(row,value='interp.',xsize=self.xsize[1],/align_center)
             volInterp = dm_widget_droplist(row,value=['Nearest Neighb','Trilinear'],xsize=xsize,uname='volInterp')
             volColor  = dm_widget_droplist(row,value=self.dispcmap,xsize=xsize*2,uname='volColor')
             row       = widget_base(volTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=3*self.xsize[1]+self.xsize[2]+xsize*2,/align_center)
             volPlot   = widget_button(row,/pushbutton_events,value='Plot Volume',xsize=xsize*2,uname='volPlot')    
             row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='perp to',xsize=self.xsize[1],/align_center)
             dispPerp  = dm_widget_droplist(row,value=self.traj_type[0:2],xsize=self.xsize[2],uname='dispPerp')
             label     = widget_label(row,value='thick',xsize=self.xsize[1],/align_center)
             dispPfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispPfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             dispPto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispPto')
          endif 
          ;slice
          row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='H Label',xsize=self.xsize[1]+self.xsize[2],/align_center,uname='h_label')
          dispHfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispHfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          dispHto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispHto')
          label     = widget_label(row,value='step',xsize=xsize,/align_center)
          dispHstep = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispHstep')
          row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='V Label',xsize=self.xsize[1]+self.xsize[2],/align_center,uname='v_label')
          dispVfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispVfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          dispVto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispVto')
          label     = widget_label(row,value='step',xsize=xsize,/align_center)
          dispVstep = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispVstep')
          if ((self.samp_typ eq 1) or (self.samp_typ eq 3)) and self.extravaxis_yn then begin
             row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
             label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label     = widget_label(row,value='T Label',xsize=self.xsize[1]+self.xsize[2],/align_center,uname='t_label')
             disptfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispTfrom')
             label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             disptto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispTto')
          endif
          row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='intensity range',xsize=self.xsize[1]+self.xsize[2],/align_center)
          dispIfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispIfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          dispIto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispIto')
          label     = widget_label(row,value='color    map ',xsize=xsize*1.5,/align_center)
          row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='smoothing level',xsize=self.xsize[1]+self.xsize[2],/align_center)
          dispSmooth = self->my_widget_text(row,value='',scr_xsize=xsize,uname='dispSmooth')
          label     = widget_label(row,value='shading',xsize=self.xsize[1],/align_center)
          dispShading = dm_widget_droplist(row,value=['Flat','Gouraud'],xsize=xsize,uname='dispShading')
          dispColor = dm_widget_droplist(row,value=self.dispcmap,xsize=xsize*2,uname='dispColor')
          row       = widget_base(sliTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=3*self.xsize[1]+self.xsize[2]+xsize,/align_center)
          dispFitSlice  = widget_button(row,/pushbutton_events,value='Fit',xsize=xsize,uname='dispFitSlice')
          dispPlotSlice = widget_button(row,/pushbutton_events,value='Plot Slice',xsize=xsize,uname='dispPlotSlice')
          dispSurfSlice = widget_button(row,/pushbutton_events,value='Surf Slice',xsize=xsize,uname='dispSurfSlice')
          ;cut
          row       = widget_base(cutTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='along',xsize=self.xsize[1],/align_center)
          case self.samp_typ of
            0: cutX = dm_widget_droplist(row,value=[self.view_u1lab[0],self.view_u2lab[0]],xsize=self.xsize[2],uname='cutX') ;powder
            1: cutX = dm_widget_droplist(row,value=self.traj_type[0:2],xsize=self.xsize[2],uname='cutX')    ;single crystal
            2: cutX = dm_widget_droplist(row,value=self.traj_type[0:1],xsize=self.xsize[2],uname='cutX')    ;diffuse scattering single E
            3: cutX = dm_widget_droplist(row,value=self.traj_type[0:2],xsize=self.xsize[2],uname='cutX')    ;diffuse scattering multipe E
            else:
          endcase
          label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
          cutFrom   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutFrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          cutTo     = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutTo')
          label     = widget_label(row,value='step',xsize=xsize,/align_center)
          cutStep   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutStep')
          row       = widget_base(cutTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=self.xsize[1],/align_center)
          label     = widget_label(row,value='T1 Label',xsize=self.xsize[1]+self.xsize[2],/align_center,uname='t1_label')
          cutT1from = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutT1from')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          cutT1to   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutT1to')
          if (self.samp_typ eq 1) or ((self.samp_typ eq 2) and self.extravaxis_yn) or (self.samp_typ eq 3) then begin
             row    = widget_base(cutTab,/row,xpad=0,ypad=0,space=0)
             label  = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label  = widget_label(row,value='T2 Label',xsize=self.xsize[1]+self.xsize[2],/align_center,uname='t2_label')
             cutT2from = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutT2from')
             label  = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             cutT2to   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutT2to')
          endif
          if (self.samp_typ eq 1 or self.samp_typ eq 3) and self.extravaxis_yn then begin
             row    = widget_base(cutTab,/row,xpad=0,ypad=0,space=0)
             label  = widget_label(row,value='',xsize=self.xsize[1],/align_center)
             label  = widget_label(row,value='T3 Label',xsize=self.xsize[1]+self.xsize[2],/align_center,uname='t3_label')
             cutT3from = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutT3from')
             label  = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             cutT3to   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutT3to')
          endif
          label     = widget_label(row,value='symbol',xsize=xsize,/align_center)
          cutColor  = dm_widget_droplist(row,value=self.colors,xsize=xsize,uname='cutColor')
          row       = widget_base(cutTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='y',xsize=self.xsize[1],/align_center)
          case self.samp_typ of
            0: cutY = dm_widget_droplist(row,value=['Intensity',self.view_type[0:3]],xsize=self.xsize[2],uname='cutY') ;powder
            1: cutY = dm_widget_droplist(row,value=['Intensity',self.traj_type[0:3]],xsize=self.xsize[2],uname='cutY') ;single crystal
            2: cutY = dm_widget_droplist(row,value=['Intensity',self.traj_type[0:1]],xsize=self.xsize[2],uname='cutY') ;diffuse scattering single E
            3: cutY = dm_widget_droplist(row,value=['Intensity',self.traj_type[0:3]],xsize=self.xsize[2],uname='cutY') ;diffuse scattering multiple E
            else:
          endcase
          label     = widget_label(row,value='range',xsize=self.xsize[1],/align_center)
          cutYfrom  = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutYfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          cutYto    = self->my_widget_text(row,value='',scr_xsize=xsize,uname='cutYto')
          cutSymbol = dm_widget_droplist(row,value=self.symbols,xsize=xsize,uname='cutSymbol')
          cutLine   = dm_widget_droplist(row,value=self.lines,xsize=xsize,uname='cutLine')
          row       = widget_base(cutTab,/row,xpad=0,ypad=0,space=0)
          label     = widget_label(row,value='',xsize=3*self.xsize[1]+self.xsize[2]+xsize,/align_center)
          cutFit    = widget_button(row,/pushbutton_events,value='Fit',xsize=xsize,uname='cutFit')
          cutPlot   = widget_button(row,/pushbutton_events,value='Plot Cut',xsize=xsize,uname='cutPlot')
          cutOplot  = widget_button(row,/pushbutton_events,value='Oplot Cut',xsize=xsize,uname='cutOplot')
          ;trajectory
          trajTab   = widget_base(extra,/row,title='Detector Trajectory',uname='trajTab',xpad=0,ypad=10)
          left      = widget_base(trajTab,/col,xpad=0,ypad=0)
          row       = widget_base(left,/row,ypad=0,xpad=0,space=0)
          label     = widget_label(row,value='x',xsize=self.xsize[1],/align_center)
          trajX     = dm_widget_droplist(row,value=self.traj_type,xsize=self.xsize[2],uname='trajX')
          label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
          trajXfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajXfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          trajXto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajXto')
          row       = widget_base(left,/row,ypad=0,xpad=0,space=0)
          label     = widget_label(row,value='y',xsize=self.xsize[1],/align_center)
          trajY     = dm_widget_droplist(row,value=self.traj_type,xsize=self.xsize[2],uname='trajY')
          label     = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
          trajYfrom = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajYfrom')
          label     = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
          trajYto   = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajYto')
          if z1_yn then begin
             row    = widget_base(left,/row,ypad=0,xpad=0,space=0)
             label  = widget_label(row,value=(['z','z1'])[z2_yn],xsize=self.xsize[1],/align_center)
             trajZ1 = dm_widget_droplist(row,value=self.traj_type,xsize=self.xsize[2],uname='trajZ1')
             label  = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             trajZ1from = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajZ1from')
             label  = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             trajZ1to = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajZ1to')
          endif 
          if z2_yn then begin
             row    = widget_base(left,/row,ypad=0,xpad=0,space=0)
             label  = widget_label(row,value='z2',xsize=self.xsize[1]-1,/align_center)
             trajZ2 = dm_widget_droplist(row,value=self.traj_type,xsize=self.xsize[2],uname='trajZ2')
             label  = widget_label(row,value='from:',xsize=self.xsize[1],/align_center)
             trajZ2from = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajZ2from')
             label  = widget_label(row,value='to:',xsize=self.xsize[1],/align_center)
             trajZ2to = self->my_widget_text(row,value='',scr_xsize=xsize,uname='trajZ2to')
          endif
          geom     = widget_info(left,/geom)
          right    = widget_base(trajTab,/col,/align_bottom,xpad=0,xsize=self.xsize[0]-geom.xsize-8,ypad=0)
          plotTra  = widget_button(right,/pushbutton_events,value='Plot Trajectories',uname='plotTra',sensitive=0)
          oplotTra = widget_button(right,/pushbutton_events,value='Oplot Trajectories',uname='oplotTra',sensitive=0)
          if float(!version.release) ge 5.6 then $     ;tips for some widget buttons
             widget_control,oplotTra,tooltip='Overplot the trajectories in the current plot window.'
          self.xsize[1] = oldxsize1
       endelse
       if float(!version.release) ge 5.6 then begin  ;tips for some widget buttons
          widget_control,dispFitSlice,tooltip='Fit the grouped slice data in PAN(Peak Analysis).'
          widget_control,dispPlotSlice,tooltip='Contour plot the slice.'
          widget_control,dispSurfSlice,tooltip='Surface plot the slice.'
          widget_control,cutOplot,tooltip='Overplot the cut in the current plot window.'
          widget_control,cutFit,tooltip='Fit the cut data in PAN(Peak Analysis).'
       endif
    endif
    ;set all the values
    if (~ ptr_valid(self.dataPtr)) then begin
       self->set_display_button    ;include some menu buttons and all display buttons
       trajs = ['trajX','trajY']
       if z1_yn then trajs = [trajs,'trajZ1']
       if z2_yn then trajs = [trajs,'trajZ2']
       case self.samp_typ of
         0:begin  ;powder
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutX'),set_value=[self.view_u1lab[0],self.view_u2lab[0]],select=self.cut_x[self.samp_typ]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutY'),set_value=['Intensity',self.view_type[0:3]],select=self.cut_y[self.samp_typ]
           self->my_widget_control,trajs,self.dispBas,set_value=[self.view_type[0:3],'None'] & nvalue = 5
           end
         1:begin  ;single crystal
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispPerp'),set_value=self.traj_type[[0,1,3]],select=self.disp_perp
           self->my_widget_control,'dispPfrom',self.dispBas,set_value=dm_to_string(self.disp_pran[0])
           self->my_widget_control,'dispPto',  self.dispBas,set_value=dm_to_string(self.disp_pran[1])
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutX'),set_value=self.traj_type[[0,1,3]],select=self.cut_x[self.samp_typ]
           if self.extravaxis_yn then begin
              index = [0,1,2,3] 
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='volsel_T0'),set_value=self.traj_type[index],select=self.volsel_t1
           endif else index = [0,1,3]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutY'),set_value=['Intensity',self.traj_type[index]],select=self.cut_y[self.samp_typ]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselX'),set_value=self.traj_type[index],select=self.volsel_x
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselY'),set_value=self.traj_type[index],select=self.volsel_y
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselZ'),set_value=self.traj_type[index],select=self.volsel_z
           if self.extravaxis_yn then index = [0,1,2,3,4] else index = [0,1,3,4] & nvalue = n_elements(index)
           self->my_widget_control,trajs,self.dispBas,set_value=self.traj_type[index]
           end
         2:begin  ;diffuse scattering single E
           if self.extravaxis_yn then begin
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispPerp'),set_value=self.traj_type[0:2],select=self.disp_perp
              self->my_widget_control,'dispPfrom',self.dispBas,set_value=dm_to_string(self.disp_pran[0])
              self->my_widget_control,'dispPto',  self.dispBas,set_value=dm_to_string(self.disp_pran[1])
              index = [0,1,2]
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselX'),set_value=self.traj_type[index],select=self.volsel_x
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselY'),set_value=self.traj_type[index],select=self.volsel_y
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselZ'),set_value=self.traj_type[index],select=self.volsel_z
           endif else $
              index = [0,1]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutX'),set_value=self.traj_type[index],select=self.cut_x[self.samp_typ]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutY'),set_value=['Intensity',self.traj_type[index]],select=self.cut_y[self.samp_typ]
           if self.extravaxis_yn then index = [0,1,2,4] else index = [0,1,4] & nvalue = n_elements(index)
           self->my_widget_control,trajs,self.dispBas,set_value=self.traj_type[index]
           end
         3:begin  ;diffuse scattering multiple E
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispPerp'),set_value=self.traj_type[[0,1,3]],select=self.disp_perp
           self->my_widget_control,'dispPfrom',self.dispBas,set_value=dm_to_string(self.disp_pran[0])
           self->my_widget_control,'dispPto',  self.dispBas,set_value=dm_to_string(self.disp_pran[1])
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutX'),set_value=self.traj_type[[0,1,3]],select=self.cut_x[self.samp_typ]
           if self.extravaxis_yn then begin
              index = [0,1,2,3] 
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='volsel_T0'),set_value=self.traj_type[index],select=self.volsel_t1
           endif else index = [0,1,3]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutY'),set_value=['Intensity',self.traj_type[index]],select=self.cut_y[self.samp_typ]
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselX'),set_value=self.traj_type[index],select=self.volsel_x
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselY'),set_value=self.traj_type[index],select=self.volsel_y
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselZ'),set_value=self.traj_type[index],select=self.volsel_z
           if self.extravaxis_yn then index = [0,1,2,3,4] else index = [0,1,3,4] & nvalue = n_elements(index)
           self->my_widget_control,trajs,self.dispBas,set_value=self.traj_type[index]
           end
         else:
       endcase
       ;intelligent cut step
       if ptr_valid(self.dataStrPtr) then emax = max((*self.dataStrPtr).e,min=emin)
       notE = [([self.view_u1typ,self.view_u2typ])[self.cut_x[0]] ne 0,$
               ([0,0,1])[self.cut_x[1]] ne 1,1b,([0,0,1])[self.cut_x[3]] ne 1]
       if notE[self.samp_typ] then begin
          if (~ finite(self.cut_step)) then self.cut_step=0.01
       endif else begin
          if finite(self.cut_step) and ptr_valid(self.dataStrPtr) and (self.ftype eq 0) then begin
             nen = ((finite(self.cut_ran[1])?(self.cut_ran[1]):(emax))-(finite(self.cut_ran[0])?(self.cut_ran[0]):(emin)))/(1e-7>(self.cut_step))
             tmp = where((*self.dataStrPtr).e le (finite(self.cut_ran[1])?(self.cut_ran[1]):(emax)) $
                     and (*self.dataStrPtr).e ge (finite(self.cut_ran[0])?(self.cut_ran[0]):(emin)), nen1)
             if nen gt nen1 then self.cut_step = !values.f_nan
          endif
       endelse
       viewlbs = [self.view_u1lab[(self.samp_typ ge 1)],self.view_u2lab[(self.samp_typ ge 1)],self.view_u3lab]
       if self.samp_typ eq 1 or self.samp_typ eq 3 then viewlbs[2] = 'E'
       head    = ['horizontal range ','vertical range ','thickness range ']
       names0  = ['h_label','v_label','t_label']
       names1  = ['t1_label','t2_label','t3_label']
       if (self.samp_typ eq 1 or self.samp_typ eq 3 ) and self.extravaxis_yn then $
          n_lab = 3 $
       else $
          n_lab = 2
       for i=0,n_lab-1 do begin
           if self.extravaxis_yn then $
              labels = head[i]+[viewlbs[i],viewlbs[(self.disp_perp+i+1) mod 3],$
                       viewlbs[(self.disp_perp+i+1) mod 3],viewlbs[(self.disp_perp+i+1) mod 3]] $
           else $
              labels = head[i]+[viewlbs[i],viewlbs[(self.disp_perp+i+1) mod 3],viewlbs[i],$
                       viewlbs[(self.disp_perp+i+1) mod 3]]
           if i eq 2 then labels[*] = head[2]+self.view_u3lab
           self->my_widget_control,names0[i],self.dispBas,set_value=labels[self.samp_typ]
           ;intelligent slice step
           notE = [([self.view_u1typ,self.view_u2typ,1])[i] ne 0,$
                   ([0,0,1])[(self.disp_perp+i+1) mod 3] ne 1,$
                   1b,([0,0,1])[(self.disp_perp+i+1) mod 3] ne 1]
           if notE[self.samp_typ] then begin
              if (i eq 0) and (~ finite(self.disp_hstep)) then self.disp_hstep=0.035
              if (i eq 1) and (~ finite(self.disp_vstep)) then self.disp_vstep=0.035
           endif else if ptr_valid(self.dataStrPtr) and (self.ftype eq 0) then begin
              if (i eq 0) and (finite(self.disp_hstep)) and ptr_valid(self.dataStrPtr) then begin
                 nen = ((finite(self.disp_hran[1])?(self.disp_hran[1]):(emax))-(finite(self.disp_hran[0])?(self.disp_hran[0]):(emin)))/(1e-7>(self.disp_hstep))
                 tmp = where((*self.dataStrPtr).e le (finite(self.disp_hran[1])?(self.disp_hran[1]):(emax)) $
                         and (*self.dataStrPtr).e ge (finite(self.disp_hran[0])?(self.disp_hran[0]):(emin)), nen1)
                 if nen gt nen1 then self.disp_hstep = !values.f_nan
              endif
              if (i eq 1) and (finite(self.disp_vstep)) and ptr_valid(self.dataStrPtr) then begin
                 nen = ((finite(self.disp_vran[1])?(self.disp_vran[1]):(emax))-(finite(self.disp_vran[0])?(self.disp_vran[0]):(emin)))/(1e-7>(self.disp_vstep))
                 tmp = where((*self.dataStrPtr).e le (finite(self.disp_vran[1])?(self.disp_vran[1]):(emax)) $
                         and (*self.dataStrPtr).e ge (finite(self.disp_vran[0])?(self.disp_vran[0]):(emin)), nen1)
                 if nen gt nen1 then self.disp_vstep = !values.f_nan
              endif
           endif
           labels = 'thickness range '+[viewlbs[(self.cut_x[0]+i+1) mod 2],viewlbs[(self.cut_x[1]+i+1) mod 3],$
                    viewlbs[(self.cut_x[2]+i+1) mod (2+self.extravaxis_yn)],viewlbs[(self.cut_x[3]+i+1) mod 3]]
           if i eq 2 then labels[*] = 'thickness range '+self.view_u3lab
           self->my_widget_control,names1[i],self.dispBas,set_value=labels[self.samp_typ]
       endfor
       self->my_widget_control,'cutT1from', self.dispBas,set_value=dm_to_string(self.cut_t1ran[0])
       self->my_widget_control,'cutT1to',   self.dispBas,set_value=dm_to_string(self.cut_t1ran[1])
       if (self.samp_typ eq 1) or ((self.samp_typ eq 2) and self.extravaxis_yn) or (self.samp_typ eq 3) then begin
          ;intelligent volume step
          if self.extravaxis_yn then index = [0,0,0,1] else index = [0,0,1]
          isE = [0b,index[self.volsel_x] eq 1 ,0b,index[self.volsel_x] eq 1]
          if (~ finite(self.vol_xstep)) then begin
             self.vol_xstep = 0.05
             if isE[self.samp_typ] and ptr_valid(self.dataStrPtr) then begin
                self.vol_xstep = ((finite(self.vol_xran[1])?(self.vol_xran[1]):(emax))-(finite(self.vol_xran[0])?(self.vol_xran[0]):(emin)))/150.
                self.vol_xstep = 0.005>(fix(self.vol_xstep*100.+0.5,type=3)/100.)
             endif
          endif
          isE = [0b,index[self.volsel_y] eq 1 ,0b,index[self.volsel_y] eq 1]
          if (~ finite(self.vol_ystep)) then begin
             self.vol_ystep = 0.05
             if isE[self.samp_typ] and ptr_valid(self.dataStrPtr) then begin
                self.vol_ystep = ((finite(self.vol_yran[1])?(self.vol_yran[1]):(emax))-(finite(self.vol_yran[0])?(self.vol_yran[0]):(emin)))/150.
                self.vol_ystep = 0.005>(fix(self.vol_ystep*100.+0.5,type=3)/100.)
             endif
          endif
          isE = [0b,index[self.volsel_z] eq 1 ,0b,index[self.volsel_z] eq 1]
          if (~ finite(self.vol_zstep)) then begin
             self.vol_zstep = 0.05
             if isE[self.samp_typ] and ptr_valid(self.dataStrPtr) then begin
                self.vol_zstep = ((finite(self.vol_zran[1])?(self.vol_zran[1]):(emax))-(finite(self.vol_zran[0])?(self.vol_zran[0]):(emin)))/150.
                self.vol_zstep = 0.005>(fix(self.vol_zstep*100.+0.5,type=3)/100.)
             endif
          endif
          self->my_widget_control,'volxfrom', self.dispBas,set_value=dm_to_string(self.vol_xran[0])
          self->my_widget_control,'volxto',   self.dispBas,set_value=dm_to_string(self.vol_xran[1])
          self->my_widget_control,'volxstep', self.dispBas,set_value=dm_to_string(self.vol_xstep)
          self->my_widget_control,'volyfrom', self.dispBas,set_value=dm_to_string(self.vol_yran[0])
          self->my_widget_control,'volyto',   self.dispBas,set_value=dm_to_string(self.vol_yran[1])
          self->my_widget_control,'volystep', self.dispBas,set_value=dm_to_string(self.vol_ystep)
          self->my_widget_control,'volzfrom', self.dispBas,set_value=dm_to_string(self.vol_zran[0])
          self->my_widget_control,'volzto',   self.dispBas,set_value=dm_to_string(self.vol_zran[1])
          self->my_widget_control,'volzstep', self.dispBas,set_value=dm_to_string(self.vol_zstep)
          self->my_widget_control,'volIfrom', self.dispBas,set_value=dm_to_string(self.vol_iran[0])
          self->my_widget_control,'volIto',   self.dispBas,set_value=dm_to_string(self.vol_iran[1])
          self->my_widget_control,'volSmooth',self.dispBas,set_value=dm_to_string(self.vol_smooth,/int)
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volColor'), select=self.vol_color
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volInterp'),select=self.vol_interp
          self->my_widget_control,'cutT2from',self.dispBas,set_value=dm_to_string(self.cut_t2ran[0])
          self->my_widget_control,'cutT2to',  self.dispBas,set_value=dm_to_string(self.cut_t2ran[1])
       endif
       if (self.samp_typ eq 1 or self.samp_typ eq 3) and self.extravaxis_yn then begin
          self->my_widget_control,'dispTfrom',self.dispBas,set_value=dm_to_string(self.disp_tran[0])
          self->my_widget_control,'dispTto',  self.dispBas,set_value=dm_to_string(self.disp_tran[1])
          self->my_widget_control,'cutT3from',self.dispBas,set_value=dm_to_string(self.cut_t3ran[0])
          self->my_widget_control,'cutT3to',  self.dispBas,set_value=dm_to_string(self.cut_t3ran[1])
          self->my_widget_control,'volselfrom_T0',self.dispBas,set_value=dm_to_string(self.volsel_tft1[0])
          self->my_widget_control,'volselto_T0',  self.dispBas,set_value=dm_to_string(self.volsel_tft1[1])
       endif
       if self.traj_x[self.samp_typ]  ge nvalue then self.traj_x[self.samp_typ]  = (self.samp_typ eq 0)
       if self.traj_y[self.samp_typ]  ge nvalue then self.traj_y[self.samp_typ]  = 1-(self.samp_typ eq 0)
       if (self.traj_z1[self.samp_typ] ge nvalue) or (self.traj_z1[self.samp_typ] lt 2) then self.traj_z1[self.samp_typ] = nvalue-1
       if (self.traj_z2[self.samp_typ] ge nvalue) or (self.traj_z2[self.samp_typ] lt 2) then self.traj_z2[self.samp_typ] = nvalue-1
       if self.traj_x[self.samp_typ] eq self.traj_y[self.samp_typ] then begin
          self.traj_x[self.samp_typ] = (self.samp_typ eq 0)
          self.traj_y[self.samp_typ] = 1-(self.samp_typ eq 0)
       endif
       if (self.traj_z1[self.samp_typ] eq self.traj_z2[self.samp_typ]) and (self.traj_z1[self.samp_typ] ne nvalue-1) then begin
          self.traj_z2[self.samp_typ] = (nvalue-1)<(self.traj_z2[self.samp_typ]+1) 
       endif
       dm_set_droplist,widget_info(self.dispBas,find_by_uname='trajX'), select=self.traj_x[self.samp_typ]
       dm_set_droplist,widget_info(self.dispBas,find_by_uname='trajY'), select=self.traj_y[self.samp_typ]
       self->my_widget_control,'trajXfrom', self.dispBas,set_value=dm_to_string(self.traj_xran[0])
       self->my_widget_control,'trajXto',   self.dispBas,set_value=dm_to_string(self.traj_xran[1])
       self->my_widget_control,'trajYfrom', self.dispBas,set_value=dm_to_string(self.traj_yran[0])
       self->my_widget_control,'trajYto',   self.dispBas,set_value=dm_to_string(self.traj_yran[1])
       if z1_yn then begin
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='trajZ1'),select=self.traj_z1[self.samp_typ]
          self->my_widget_control,'trajZ1from',self.dispBas,set_value=dm_to_string(self.traj_z1ran[0])
          self->my_widget_control,'trajZ1to',  self.dispBas,set_value=dm_to_string(self.traj_z1ran[1])
       endif
       if z2_yn then begin
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='trajZ2'),select=self.traj_z2[self.samp_typ]
          self->my_widget_control,'trajZ2from',self.dispBas,set_value=dm_to_string(self.traj_z2ran[0])
          self->my_widget_control,'trajZ2to',  self.dispBas,set_value=dm_to_string(self.traj_z2ran[1])
       endif  
    endif else begin
       nitem    = n_elements((*self.dataPtr).label)
       sel_list = (*self.dataPtr).label[2:nitem-1]
       ;volume 
       if nitem ge 5 then begin 
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselX'),set_value=sel_list,select=self.volsel_x
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselY'),set_value=sel_list,select=self.volsel_y
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volselZ'),set_value=sel_list,select=self.volsel_z
          self->my_widget_control,'volxfrom', self.dispBas,set_value=dm_to_string(self.vol_xran[0])
          self->my_widget_control,'volxto',   self.dispBas,set_value=dm_to_string(self.vol_xran[1])
          self->my_widget_control,'volxstep', self.dispBas,set_value=dm_to_string(self.vol_xstep)
          self->my_widget_control,'volyfrom', self.dispBas,set_value=dm_to_string(self.vol_yran[0])
          self->my_widget_control,'volyto',   self.dispBas,set_value=dm_to_string(self.vol_yran[1])
          self->my_widget_control,'volystep', self.dispBas,set_value=dm_to_string(self.vol_ystep)
          self->my_widget_control,'volzfrom', self.dispBas,set_value=dm_to_string(self.vol_zran[0])
          self->my_widget_control,'volzto',   self.dispBas,set_value=dm_to_string(self.vol_zran[1])
          self->my_widget_control,'volzstep', self.dispBas,set_value=dm_to_string(self.vol_zstep)
          for i=0,nitem-6 do begin   
              dm_set_droplist,widget_info(self.dispBas,find_by_uname='volsel_T'+dm_to_string(i)),set_value=sel_list,select=(*self.volsel_t)[i]
              self->my_widget_control,'volselfrom_T'+dm_to_string(i),self.dispBas,set_value=dm_to_string((*self.volsel_tft)[i,0])
              self->my_widget_control,'volselto_T'+dm_to_string(i),self.dispBas,set_value=dm_to_string((*self.volsel_tft)[i,1])
          endfor
          self->my_widget_control,'volIfrom', self.dispBas,set_value=dm_to_string(self.vol_iran[0])
          self->my_widget_control,'volIto',   self.dispBas,set_value=dm_to_string(self.vol_iran[1])
          self->my_widget_control,'volSmooth',self.dispBas,set_value=dm_to_string(self.vol_smooth,/int)
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volColor'), select=self.vol_color
          dm_set_droplist,widget_info(self.dispBas,find_by_uname='volInterp'),select=self.vol_interp
       endif
       ;slice
       dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispselH'),set_value=sel_list,select=self.dispsel_h
       dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispselV'),set_value=sel_list,select=self.dispsel_v
       for i=0,nitem-5 do begin   
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispsel_T'+dm_to_string(i)),set_value=sel_list,select=(*self.dispsel_t)[i]
           self->my_widget_control,'dispselfrom_T'+dm_to_string(i),self.dispBas,set_value=dm_to_string((*self.dispsel_tft)[i,0])
           self->my_widget_control,'dispselto_T'+dm_to_string(i),self.dispBas,set_value=dm_to_string((*self.dispsel_tft)[i,1])
       endfor
       ;cut
       dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutX'),set_value=sel_list,select=self.cut_x[0]
       dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutY'),set_value=['Intensity',sel_list],select=self.cut_y[0]
       for i=0,nitem-4 do begin   
           dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutsel_T'+dm_to_string(i)),set_value=sel_list,select=(*self.cutsel_t)[i]
           self->my_widget_control,'cutselfrom_T'+dm_to_string(i),self.dispBas,set_value=dm_to_string((*self.cutsel_tft)[i,0])
           self->my_widget_control,'cutselto_T'+dm_to_string(i),self.dispBas,set_value=dm_to_string((*self.cutsel_tft)[i,1])
       endfor
    endelse
    self->my_widget_control,'dispHfrom', self.dispBas,set_value=dm_to_string(self.disp_hran[0])
    self->my_widget_control,'dispHto',   self.dispBas,set_value=dm_to_string(self.disp_hran[1])
    self->my_widget_control,'dispHstep', self.dispBas,set_value=dm_to_string(self.disp_hstep)
    self->my_widget_control,'dispVfrom', self.dispBas,set_value=dm_to_string(self.disp_vran[0])
    self->my_widget_control,'dispVto',   self.dispBas,set_value=dm_to_string(self.disp_vran[1])
    self->my_widget_control,'dispVstep', self.dispBas,set_value=dm_to_string(self.disp_vstep)
    self->my_widget_control,'dispIfrom', self.dispBas,set_value=dm_to_string(self.disp_iran[0])
    self->my_widget_control,'dispIto',   self.dispBas,set_value=dm_to_string(self.disp_iran[1])
    dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispColor'),  select=self.disp_color
    dm_set_droplist,widget_info(self.dispBas,find_by_uname='dispShading'),select=self.disp_shading
    dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutColor'),   select=self.cut_color
    dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutSymbol'),  select=self.cut_symbol
    dm_set_droplist,widget_info(self.dispBas,find_by_uname='cutLine'),    select=self.cut_line
    self->my_widget_control,'dispSmooth',self.dispBas,set_value=dm_to_string(self.disp_smooth,/int)
    self->my_widget_control,'cutFrom',   self.dispBas,set_value=dm_to_string(self.cut_ran[0])
    self->my_widget_control,'cutTo',     self.dispBas,set_value=dm_to_string(self.cut_ran[1])
    self->my_widget_control,'cutStep',   self.dispBas,set_value=dm_to_string(self.cut_step)
    self->my_widget_control,'cutYfrom',  self.dispBas,set_value=dm_to_string(self.cut_yran[0])
    self->my_widget_control,'cutYto',    self.dispBas,set_value=dm_to_string(self.cut_yran[1])
end

;reset menu item sensitivity and ticks accordingly
;keywords:
;   geomsensitive: if set to 1: instrument geometry menus will be sensitive
;                  if set to 0: instrument geometry menus will be insensitive
;                  if not set, no action
;   backgroundmenuonly: if set, only reset the background menu
pro dcs_mslice::reset_menu,geomsensitive=geomsensitive,backgroundmenuonly=backgroundmenuonly
    ;background menu and parameter menu
    emptBut  = widget_info(self.tlb,find_by_uname='emptBut')
    emptBut2 = widget_info(self.tlb,find_by_uname='emptBut2')
    if ((self.extraempcan_yn eq 1) and (emptBut2 eq 0)) or ((self.extraempcan_yn eq 0) and (emptBut2 ne 0)) then begin
       ;rebuild the background menu and parameter menu
       bkgrmenu = widget_info(self.tlb,find_by_uname='bkgrMenu')
       parmmenu = widget_info(self.tlb,find_by_uname='parmMenu')
       self->my_widget_control,['emptBut2','bgflBut','bgtcBut','bgdkBut','bgovBut','bgclBut','vandBut','vdclBut'],bkgrmenu,/destroy
       self->my_widget_control,['emfcBut','knpmBut','ddirBut','wdirBut'],parmmenu,/destroy
       if self.extraempcan_yn then begin
          self->my_widget_control,'emptBut',bkgrmenu,set_value='Load Empty Can File(s) 1'
          void  = widget_button(bkgrmenu,value='Load Empty Can File(s) 2',uname='emptBut2')
          void  = widget_button(parmmenu,value='Change 2-Set Empty Can Factors...',uname='emfcBut',sensitive=0)
          void  = widget_button(parmmenu,value='Change Smoothing Kernel...',uname='knpmBut')
          void  = widget_button(parmmenu,value='Set Data Directory...',uname='ddirBut',/separator)
          void  = widget_button(parmmenu,value='Set Working Directory...',uname='wdirBut')
       endif else begin
          self->my_widget_control,'emptBut',bkgrmenu,set_value='Load Empty Can File(s)'
          void  = widget_button(parmmenu,value='Change Smoothing Kernel...',uname='knpmBut')
          void  = widget_button(parmmenu,value='Set Data Directory...',uname='ddirBut',/separator)
          void  = widget_button(parmmenu,value='Set Working Directory...',uname='wdirBut')
       endelse
       void     = widget_button(bkgrmenu,value='Load Detector Background File(s)',uname='bgflBut')
       void     = widget_button(bkgrmenu,value='Load Time Channel Background File(s)',uname='bgtcBut')
       void     = widget_button(bkgrmenu,value='Load Dark Count File(s)',uname='bgdkBut')
       void     = widget_button(bkgrmenu,value='Set Overall Background Rate...',uname='bgovBut')
       void     = widget_button(bkgrmenu,value='Clear Background',uname='bgclBut')
       void     = widget_button(bkgrmenu,value='Load Vanadium File(s)',uname='vandBut',/separator)
       void     = widget_button(bkgrmenu,value='Clear Vanadium',uname='vdclBut')
    endif
    if self.extraempcan_yn then begin
       self->my_widget_control,'emptBut2',bkgrmenu,sensitive=ptr_valid(self.bgdata)
       self->my_widget_control,'emfcBut',bkgrmenu,sensitive=ptr_valid(self.bgdata2)
    endif
    self->my_widget_control,'vdclBut',sensitive=ptr_valid(self.eff)
    self->my_widget_control,'bgclBut',sensitive=(total(ptr_valid([self.bgtchnPtr,self.bgratePtr,self.bgdetsPtr,self.bgdata])) ne 0)
    if keyword_set(backgroundmenuonly) then return
    ;file menu: add a click to the current file type
    ftypemenu  = widget_info(self.tlb,find_by_uname='ftypeMenu')
    checked    = widget_info(ftypemenu,find_by_uname='ftype_'+dm_to_string(self.ftype))
    for i=0,n_elements(*self.ftypename)-1 do begin
        if i ne self.ftype then begin
           tmp = widget_info(ftypemenu,find_by_uname='ftype_'+dm_to_string(i))
           unchecked = (n_elements(unchecked) eq 0)?tmp:[unchecked,tmp]
        endif
    endfor
    dm_toggle_menubut,check=checked,uncheck=unchecked
    self->my_widget_control,['savesliceMenu','savecutMenu','saveprojMenu'],sensitive=(ptr_valid(self.projStrPtr) and ptr_valid(self.dataStrPtr))
    ;mask menu
    self->my_widget_control,['mkdbMenu'],sensitive=(self.ftype eq 0)
    if self.ftype eq 0 then self->dm_mask_detbanks
    self->my_widget_control,['mkclBut','mksvBut'],sensitive=ptr_valid(self.mask)
    self->my_widget_control,['mkplBut','mkbpBut'],sensitive=ptr_valid(self.dataStrPtr)
    self->my_widget_control,'mkplangl',sensitive=self.ftype eq 0
    ;option menu
    binSum   = widget_info(self.tlb,find_by_uname='binSum')
    binAvg   = widget_info(self.tlb,find_by_uname='binAvg')
    bincsYes = widget_info(self.tlb,find_by_uname='bincsYes')
    bincsNo  = widget_info(self.tlb,find_by_uname='bincsNo')
    binexYes = widget_info(self.tlb,find_by_uname='binexYes')
    binexNo  = widget_info(self.tlb,find_by_uname='binexNo')
    monSum   = widget_info(self.tlb,find_by_uname='monSum')
    monFit   = widget_info(self.tlb,find_by_uname='monFit')
    monFC0   = widget_info(self.tlb,find_by_uname='monFC0')
    monBM1   = widget_info(self.tlb,find_by_uname='monBM1')
    monTIME  = widget_info(self.tlb,find_by_uname='monTIME')
    viewOrth = widget_info(self.tlb,find_by_uname='vaxisOrtho')
    viewArbi = widget_info(self.tlb,find_by_uname='vaxisArbi')
    abspNon  = widget_info(self.tlb,find_by_uname='abspNone')
    abspAnn  = widget_info(self.tlb,find_by_uname='abspAnnulus')
    abspCyl  = widget_info(self.tlb,find_by_uname='abspCylinder')
    abspSph  = widget_info(self.tlb,find_by_uname='abspSphere')
    abspFla  = widget_info(self.tlb,find_by_uname='abspFlat')
    epadNone = widget_info(self.tlb,find_by_uname='epadNone')
    epadAuto = widget_info(self.tlb,find_by_uname='epadAuto')
    epadSpec = widget_info(self.tlb,find_by_uname='epadSpec')
    bgBlack  = widget_info(self.tlb,find_by_uname='pbgBlack')
    bgWhite  = widget_info(self.tlb,find_by_uname='pbgWhite')
    
    dm_set_button,widget_info(self.tlb,find_by_uname='epadCheckfirst'),self.eadjust_checkfirst,$
       onstring='Check the First File Only',offstring='Check All Files'
    self->my_widget_control,'epadCheckfirst',sensitive=(self.eadjust ne 0)
    self->set_instrgeom,geom=self.instrgeom,sensitive=geomsensitive
    ntype = n_elements(self.intn_typ_str)
    if self.intn_typ ge ntype or self.intn_typ lt 0 then self.intn_typ=0
    inttypid = lonarr(ntype)
    for i=0,ntype-1 do $
        inttypid[i] = widget_info(self.tlb,find_by_uname='intnTyp'+dm_to_string(i,/int))
    dm_toggle_menubut,check=inttypid[self.intn_typ],uncheck=inttypid[where(indgen(ntype) ne self.intn_typ)]
    dm_set_button,widget_info(self.tlb,find_by_uname='intnEint'),self.eint_yn,$
       onstring='E-Integrated Intensity Enabled',offstring='E-Integrated Intensity Disabled'
    dm_toggle_menubut,check=([binAvg,binSum])[self.bin_avgsum],uncheck=([binAvg,binSum])[1-self.bin_avgsum]
    dm_toggle_menubut,check=([bincsNo,bincsYes])[self.bin_conststep],uncheck=([bincsNo,bincsYes])[1-self.bin_conststep]
    dm_toggle_menubut,check=([binexNo,binexYes])[self.bin_extlib],uncheck=([binexNo,binexYes])[1-self.bin_extlib]
    dm_toggle_menubut,check=([monSum,monFit])[self.mon_sumfit],uncheck=([monSum,monFit])[1-self.mon_sumfit]
    case self.dcs_mon of
            0:   dm_toggle_menubut,check=monFC0, uncheck=[monBM1,monTIME]
            1:   dm_toggle_menubut,check=monBM1, uncheck=[monFC0,monTIME]
            2:   dm_toggle_menubut,check=monTIME,uncheck=[monFC0,monBM1]
            else:
    endcase
    if ptr_valid(self.absp_info) then begin
       case (*(self.absp_info))[0] of
            0:   dm_toggle_menubut,check=abspCyl,uncheck=[abspNon,abspAnn,abspSph,abspFla]
            1:   dm_toggle_menubut,check=abspAnn,uncheck=[abspNon,abspCyl,abspSph,abspFla]
            2:   dm_toggle_menubut,check=abspSph,uncheck=[abspNon,abspCyl,abspAnn,abspFla]
            3:   dm_toggle_menubut,check=abspFla,uncheck=[abspNon,abspCyl,abspAnn,abspSph]
            else:dm_toggle_menubut,check=abspNon,uncheck=[abspCyl,abspAnn,abspSph,abspFla]
       endcase
    endif else   dm_toggle_menubut,check=abspNon,uncheck=[abspCyl,abspAnn,abspSph,abspFla]
    dm_toggle_menubut,check=([viewOrth,viewArbi])[self.view_ortharbi],uncheck=([viewOrth,viewArbi])[1-self.view_ortharbi]
    self->change_powderline,self.powd_typ,/menuonly
    case self.eadjust of 
            0:   dm_toggle_menubut,check=epadNone,uncheck=[epadAuto,epadSpec]
            1:   dm_toggle_menubut,check=epadAuto,uncheck=[epadNone,epadSpec]
            2:   dm_toggle_menubut,check=epadSpec,uncheck=[epadAuto,epadNone]
            else:dm_toggle_menubut,check=epadNone,uncheck=[epadAuto,epadSpec]
    endcase
    dm_toggle_menubut,check=([bgBlack,bgWhite])[self.pbgcol_bw],uncheck=([bgBlack,bgWhite])[1-self.pbgcol_bw]
    self->my_widget_control,'userFunction',set_button=(strlen(self.userf_str) ne 0)
    dm_set_button,widget_info(self.tlb,find_by_uname='extraVaxis'),self.extravaxis_yn,$
       onstring='Allow Extra Viewing Axis',$
       offstring='Disallow Extra Viewing Axis'
    dm_set_button,widget_info(self.tlb,find_by_uname='foldRight'),self.foldright,$
       onstring='Fold to the Plus Side',$
       offstring='Fold to the Side with Larger Range'
    dm_set_button,widget_info(self.tlb,find_by_uname='darkCorr'),self.dcor_yn,$
       onstring='Dark Angle Correction Enabled',$
       offstring='Dark Angle Correction Disabled'
    dm_set_button,widget_info(self.tlb,find_by_uname='edepEff'),self.edep_yn,$
       onstring='E-Dependent Detector Efficiency Correction Enabled',$
       offstring='E-Dependent Detector Efficiency Correction Disabled'
    dm_set_button,widget_info(self.tlb,find_by_uname='edepEff'),self.edep_yn,$
       onstring='E-Dependent Detector Efficiency Correction Enabled',$
       offstring='E-Dependent Detector Efficiency Correction Disabled'
    dm_set_button,widget_info(self.tlb,find_by_uname='errBar'),1-self.errbar,$
       onstring='Error Bar Disabled',$
       offstring='Error Bar Enabled'
    self->my_widget_control,'moncMenu',sensitive=(self.ftype eq 0)
    self->change_samptyp,self.samp_typ
end

;reset projection and display window
pro dcs_mslice::reset_pdt
    WIDGET_CONTROL,/HOURGLASS

    self->reset_droplist        ;adjust droplist values if necessary

    widget_control,self.tlb,update=0
    geom1_old = widget_info(self.projBas,/geometry)
    geom2_old = widget_info(self.dispBas,/geometry)
    widget_control,self.dispBas,/destroy
    widget_control,self.projBas,/destroy
    self->reset_projection
    self->reset_display
    geom1  = widget_info(self.projBas,/geometry)
    geom2  = widget_info(self.dispBas,/geometry)
    if float(!version.release) ge 5.5 then begin
       self.geom[1] = self.geom[1]+geom1.ysize+geom2.ysize-(geom1_old.ysize+geom2_old.ysize)
       widget_control,self.tlb,ysize=self.geom[1]
    endif else begin
       self.geom[1] = self.geom[1]+geom1.scr_ysize+geom2.scr_ysize-(geom1_old.scr_ysize+geom2_old.scr_ysize)
       widget_control,self.tlb,scr_ysize=self.geom[1]
    endelse
    widget_control,self.tlb,/update
    widget_control,self.projBas,/update,/map
    widget_control,self.dispBas,/update,/map
end

;save cuts in files
;keyword:
;   dave:   if set, save cut as dave files
;   dPtr:   if present, will pass the cut data as a dave pointer
;   xye:    if set, save cut as x, y, yerr 3-column file
pro dcs_mslice::save_cut,dave=dave,dPtr=dPtr,xye=xye
    if (~ self->pd_valid()) then return
    ext = ''
    if keyword_set(dave) then ext = 'dave'
    if keyword_set(xye)  then ext = 'xye'
    if (~ arg_present(dPtr)) then begin
       if strlen(ext) eq 0 then return
       path = self.dirs[1]
       file = dm_choose_file(ext,/write,dialog_parent=self.tlb,title='Save cut as '+ext+' file.',path=path)
       file = file[0]
       if strlen(file) eq 0 then return
       if file_test(path,/directory) then self.dirs[1]=path
       openw,unit,file,/get_lun,error=openerr
       if openerr ne 0 then begin
          ok = dialog_message("Can't write in "+file,/error,dialog_parent=self.tlb)
          return
       endif
    endif

    ;select data
    self->dm_choose_cutdata,xdat=xdat,ydat=ydat,yerr=yerr,xtit=xtit,ytit=ytit
    if (~ self.errbar) and (n_elements(yerr) ne 0) then tmp=temporary(yerr)   ;destroy yerr

    if keyword_set(xye) then begin
       for i=0L,n_elements(xdat)-1 do begin
           if n_elements(yerr) ne 0 then printf,unit,xdat[i],ydat[i],yerr[i] $
           else printf,unit,xdat[i],ydat[i]
       endfor
    endif

    if keyword_set(dave) or arg_present(dPtr) then begin
       if n_elements(yerr) ne 0 then $
          ret = create_dave_pointer(    $
                dPtr,                   $
                qty     = ydat,         $
                qtlabel = ytit,         $
                err     = yerr,         $
                xvals   = xdat,         $
                xtype   = 'POINT',      $
                xlabel  = xtit          $
          ) $
       else $
          ret = create_dave_pointer(    $
                dPtr,                   $
                qty     = ydat,         $
                qtlabel = ytit,         $
                xvals   = xdat,         $
                xtype   = 'POINT',      $
                xlabel  = xtit          $
          )
       if n_elements(unit) ne 0 then begin
          davePtr = dPtr
          save,davePtr,filename=file
          heap_free,dPtr
       endif
    endif

    if n_elements(unit) ne 0 then free_lun,unit
end

;save slice in files
;keywords:
;   dave:   if set, save slice as grouped DAVE file, for PAN fitting
;   dPtr:   if present, will pass the slice data as a dave pointer
;   iexy:   if set, save slice as Int, Err, X, Y 4-column file
;   spe:    if set, save slice as spe file
;   xyie:   if set, save slice as X, Y, Int, Err matrix file
pro dcs_mslice::save_slice,dave=dave,dPtr=dPtr,iexy=iexy,spe=spe,xyie=xyie
    if (~ self->pd_valid()) then return
    ext = ''
    if keyword_set(dave) then ext = 'dave'
    if keyword_set(iexy) then ext = 'iexy'
    if keyword_set(spe)  then ext = 'spe'
    if keyword_set(xyie) then ext = 'xyie'

    dcs_filler = -1e20
    if finite(self.mask_filler) then dcs_filler = self.mask_filler 
    
    if (~ arg_present(dPtr)) then begin
       if strlen(ext) eq 0 then return
       path = self.dirs[1]
       file = dm_choose_file(ext,/write,dialog_parent=self.tlb,title='Save slice as '+ext+' file.',path=path)
       file = file[0]
       if strlen(file) eq 0 then return
       if file_test(path,/directory) then self.dirs[1]=path
       if (~ keyword_set(spe)) then begin
          openw,unit,file,/get_lun,error=openerr
          if openerr ne 0 then begin
             ok = dialog_message("Can't write in "+file,/error,dialog_parent=self.tlb)
             return
          endif
       endif
    endif

    ;select data
    tmp = self.disp_smooth
    if self.disp_smooth gt 0 then begin
       info = 'Do you want to '+(['save','send'])[arg_present(dPtr)]+' the smoothed data?'
       ok = dialog_message(info,/question,dialog_parent=self.tlb)
       if strlowcase(ok) eq 'no' then self.disp_smooth = 0
    endif
    self->dm_choose_slicedata,xdat=xdat,ydat=ydat,zdat=zdat,zerr=zerr,xtit=xtit,ytit=ytit,zran=zran,zmin=zmin
    self.disp_smooth = tmp
    index = where(finite(zdat) eq 0, count)
    if count ne 0 then begin
       notok = 1b
       while(notok) do begin
          dcs_filler = dm_dialog_input('filler value:',/float,default=dcs_filler,$
                       info='This value is used for empty grids that contain no data.',dialog_parent=self.tlb)
          notok = ~finite(dcs_filler)
       endwhile
       if self.samp_typ ne -1 then self.mask_filler = dcs_filler
       zdat[index] = dcs_filler
       zerr[index] = -1.0               ;signifies missing data
    endif
    
    widget_control,/hourglass
    
    nx = n_elements(xdat) & ny = n_elements(ydat)

    if keyword_set(iexy) then begin
       for i=0L,nx-1L do begin
           for j=0L,ny-1L do begin
               printf,unit,zdat[i,j],zerr[i,j],xdat[i],ydat[j]
           endfor
       endfor
    endif

    if keyword_set(spe) then begin
       dm_write_spe,file,xdat,ydat,zdat,zerr,xtit=xtit,ytit=ytit,ztit=self.intn_typ_str[self.intn_typ],group_leader=self.tlb
    endif

    if keyword_set(xyie) then begin
       printf,unit,nx,ny
       printf,unit,' '
       for i=0L,nx-1L do printf,unit,xdat[i]
       printf,unit,' '
       for i=0L,ny-1L do printf,unit,ydat[i]
       printf,unit,' '
       printf,unit,zdat,format='('+dm_to_string(nx)+'(g13.6," "))'
       printf,unit,' '
       printf,unit,zerr,format='('+dm_to_string(nx)+'(g13.6," "))'
    endif

    if keyword_set(dave) or arg_present(dPtr) then begin
       if count ne 0 then begin ;creat mask
          mask = bytarr(nx,ny)+1b
          mask[index] = 0b
          mask = bits2bytes(transpose(mask))
       endif

       ;contruct DAVE pointer
       xunit = ''
       yunit = ''
       if self.samp_typ eq 0 then begin
          xtit  = self.view_label[self.view_u1typ]
          xunit = self.view_unit[self.view_u1typ]
          ytit  = self.view_label[self.view_u2typ]
          yunit = self.view_unit[self.view_u2typ]
       endif

       dPtr = dm_define_pointer()
       ptr_free,(*(*dPtr).dataStrPtr).commonStr.histPtr
       (*(*dPtr).dataStrPtr).commonStr.histPtr = $
         ptr_new({qty:transpose(temporary(zdat)),err:transpose(temporary(zerr)),x:ydat,y:xdat})
	     (*(*dPtr).dataStrPtr).commonStr.xunits    = yunit
	     (*(*dPtr).dataStrPtr).commonStr.xlabel    = ytit
	     (*(*dPtr).dataStrPtr).commonStr.yunits    = xunit
	     (*(*dPtr).dataStrPtr).commonStr.ylabel    = xtit
	     (*(*dPtr).dataStrPtr).commonStr.histlabel = self.intn_typ_str[self.intn_typ]

       if count ne 0 then begin ;mask
          ;add PAN_MASK and DCS_FILLER info to specificPtr
          specific = (*(*(*dPtr).dataStrPtr).specificPtr)
          tags = strupcase(tag_names(specific))
          loc = where(tags eq 'PAN_MASK',count)
          if count eq 0 then specific = create_struct('PAN_MASK',ptr_new(mask),specific) $
          else *specific.pan_mask = mask
          loc = where(tags eq 'DCS_FILLER',count)
          if count eq 0 then specific = create_struct('DCS_FILLER',dcs_filler,specific) $
          else specific.dcs_filler = dcs_filler
    	  (*(*(*dPtr).dataStrPtr).specificPtr) = specific
       endif

       if n_elements(unit) ne 0 then begin
          davePtr = dPtr
          save,davePtr,filename=file
          heap_free,dPtr
       endif
    endif

    if n_elements(unit) ne 0 then free_lun,unit
end

;save the projection volume
pro dcs_mslice::save_projection,txt=txt
    if self.samp_typ eq - 1 then begin
       title = 'Save all data as a text file.' 
       nitem = n_elements((*self.dataPtr).label)
       ndat  = n_elements((*self.dataPtr).data[*,0])
    endif else begin
       title = 'Save projection as a text file.'
       nen   = n_elements((*self.projStrPtr).en)
       n2th  = n_elements((*self.projStrPtr).two_theta)
    endelse
    
    path = self.dirs[1]
    file = dm_choose_file(ext,/write,dialog_parent=self.tlb,title=title ,path=path)
    file = file[0]
    if strlen(file) eq 0 then return
    if file_test(path,/directory) then self.dirs[1]=path
    openw,unit,file,/get_lun,error=openerr
    if openerr ne 0 then begin
       ok = dialog_message("Can't write in "+file,/error,dialog_parent=self.tlb)
       return
    endif
    
    widget_control,/hourglass
    
    if !version.os_family eq 'Windows' then sep=string(9b) else sep='     '
    
    case self.samp_typ of
      -1: begin
          tmp = ''
          for i=2,nitem-1 do tmp = tmp+sep+(*self.dataPtr).label[i]
          tmp = tmp+sep+(*self.dataPtr).label[0]+sep+(*self.dataPtr).label[1]
          printf,unit,tmp
          for i=0L,ndat-1L do $
              printf,unit,reform((*self.dataPtr).data[i,[2+indgen(nitem-2),0,1]])
          end
       0: begin
          printf,unit,sep,'|Q|(A^-1)',sep,'E(meV)',sep,'I',sep,'dI'
          for i=0L,nen-1 do begin
              for j=0L,n2th-1 do begin
                  printf,unit,(*self.projStrPtr).Q[i,j],(*self.projStrPtr).en[i],' ',(*self.projStrPtr).qty[i,j],' ',(*self.projStrPtr).err[i,j]
              endfor
          endfor
          end
       1: begin
          if self.extravaxis_yn then begin
             printf,unit,sep,self.traj_type[0],sep,self.traj_type[1],sep,self.traj_type[2],sep,'E(meV)',sep,'I',sep,'dI'
             for i=0L,nen-1 do begin
                 for j=0L,n2th-1 do begin
                     printf,unit,(*self.projStrPtr).V[0,i,j],(*self.projStrPtr).V[1,i,j],(*self.projStrPtr).V[2,i,j],(*self.projStrPtr).en[i],' ',(*self.projStrPtr).qty[i,j],' ',(*self.projStrPtr).err[i,j]
                 endfor
             endfor
          endif else begin
             printf,unit,sep,self.traj_type[0],sep,self.traj_type[1],'E(meV)',sep,'I',sep,'dI'
             for i=0L,nen-1 do begin
                 for j=0L,n2th-1 do begin
                     printf,unit,(*self.projStrPtr).V[0,i,j],(*self.projStrPtr).V[1,i,j],(*self.projStrPtr).en[i],' ',(*self.projStrPtr).qty[i,j],' ',(*self.projStrPtr).err[i,j]
                 endfor
             endfor
          endelse
          end
       2: begin
          tmp = size((*self.projStrPtr).V)
          if tmp[0] eq 2 then nfile=1L else nfile=tmp[3]
          if self.extravaxis_yn then begin
             printf,unit,sep,self.traj_type[0],sep,self.traj_type[1],sep,self.traj_type[2],sep,'I',sep,'dI'
             for i=0L,n2th-1 do begin
                 for j=0L,nfile-1 do begin
                     printf,unit,(*self.projStrPtr).V[0,i,j],(*self.projStrPtr).V[1,i,j],(*self.projStrPtr).V[2,i,j],' ',(*self.projStrPtr).qty[i,j],' ',(*self.projStrPtr).err[i,j]
                 endfor
             endfor
          endif else begin
             printf,unit,sep,self.traj_type[0],sep,self.traj_type[1],sep,'I',sep,'dI'
             for i=0L,n2th-1 do begin
                 for j=0L,nfile-1 do begin
                     printf,unit,(*self.projStrPtr).V[0,i,j],(*self.projStrPtr).V[1,i,j],' ',(*self.projStrPtr).qty[i,j],' ',(*self.projStrPtr).err[i,j]
                 endfor
             endfor
          endelse
          end
       3: begin
          tmp = size((*self.projStrPtr).V)
          if tmp[0] eq 3 then nfile=1L else nfile=tmp[4]
          if self.extravaxis_yn then begin
             printf,unit,sep,self.traj_type[0],sep,self.traj_type[1],sep,self.traj_type[2],sep,'E(meV)',sep,'I',sep,'dI'
             for i=0L,nen-1 do begin
                 for j=0L,n2th-1 do begin
                     for k=0L,nfile-1 do begin
                         printf,unit,(*self.projStrPtr).V[0,i,j,k],(*self.projStrPtr).V[1,i,j,k],(*self.projStrPtr).V[2,i,j,k],(*self.projStrPtr).en[i],' ',(*self.projStrPtr).qty[i,j,k],' ',(*self.projStrPtr).err[i,j,k]
                     endfor
                 endfor
             endfor
          endif else begin
             printf,unit,sep,self.traj_type[0],sep,self.traj_type[1],sep,'E(meV)',sep,'I',sep,'dI'
             for i=0L,nen-1 do begin
                 for j=0L,n2th-1 do begin
                     for k=0L,nfile-1 do begin
                         printf,unit,(*self.projStrPtr).V[0,i,j,k],(*self.projStrPtr).V[1,i,j,k],(*self.projStrPtr).en[i],' ',(*self.projStrPtr).qty[i,j,k],' ',(*self.projStrPtr).err[i,j,k]
                     endfor
                 endfor
             endfor
          endelse
          end   
    endcase   
    
    free_lun,unit
end

;save the volume data
pro dcs_mslice::save_volume,txt=txt
    if keyword_set(txt) then begin
       file = dm_choose_file('txt',/write,dialog_parent=self.tlb,title='Save volume data as a multi-column text file.',path=path)
       if strlen(file) eq 0 then return
       if !version.os_family eq 'Windows' then sep=string(9b) else sep='     '
       openw,unit,file,/get_lun,error=openerr
       if openerr ne 0 then begin
          ok = dialog_message("Can't write in "+file,/error,dialog_parent=self.tlb)
          return
       endif
    endif
    
    ;select data
    tmp = self.vol_smooth
    if self.vol_smooth gt 0 then begin
       info = 'Do you want to save the smoothed data?'
       ok = dialog_message(info,/question,dialog_parent=self.tlb)
       if strlowcase(ok) eq 'no' then self.vol_smooth = 0
    endif  
    self->dm_choose_volumedata,xdat=xdat,ydat=ydat,zdat=zdat,idat=idat,ierr=ierr,xtit=xtit,ytit=ytit,ztit=ztit

    self.vol_smooth = tmp
    itmp = where(finite(idat) eq 0,count)
    if count ne 0 then begin
       notok  = 1b
       filler = -1e20
       while(notok) do begin
           filler = dm_dialog_input('filler value:',/float,default=filler,$
                    info='This value is used for empty grids that contain no data.',dialog_parent=self.tlb)
            notok = ~finite(filler)
       endwhile
       idat[itmp] = filler
       ierr[itmp] = -1 
       itmp = -1
    endif
    widget_control,/hourglass  ;busy signal
    if keyword_set(txt) then begin
       printf,unit,sep,xtit,sep,ytit,sep,ztit,sep,'I',sep,'dI'
       for i=0L,n_elements(xdat)-1 do begin
           for j=0L,n_elements(ydat)-1 do begin
               for k=0L,n_elements(zdat)-1 do begin
                   printf,unit,xdat[i],' ',ydat[j],' ',zdat[k],' ',idat[i,j,k],' ',ierr[i,j,k]
               endfor
           endfor    
       endfor 
       free_lun,unit
    endif
end

;set the buttons after projection is changed
pro dcs_mslice::set_display_button
    pd_valid = 0b
    if ptr_valid(self.projStrPtr) then begin
       if n_elements((*self.projStrPtr).qty) gt 1 then pd_valid=1b
    endif
    self->my_widget_control,'savevolumeMenu',sensitive=(pd_valid and ((self.samp_typ eq 1) or (self.samp_typ eq 2 and self.extravaxis_yn) or (self.samp_typ eq 3)))
    self->my_widget_control,['savesliceMenu','savecutMenu','saveprojMenu'],sensitive=pd_valid
    self->my_widget_control,'viewSpurion',sensitive=ptr_valid(self.dataStrPtr) and ptr_valid(self.projStrPtr) and (self.samp_typ lt 2) and (self.instrgeom eq 0)
    self->my_widget_control,['dispFitSlice','dispPlotSlice','dispSurfSlice','cutFit','cutPlot','cutOplot','volPlot'],self.dispBas,sensitive=pd_valid
    self->my_widget_control,'plotTra', self.dispBas,sensitive=ptr_valid(self.projStrPtr)
    self->my_widget_control,'oplotTra',self.dispBas,sensitive=(ptr_valid(self.projStrPtr) and (self.samp_typ ne 0))
end

pro dcs_mslice::set_filehistory,info,dir,files,choose=choose
    if n_elements(choose) eq 1 then begin
       self.filesel->set_path,(*(self.file_hist[choose]))[2]
       ftype = strmid((*(self.file_hist[choose]))[0],6,strlen((*(self.file_hist[choose]))[0])-6)
       if ftype ne self.ftype then ok=self->dm_filetools((*(self.file_hist[choose]))[0])
       self.filesel->getproperty,dFiles=dFiles
       for i=3,n_elements(*(self.file_hist[choose]))-1 do begin
           ind = where(dFiles eq (*(self.file_hist[choose]))[i],count)
           if count ne 0 then begin
              if n_elements(selected) eq 0 then selected=ind $
              else selected=[selected,ind]
           endif
       endfor
       if n_elements(selected) ne 0 then self.filesel->set_file,selected,/highlight
       return
    endif
    num = n_elements(self.file_hist)    ;capacity
    if (n_elements(info) ne 0) and (n_elements(dir) ne 0) and (n_elements(files) ne 0) then begin
       for i=0,self.n_history-1 do begin  ;check if this entry already exists
           if dir ne (*(self.file_hist[i]))[2] then continue
           if n_elements(files) ne n_elements(*(self.file_hist[i]))-3 then continue
           same = 1b
           for j=0, n_elements(files)-1 do $
               if files[j] ne (*(self.file_hist[i]))[3+j] then same=0b
           if same then return ;entry exists
       endfor
       if self.n_history lt num then begin ;empty room available
          if self.n_history ne 0 then $
             self.file_hist[1:self.n_history] = self.file_hist[0:self.n_history-1]
          self.file_hist[0] = ptr_new(['ftype_'+dm_to_string(self.ftype,/int),info,dir,files])
          self.n_history = self.n_history+1
       endif else begin                    ;full
          ptr_free,self.file_hist[num-1]
          self.file_hist[1:num-1] = self.file_hist[0:num-2]
          self.file_hist[0] = ptr_new(['ftype_'+dm_to_string(self.ftype,/int),info,dir,files])
       endelse
    endif
    histmenu = widget_info(self.tlb,find_by_uname='history')
    widget_control,histmenu,sensitive=(self.n_history gt 0)
    self->my_widget_control,'clearHist',sensitive=(self.n_history gt 0)
    for i=0,num-1 do self->my_widget_control,'hist_'+dm_to_string(i,/int),histmenu,/destroy
    for i=0,self.n_history-1 do begin
        void = widget_button(histmenu,value=dm_to_string(i,/int)+' '+$
               (*(self.file_hist[i]))[1],uname='hist_'+dm_to_string(i))
    endfor
end

;set instrument geometry and adjust the proper menu items
pro dcs_mslice::set_instrgeom,geom=geom,sensitive=sensitive
    igeoDir  = widget_info(self.tlb,find_by_uname='directGeom')
    igeoInv  = widget_info(self.tlb,find_by_uname='inverseGeom')
    if n_elements(geom) ne 0 then begin
       self.instrgeom = 0>(geom)<1
       if self.instrgeom eq 0 then begin    ;direct geometry
          dm_toggle_menubut,check=igeoDir,uncheck=igeoInv
          self->my_widget_control,'edepEff',/sensitive
          self->my_widget_control,'viewSpurion',$
             sensitive=ptr_valid(self.dataStrPtr) and ptr_valid(self.projStrPtr) and (self.samp_typ lt 2)
       endif else begin                     ;inverse geometry
          dm_toggle_menubut,check=igeoInv,uncheck=igeoDir
          self->my_widget_control,['edepEff','viewSpurion'],sensitive=0
       endelse
    endif
    if n_elements(sensitive) ne 0 then begin
       widget_control,igeoDir,sensitive=sensitive
       widget_control,igeoInv,sensitive=sensitive
    endif
end

;set intensity type
pro dcs_mslice::set_intntype,type
    ntype = n_elements(self.intn_typ_str)
    self.intn_typ = 0>(type)<(ntype-1)
    inttypid = lonarr(ntype)
    for i=0,ntype-1 do $
        inttypid[i] = widget_info(self.tlb,find_by_uname='intnTyp'+dm_to_string(i,/int))
    dm_toggle_menubut,check=inttypid[self.intn_typ],uncheck=inttypid[where(indgen(ntype) ne self.intn_typ)]
    ;need to recalculate projections
    ptr_free,self.projStrPtr
    self->set_display_button
end

;cleanup, life cycle method
pro dcs_mslice::Cleanup
    ;save parameters in default parameter file if allowed
    parmname = (['mslice.prm','dcs_mslice.prm'])[self.samp_typ ge 0]
    if file_test(self.dirs[1]) then begin
       cd,self.dirs[1]
       openw,unit,parmname,/get_lun,error=openerr
       if openerr eq 0 then begin
          free_lun,unit
          self->dm_save_parameter,parmname
       endif
    endif
    if widget_info(self.tlb,/valid_id) then widget_control,self.tlb,/destroy
    ptr_free,self.ftypename,self.ftypeext,self.bgratePtr,self.bgtchnPtr,self.bgdetsPtr,self.bgdata,self.bgdata2
    ptr_free,self.powd_qvalue,self.dcor_info,self.mask,self.eff,self.kfactor,self.rebinsize,self.powd_line,self.absp_info,self.file_hist,self.absp_npts,self.absp_nptst
    ptr_free,self.detPosPtr,self.dataStrPtr,self.projStrPtr,self.dataPtr,self.info,self.dispsel_t,self.dispsel_tft,self.cutsel_t,self.cutsel_tft,self.volsel_t,self.volsel_tft
    obj_destroy,self.gr_list
    obj_destroy,self.fileSel
end

;initialization, life cycle method
;usage:
;   void=obj_new('dcs_mslice',group_leader=group_leader)
;parameters:
;   NONE
;keyword:
;   groupleader:    groupleader widget ID
;   dataPtr:        if present, starts it as a volume data viewer
function dcs_mslice::Init,dataPtr=dataPtr,group_leader=group_leader,workDir=workDir,dataDir=dataDir,DAVETool=DAVETool
    if (n_elements(dataPtr) ne 0) then begin
       ;check dataPtr
       field = strupcase(tag_names(*dataPtr))
       if (~ ptr_valid(dataPtr)) then return,0
       loc   = where(field eq 'DATA',count)
       if count eq 0 then return,0
       nitem = n_elements((*dataPtr).data[0,*])
       if nitem lt 4 then return,0
       loc   = where(field eq 'LABEL',count)
       if count eq 0 then begin
          loc   = where(field eq 'LABELS',count)
          if count ne 0 then label = (*dataPtr).labels $
          else begin 
             label = ['Intensity','Error']
             for i=0,nitem-3 do label = [label,'x'+dm_to_string(i+1)]
          endelse
          *dataPtr = create_struct(*dataPtr,'label',label)
       endif
       loc   = where(field eq 'INSTRUMENT',count)
       if count eq 0 then *dataPtr = create_struct(*dataPtr,'instrument','')
       loc   = where(field eq 'UNIT',count)
       if count eq 0 then begin
          loc   = where(field eq 'UNITS',count)
          if count ne 0 then unit = (*dataPtr).units $
          else unit = strarr(nitem)
          *dataPtr = create_struct(*dataPtr,'unit',unit)
       endif
       loc   = where(field eq 'IS_UNIQ',count)
       if count eq 0 then begin
          is_uniq = bytarr(nitem)
          *dataPtr = create_struct(*dataPtr,'is_uniq',is_uniq)
       endif
       loc   = where(field eq 'AUTHOR',count)
       if count ne 0 then author = ', '+(*dataPtr).author[0] else author = ''
    endif
    if n_elements(group_leader) ne 0 then self.group_leader=group_leader
    registerName = (['dcs_mslice','mslice'])[ptr_valid(dataPtr)] 
    if xregistered(registerName) and (~ ptr_valid(dataPtr)) then begin ;only allow one dcs_mslice running
       FORWARD_FUNCTION LookupManagedWidget
       id = LookupManagedWidget(registername)
       widget_control,id,iconify=0
       return,0
    endif
    
    ;error handler
    catch, myerror
    if myerror ne 0 then begin
       catch,/cancel
       ok = dialog_message(!error_state.msg,/error)
       ptr_free,self.ftypename,self.ftypeext,self.bgratePtr,self.bgtchnPtr,self.bgdetsPtr,self.bgdata,self.bgdata2
       ptr_free,self.powd_qvalue,self.dcor_info,self.mask,self.eff,self.kfactor,self.rebinsize,self.powd_line,self.absp_info,self.file_hist,self.absp_npts,self.absp_nptst
       ptr_free,self.detPosPtr,self.dataStrPtr,self.projStrPtr,self.dataPtr,self.info,self.dispsel_t,self.dispsel_tft,self.cutsel_t,self.cutsel_tft,self.volsel_t,self.volsel_tft
       obj_destroy,self.gr_list
       obj_destroy,self.fileSel
       return, 0
    end

    ;initialize the data
    self.instrgeom      = 0     ;default direct geometry
    self.dcor_yn        = 0
    self.edep_yn        = 1
    self.bitflip        = 1
    self.dcs2thcorr     = 0
    self.mask_filler    = -1e20
    self.pbgcol_bw      = 0     ;defalut black
    self.eadjust        = 0
    self.eadjust_spec   = !values.f_nan
    self.eadjust_checkfirst = 0
    self.eint_yn        = 0
    self.intn_typ       = 0     ;default S(q,w)
    self.intn_typ_str   = ['S(Q,omega)','Symmetrized S(Q,omega)','d2(sigma)/d(Omega)d(Ef)',$
                           'Chi(Q,omega)','GDOS','VDOS','Unadjusted by ki/kf factor']
    self.powd_typ       = 0     ;no powder line
    self.dcor_eqran     = [!values.f_nan,!values.f_nan,!values.f_nan,!values.f_nan]
    self.xsize          = [549L,40L,120L,9L,92L]
    self.ssfac          = 1.0
    self.view_ortharbi  = 1     ;arbitrary
    self.bin_avgsum     = 0     ;average
    self.binthreshold   = 0.0   ;treshold level
    self.bin_conststep  = 0     ;average value
    self.bin_extlib     = 1     ;use external library
    self.mon_sumfit     = 1     ;fit
    self.dcs_mon        = 1     ;use bm1 as the monitor
    self.temperature    = 300.0
    self.bgeran         = [-20,-10]
    self.intn_lab       = 'Intensity (arb. units)'
    self.samp_typ       = 0     ;default powder
    self.rebinsize      = ptr_new([1s,2s,4s,5s,8s,10s,20s,25s,40s,50s])
    self.e_bin          = [1,1,1,1,1]
    self.e_range        = [!values.f_nan,!values.f_nan]
    self.latt_parm      = [10d,10d,10d,90d,90d,90d]
    self.orie_u         = [1e,0e,0e]
    self.orie_v         = [0e,1e,0e]
    self.orie_offset    = !values.f_nan
    self.view_type      = ['Energy','|Q|','2Theta','Det Group Number','Time Channel']
    self.view_label     = ['E','|Q|','2theta','Detector Group','Time']
    self.view_unit      = ['meV',string('c5'XB)+'!U-1 !N',string('b0'XB),'','us']
    self.view_u1typ     = 1
    self.view_u1        = [1,0,0]
    self.view_u2        = [0,1,0]
    self.view_u3        = [0,0,1]
    self.view_u1lab     = ['|Q|','H']
    self.view_u2lab     = ['E','K']
    self.view_u3lab     = 'L'
    self.vol_smooth     = 1
    self.vol_interp     = 1     ;0:nearest neighbor 1:trilinear
    self.vol_color      = 13    ;rainbow
    self.volsel_x       = 0
    self.volsel_y       = 1
    self.volsel_z       = 2+self.extravaxis_yn
    self.vol_xran       = [!values.f_nan,!values.f_nan]
    self.vol_xstep      = !values.f_nan
    self.vol_yran       = [!values.f_nan,!values.f_nan]
    self.vol_ystep      = !values.f_nan
    self.vol_zran       = [!values.f_nan,!values.f_nan]
    self.vol_zstep      = !values.f_nan
    self.vol_iran       = [!values.f_nan,!values.f_nan]
    self.volsel_t1      = 2
    self.volsel_tft1    = [!values.f_nan,!values.f_nan]
    self.disp_perp      = 2
    self.dispsel_h      = 0
    self.dispsel_v      = 1
    self.disp_pran      = [!values.f_nan,!values.f_nan]
    self.disp_hran      = [!values.f_nan,!values.f_nan]
    self.disp_vran      = [!values.f_nan,!values.f_nan]
    self.disp_tran      = [!values.f_nan,!values.f_nan]
    self.disp_hstep     = 0.035
    self.disp_vstep     = !values.f_nan
    self.disp_iran      = [!values.f_nan,!values.f_nan]
    self.disp_smooth    = 1
    self.disp_sm_kernel = [[0.1e,0.2e,0.1e],[0.2e,0.8e,0.2e],[0.1e,0.2e,0.1e]]
    self.disp_shading   = 1
    self.dispcmap       = ['B-W LINEAR','BLUE/WHITE','GRN-RED-BLU-WHTt','RED TEMPERATURE','BLUE/GREEN/RED/YELLOW','STD GAMMA-II',$
                           'PRISM','RED-PURPLE','GREEN/WHITE LINEAR','GRN/WHT EXPONENTIAL','GREEN-PINK','BLUE-RED','16 LEVEL',$
                           'RAINBOW','STEPS','STERN SPECIAL','Haze','Blue-Pastel-Red','Pastels','Hue Sat Lightness 1',$
                           'Hue Sat Lightness 2','Hue Sat Value 1','Hue Sat Value 2','Purple-Red+Stripes','Beach','Mac Style',$
                           'Eos A','Eos B','Hardcandy','Nature','Ocean','Peppermint','plasma','Blue-Red','Rainbow','Blue Waves',$
                           'Volcano','Waves','Rainbow18','Rainbow+white','Rainbow+black']
    self.disp_color     = 13    ;rainbow
    self.cut_x          = [0s,0s,0s,0s]
    self.cut_y          = [0s,0s,0s,0s]
    self.cut_ran        = [!values.f_nan,!values.f_nan]
    self.cut_step       = 0.035
    self.cut_t1ran      = [!values.f_nan,!values.f_nan]
    self.cut_t2ran      = [!values.f_nan,!values.f_nan]
    self.cut_t3ran      = [!values.f_nan,!values.f_nan]
    self.cut_yran       = [!values.f_nan,!values.f_nan]
    self.cut_symbol     = 0s
    self.symbols        = ['circle','filled circle','diamond','filled diamond','triangle','filled triangle',$
                           'inverted triangle','filled inverted triangle','square','filled square',$
                           'plus','star','dot','cross','greater','less','no symbol']
    self.cut_color      = 0
    self.colors         = ['white','red','blue','magenta','green','cyan','yellow','grey','dark grey','dark red','dark blue',$
                           'dark magenta','dark green','dark cyan','dark yellow','user define']
    self.coldefine      = 255b-[255b,255b,255b]*self.pbgcol_bw
    self.cut_line       = 0
    self.lines          = ['solid','dotted','dashed','dash dot','dash dot dot dot','long dash','no line']
    self.traj_type      = ['[Qh,0,0]','[0,Qk,0]','[0,0,Ql]','E (meV)','None']
    self.traj_x         = [1,0,0,0]
    self.traj_xran      = [!values.f_nan,!values.f_nan]
    self.traj_y         = [0,1,1,1]
    self.traj_yran      = [!values.f_nan,!values.f_nan]
    self.traj_z1        = [5,4,3,4]
    self.traj_z1ran     = [!values.f_nan,!values.f_nan]
    self.traj_z2        = [0,4,3,4]
    self.traj_z2ran     = [!values.f_nan,!values.f_nan]
    self.gr_list        = obj_new('IDL_Container')
    self.errbar         = 1b
    self.useitool       = 0b
    self.negative2th    = 0b
    self.empcanfactor   = [1.0,1.0]
    self.eief           = !values.f_nan
    self.estartend      = [!values.f_nan,!values.f_nan,81.]
    self.a2[*]          = !values.f_nan
    self.absp_npts      = [ptr_new([[30,60],[60,120],[90,180]]),$
                           ptr_new([[30,60],[60,120],[90,180]]),$
                           ptr_new([[30,60,60],[60,120,120],[90,180,180]]),$
                           ptr_new([[60,30],[120,60],[180,90]])]
    self.absp_nptst     = [ptr_new(['nr=30, nphi=60','nr=60, nphi=120','nr=90, nphi=180']),$
                           ptr_new(['nr=30, nphi=60','nr=60, nphi=120','nr=90, nphi=180']),$
                           ptr_new(['nr=30, nphi=60, ntheta=60','nr=60, nphi=120, ntheta=120','nr=90, nphi=180, ntheta=180']),$
                           ptr_new(['nw=60, nt=30','nw=120, nt=60','nw=180, nt=90'])]
    mod_date            = '10/2010'
    self.info           = ptr_new(['--Mslice for DCS --', $
                           'Based on the NCNR DCS data reduction programs by John Copley',$
                           'and the matlab version of mslice by Radu Coldea.',$
                           'Yiming Qiu(yiming.qiu@nist.gov)   '+mod_date])
    
    if float(!version.release) ge 5.5 then self.pathsep=call_function("path_sep") $
    else begin
       defsysv,'!delimiter',exists=exists
       if (~ exists) then begin
          case !version.os_family of
            'Windows':  defsysv,'!delimiter','\'
            'unix':     defsysv,'!delimiter','/'
            'MacOS':    defsysv,'!delimiter',':'
            else:       defsysv,'!delimiter','/'
          endcase
       endif
       self.pathsep = !delimiter
    endelse
    
    cd,current=current
    self.dirs = [current,current] 
    defsysv,'!dave_defaults',exists=exists
    if exists then $
       if ptr_valid(!dave_defaults) then begin
          if (file_test((*!dave_defaults).datDir,/directory))[0]  then self.dirs[0] = (*!dave_defaults).datDir
          if (file_test((*!dave_defaults).workDir,/directory))[0] then self.dirs[1] = (*!dave_defaults).workDir
       endif
    if n_elements(dataDir) ne 0 then $
       if strlen(dataDir) ne 0 then self.dirs[0] = dataDir
    if n_elements(workDir) ne 0 then $
       if strlen(workDir) ne 0 then self.dirs[1] = workDir
    self.helpfile       = ['dcs_mslice.pdf','dcs_mslice_addftype.pdf']
    defsysv,'!DAVE_PDFHELP_DIR',exists=exists
    if exists then self.helpfile = !DAVE_PDFHELP_DIR+self.pathsep+self.helpfile $
    else begin
       self.helpfile[0] = file_which(self.helpfile[0],/include)
       self.helpfile[1] = file_which(self.helpfile[1],/include)
    endelse
    defsysv,'!dcs_hsq2mn',exists=exists
    if (~ exists) then defsysv,'!dcs_hsq2mn',double(81.804200) ;h^2/2./m_n/e*1e23

    if n_elements(DAVETool) ne 0 then $
       if obj_valid(DAVETool) then self.DAVETool = DAVETool

    ;generate widgets
    if ptr_valid(dataPtr) then begin
       self.mask_filler = !values.f_nan
       self.samp_typ  = -1
       self.dataPtr   = dataPtr 
       n_item         = n_elements((*dataPtr).label)
       if n_item gt 5 then begin
          self.volsel_t   = ptr_new(indgen(n_item-5)+3)
          self.volsel_tft = ptr_new(fltarr(n_item-5,2))
          (*self.volsel_tft)[*] = !values.f_nan 
       endif
       if n_item gt 4 then begin
          self.dispsel_t   = ptr_new(indgen(n_item-4)+2)
          self.dispsel_tft = ptr_new(fltarr(n_item-4,2))
          (*self.dispsel_tft)[*] = !values.f_nan 
       endif
       if n_item gt 3 then begin
          self.cutsel_t    = ptr_new(indgen(n_item-3)+1)
          self.cutsel_tft  = ptr_new(fltarr(n_item-3,2))
          (*self.cutsel_tft)[*] = !values.f_nan 
       endif
       *(self.info)   = ['--Mslice'+(strlen((*dataPtr).instrument) eq 0?'':' for '+(*dataPtr).instrument)+' --',$
                         'Yiming Qiu(yiming.qiu@nist.gov)'+author+'   '+mod_date]
       self.tlb = widget_base(title=(*self.info)[0],/column,kill_notify='dcs_mslice_Exit',mbar=bar,$
                  group_leader=self.group_leader,ypad=0,xpad=0,uname='tlb',/tlb_size_event,map=0)
       filemenu = widget_button(bar,value='File',/menu)
       parmmenu = widget_button(bar,value='Parameter',/menu)
       maskmenu = widget_button(bar,value='Mask',/menu)
       optnmenu = widget_button(bar,value='Option',/menu)
       viewmenu = widget_button(bar,value='Window',/menu,uname='viewMenu') 
       helpmenu = widget_button(bar,value='Help',/menu) 
       binmmenu = widget_button(optnmenu,value='Binning Method',/menu)  
       pbgcmenu = widget_button(optnmenu,value='Plot Window Background Color',/menu)
       fvolmenu = widget_button(filemenu,value='Save Volume Data As',uname='savevolumeMenu',/menu,sensitive=(n_item ge 5))
       fslimenu = widget_button(filemenu,value='Save Slice Data As',uname='savesliceMenu',/menu)
       fcutmenu = widget_button(filemenu,value='Save Cut Data As',uname='savecutMenu',/menu) 
       fpromenu = widget_button(filemenu,value='Save All Data As',uname='saveprojMenu',/menu)   
       void     = widget_button(filemenu,value='Exit',uname='exitBut',/separator)
       void     = widget_button(fvolmenu,value='Text File',uname='savevolumetxt')
       void     = widget_button(fslimenu,value='DAVE File',uname='saveslicedave')
       void     = widget_button(fslimenu,value='SPE File',uname='saveslicespe')
       void     = widget_button(fslimenu,value='I,Err,X,Y 4-Column File',uname='savesliceiexy')
       void     = widget_button(fslimenu,value='X,Y,I,Err Matrix File',uname='saveslicexyie')
       if obj_valid(self.DAVETool) then $
          void  = widget_button(fslimenu,value='Send to DAVE Data Manager',uname='sendslicedave',/separator)
       void     = widget_button(fcutmenu,value='X,Y,Err 3-Column File',uname='savecutxye')
       void     = widget_button(fcutmenu,value='DAVE File',uname='savecutdave')
       if obj_valid(self.DAVETool) then $
          void  = widget_button(fcutmenu,value='Send to DAVE Data Manager',uname='sendcutdave',/separator)
       void     = widget_button(fpromenu,value='Text File',uname='saveprojtxt')
       void     = widget_button(parmmenu,value='Load Parameters...',uname='ldpmBut')
       void     = widget_button(parmmenu,value='Save Parameters...',uname='svpmBut')
       void     = widget_button(parmmenu,value='Change Smoothing Kernel...',uname='knpmBut',/separator)
       void     = widget_button(parmmenu,value='Change Binning Threshold...',uname='bnthBut')
       void     = widget_button(parmmenu,value='Set Working Directory...',uname='wdirBut',/separator)
       void     = dm_widget_button(maskmenu,value='Mask Data by Value',uname='maskFiller')  
       binAvg   = dm_widget_button(binmmenu,value='Average',uname='binAvg')
       binSum   = dm_widget_button(binmmenu,value='Sum',uname='binSum')
       bgBlack  = dm_widget_button(pbgcmenu,value='Black',uname='pbgBlack')
       bgWhite  = dm_widget_button(pbgcmenu,value='White',uname='pbgWhite')
       bincsmenu= widget_button(binmmenu,value='Constant Intervals',/menu,/separator)
       bincsYes = dm_widget_button(bincsmenu,value='Yes',uname='bincsYes')
       bincsNo  = dm_widget_button(bincsmenu,value='No',uname='bincsNo')
       binexmenu= widget_button(binmmenu,value='Use External Binning Library',/menu)
       binexYes = dm_widget_button(binexmenu,value='Yes',uname='binexYes')
       binexNo  = dm_widget_button(binexmenu,value='No',uname='binexNo')
       void     = widget_button(viewmenu,value='Close All Plot Windows',uname='killAll',sensitive=0)
       arrpmenu = widget_button(viewmenu,value='Arrange Plots',uname='arrPlot',sensitive=0,/menu)
       void     = widget_button(arrpmenu,value='Cascade',uname='arrCascade')
       void     = widget_button(arrpmenu,value='Table',uname='arrTable')
       void     = widget_button(helpmenu,value='About Mslice',uname='selfBut')
       
       ;menu bar separator for windows system
       if !version.os_family eq 'Windows' then $
          mbar_sep = widget_label(self.tlb,value=' ',sensitive=0,/dynamic_resize,/align_top,scr_ysize=5) $
       else $
          self.xsize[2] = self.xsize[2]+20
          
       ;load mslice.prm if existed
       parmfile = self.dirs[1]+self.pathsep+'mslice.prm'
       if file_test(parmfile,/read) then $
          self->dm_load_parameter,parmfile,/init   
       
       dm_toggle_menubut,check=([binAvg,binSum])[self.bin_avgsum],uncheck=([binAvg,binSum])[1-self.bin_avgsum]
       dm_toggle_menubut,check=([bincsNo,bincsYes])[self.bin_conststep],uncheck=([bincsNo,bincsYes])[1-self.bin_conststep]
       dm_toggle_menubut,check=([binexNo,binexYes])[self.bin_extlib],uncheck=([binexNo,binexYes])[1-self.bin_extlib]
       dm_toggle_menubut,check=([bgBlack,bgWhite])[self.pbgcol_bw],uncheck=([bgBlack,bgWhite])[1-self.pbgcol_bw]
       self->reset_display
       
       widget_control,self.dispBas,/update,/map
    endif else begin
       self.tlb = widget_base(title=(*self.info)[0],/column,kill_notify='dcs_mslice_Exit',mbar=bar,$
                  group_leader=self.group_leader,ypad=0,xpad=0,uname='tlb',/tlb_size_event,map=0)
       filemenu = widget_button(bar,value='File',/menu)
       parmmenu = widget_button(bar,value='Parameter',/menu,uname='parmMenu')
       bkgrmenu = widget_button(bar,value='Background',/menu,uname='bkgrMenu')
       maskmenu = widget_button(bar,value='Mask',/menu)
       optnmenu = widget_button(bar,value='Option',/menu)
       viewmenu = widget_button(bar,value='Window',/menu,uname='viewMenu')
       helpmenu = widget_button(bar,value='Help',/menu)
       igeomenu = widget_button(optnmenu,value='Instrument Geometry',uname='igeomMenu',/menu)
       itypmenu = widget_button(optnmenu,value='View Intensity As',/menu)
       binmmenu = widget_button(optnmenu,value='Binning Method',/menu)
       moncmenu = widget_button(optnmenu,value='Monitor Count',uname='moncMenu',/menu)
       abspmenu = widget_button(optnmenu,value='Absorption Correction',/menu)
       vaxsmenu = widget_button(optnmenu,value='Viewing Axis',/menu,uname='vaxisMenu',sensitive=0)
       powdmenu = widget_button(optnmenu,value='Show Powder Line',/menu,uname='showPowder')
       epadmenu = widget_button(optnmenu,value='Elastic Peak Position Adjustment',/menu,uname='eAdjust')
       pbgcmenu = widget_button(optnmenu,value='Plot Window Background Color',/menu)
       void     = widget_button(filemenu,value='File Type',uname='ftypeMenu',/menu)
       void     = widget_button(filemenu,value='File Tools',uname='ftoolMenu',/menu,sensitive=0)
       fvolmenu = widget_button(filemenu,value='Save Volume Data As',uname='savevolumeMenu',/menu,/separator,sensitive=0)
       fslimenu = widget_button(filemenu,value='Save Slice Data As',uname='savesliceMenu',/menu,sensitive=0)
       fcutmenu = widget_button(filemenu,value='Save Cut Data As',uname='savecutMenu',/menu,sensitive=0)
       fpromenu = widget_button(filemenu,value='Save Projection As',uname='saveprojMenu',/menu,sensitive=0)
       void     = widget_button(filemenu,value='Recent Files',uname='history',sensitive=0,/menu,/separator)
       void     = widget_button(filemenu,value='Clear File History',uname='clearHist',sensitive=0)
       void     = widget_button(filemenu,value='Exit',uname='exitBut',/separator)
       void     = widget_button(fvolmenu,value='Text File',uname='savevolumetxt')
       void     = widget_button(fslimenu,value='DAVE File',uname='saveslicedave')
       void     = widget_button(fslimenu,value='SPE File',uname='saveslicespe')
       void     = widget_button(fslimenu,value='I,Err,X,Y 4-Column File',uname='savesliceiexy')
       void     = widget_button(fslimenu,value='X,Y,I,Err Matrix File',uname='saveslicexyie')
       if obj_valid(self.DAVETool) then $
          void  = widget_button(fslimenu,value='Send to DAVE Data Manager',uname='sendslicedave',/separator)
       void     = widget_button(fcutmenu,value='X,Y,Err 3-Column File',uname='savecutxye')
       void     = widget_button(fcutmenu,value='DAVE File',uname='savecutdave')
       if obj_valid(self.DAVETool) then $
          void  = widget_button(fcutmenu,value='Send to DAVE Data Manager',uname='sendcutdave',/separator)
       void     = widget_button(fpromenu,value='Text File',uname='saveprojtxt')
       void     = widget_button(parmmenu,value='Load Parameters...',uname='ldpmBut')
       void     = widget_button(parmmenu,value='Save Parameters...',uname='svpmBut')
       void     = widget_button(parmmenu,value='Change Self-shielding Factor...',uname='ssfcBut',/separator)
       if self.extraempcan_yn then begin
          void  = widget_button(parmmenu,value='Change 2-Set Empty Can Factors...',uname='emfcBut',sensitive=0)
       endif
       void     = widget_button(parmmenu,value='Change Smoothing Kernel...',uname='knpmBut')
       void     = widget_button(parmmenu,value='Change Binning Threshold...',uname='bnthBut')
       void     = widget_button(parmmenu,value='Set Data Directory...',uname='ddirBut',/separator)
       void     = widget_button(parmmenu,value='Set Working Directory...',uname='wdirBut')
       if self.extraempcan_yn then begin
          void  = widget_button(bkgrmenu,value='Load Empty Can File(s) 1',uname='emptBut')
          void  = widget_button(bkgrmenu,value='Load Empty Can File(s) 2',uname='emptBut2',sensitive=0)
       endif else $
          void  = widget_button(bkgrmenu,value='Load Empty Can File(s)',uname='emptBut')
       void     = widget_button(bkgrmenu,value='Load Detector Background File(s)',uname='bgflBut')
       void     = widget_button(bkgrmenu,value='Load Time Channel Background File(s)',uname='bgtcBut')
       void     = widget_button(bkgrmenu,value='Load Dark Count File(s)',uname='bgdkBut')
       void     = widget_button(bkgrmenu,value='Set Overall Background Rate...',uname='bgovBut')
       void     = widget_button(bkgrmenu,value='Clear Background',uname='bgclBut',sensitive=0)
       void     = widget_button(bkgrmenu,value='Load Vanadium File(s)',uname='vandBut',/separator)
       void     = widget_button(bkgrmenu,value='Clear Vanadium',uname='vdclBut',sensitive=0)
       mkdbmenu = widget_button(maskmenu,value='DCS Mask',uname='mkdbMenu',/menu)
       void     = dm_widget_button(mkdbmenu,value='Mask Negative Angle Detectors',uname='mkngBut')
       void     = dm_widget_button(mkdbmenu,value='Mask Upper and Lower Banks',uname='mkulBut')
       void     = dm_widget_button(mkdbmenu,value='Mask Lower Bank',uname='mkloBut',/separator)
       void     = dm_widget_button(mkdbmenu,value='Mask Central Bank',uname='mkcnBut')
       void     = dm_widget_button(mkdbmenu,value='Mask Upper Bank',uname='mkupBut')
       stmkmenu = widget_button(maskmenu,value='Set Mask',/menu)
       void     = widget_button(stmkmenu,value='Set Mask...',uname='mkstBut')
       plmkmenu = widget_button(stmkmenu,value='Plot Mask...',uname='mkplBut',sensitive=0,/menu)
       void     = widget_button(plmkmenu,value='By Detector Number',uname='mkpldnum')
       void     = widget_button(plmkmenu,value='By Angle',uname='mkplangl')
       void     = widget_button(stmkmenu,value='Mask Bragg Peaks...',uname='mkbpBut',sensitive=0)
       void     = widget_button(maskmenu,value='Read Mask File...',uname='mkrdBut',/separator)
       void     = widget_button(maskmenu,value='Save Mask File...',uname='mksvBut',sensitive=0)
       void     = widget_button(maskmenu,value='Clear Mask',uname='mkclBut',sensitive=0,/separator)
       void     = dm_widget_button(igeomenu,value='Direct',uname='directGeom')
       void     = dm_widget_button(igeomenu,value='Inverse',uname='inverseGeom')
       for i=0,n_elements(self.intn_typ_str)-1 do $
           void = dm_widget_button(itypmenu,value=self.intn_typ_str[i],uname='intnTyp'+dm_to_string(i,/int))
       void     = dm_widget_button(itypmenu,value='E-Integrated Intensity',uname='intnEint',set_button=self.eint_yn,$
                  onstring='E-Integrated Intensity Enabled',offstring='E-Integrated Intensity Disabled',/separator)
       void     = dm_widget_button(binmmenu,value='Average',uname='binAvg')
       void     = dm_widget_button(binmmenu,value='Sum',uname='binSum')
       bincsmenu= widget_button(binmmenu,value='Constant Intervals',/menu,/separator)
       void     = dm_widget_button(bincsmenu,value='Yes',uname='bincsYes')
       void     = dm_widget_button(bincsmenu,value='No',uname='bincsNo')
       binexmenu= widget_button(binmmenu,value='Use External Binning Library',/menu)
       void     = dm_widget_button(binexmenu,value='Yes',uname='binexYes')
       void     = dm_widget_button(binexmenu,value='No',uname='binexNo')
       void     = dm_widget_button(moncmenu,value='Sum',uname='monSum')
       void     = dm_widget_button(moncmenu,value='Integrated Intensity',uname='monFit')
       mondmenu = widget_button(moncmenu,value='Monitor Choice',/menu,/separator)
       void     = dm_widget_button(mondmenu,value='FC0',uname='monFC0')
       void     = dm_widget_button(mondmenu,value='BM1',uname='monBM1')
       void     = dm_widget_button(mondmenu,value='Duration',uname='monTIME')
       if lmgr(/runtime) then usertmp = 'User Macro' else usertmp = 'User Macro/Function/Procedure'
       void     = dm_widget_button(optnmenu,value=usertmp,uname='userFunction',$
                  set_button=strlen(self.userf_str),onstring='Enable '+usertmp,$
                  offstring='Disable '+usertmp,/separator)
       void     = dm_widget_button(optnmenu,value='Allow Extra Viewing Axis',uname='extraVaxis',$
                  set_button=self.extravaxis_yn,onstring='Allow Extra Viewing Axis',$
                  offstring='Disallow Extra Viewing Axis',sensitive=0)
       void     = dm_widget_button(optnmenu,value='Fold to the Plus Side',uname='foldRight',$
                  set_button=self.foldright,onstring='Fold to the Plus Side',$
                  offstring='Fold to the Side with Larger Range',sensitive=0)           
       void     = dm_widget_button(optnmenu,value='Dark Angle Correction',uname='darkCorr',$
                  set_button=self.dcor_yn,onstring='Dark Angle Correction Enabled',$
                  offstring='Dark Angle Correction Disabled',sensitive=0)
       void     = dm_widget_button(optnmenu,value='E-Dependent Detector Efficiency Correction',uname='edepEff',$
                  set_button=self.edep_yn,onstring='E-Dependent Detector Efficiency Correction Enabled',$
                  offstring='E-Dependent Detector Efficiency Correction Disabled')
       void     = dm_widget_button(optnmenu,value='Hide Error Bar',uname='errBar',set_button=1-self.errbar,$
                  onstring='Error Bar Disabled',offstring='Error Bar Enabled')
       void     = widget_button(optnmenu,value='View Spurion Caused by Detector',uname='viewSpurion',$
                  sensitive=0,/separator)
       void     = widget_button(viewmenu,value='Close All Plot Windows',uname='killAll',sensitive=0)
       arrpmenu = widget_button(viewmenu,value='Arrange Plots',uname='arrPlot',sensitive=0,/menu)
       void     = dm_widget_button(powdmenu,value='None',uname='powderNone')
       void     = dm_widget_button(powdmenu,value='Al',uname='powderAl')
       void     = dm_widget_button(powdmenu,value='Cu',uname='powderCu')
       void     = dm_widget_button(powdmenu,value='Stainless Steel',uname='powderSt')
       void     = dm_widget_button(powdmenu,value='Specify',uname='powderSp')
       void     = dm_widget_button(vaxsmenu,value='Orthogonal',uname='vaxisOrtho')
       void     = dm_widget_button(vaxsmenu,value='Arbitrary',uname='vaxisArbi')
       void     = dm_widget_button(abspmenu,value='None',uname='abspNone')
       void     = dm_widget_button(abspmenu,value='Cylinder',uname='abspCylinder')
       void     = dm_widget_button(abspmenu,value='Annulus',uname='abspAnnulus')
       void     = dm_widget_button(abspmenu,value='Sphere',uname='abspSphere')
       void     = dm_widget_button(abspmenu,value='Flat Plate',uname='abspFlat')
       void     = dm_widget_button(epadmenu,value='None',uname='epadNone')
       void     = dm_widget_button(epadmenu,value='Automatic',uname='epadAuto')
       void     = dm_widget_button(epadmenu,value='Specify',uname='epadSpec')
       void     = dm_widget_button(epadmenu,value='Check the First File Only',uname='epadCheckfirst',set_button=self.eadjust_checkfirst,$
                  onstring='Check the First File Only',offstring='Check All Files',/separator)
       void     = dm_widget_button(pbgcmenu,value='Black',uname='pbgBlack')
       void     = dm_widget_button(pbgcmenu,value='White',uname='pbgWhite')
       void     = widget_button(arrpmenu,value='Cascade',uname='arrCascade')
       void     = widget_button(arrpmenu,value='Table',uname='arrTable')
       if strlen(self.helpfile[0]) ne 0 then $
       void     = widget_button(helpmenu,value='Contents',uname='hpctBut')
       if strlen(self.helpfile[1]) ne 0 then $
       void     = widget_button(helpmenu,value='How to Add a New File Type',uname='hpftBut')
       void     = widget_button(helpmenu,value='About Mslice for DCS',uname='selfBut',separator=(total(strlen(self.helpfile)) ne 0))

       ;menu bar separator for windows system
       if !version.os_family eq 'Windows' then $
          mbar_sep = widget_label(self.tlb,value=' ',sensitive=0,/dynamic_resize,/align_top,scr_ysize=5) $
       else $
          self.xsize[2] = self.xsize[2]+20
        
       self->dm_filemenu,/init ;also set up menu items
       self->reset_file
   
       ;load dcs_mslice.prm if existed
       parmfile = self.dirs[1]+self.pathsep+'dcs_mslice.prm'   
       if file_test(parmfile,/read) then begin
          self->dm_load_parameter,parmfile,/init
       endif else begin ;only ask if the default parameter file is unavailable
          ftype = dm_dialog_input('file type:',default=self.ftype,droplist_content=ptr_new(*self.ftypename),is_droplist=[1],$
              /return_number,title='Please choose')
          if self.ftype ne ftype then ok = self->dm_filetools('ftype_'+dm_to_string(ftype),/init)
       endelse 
       
       self->reset_droplist        ;adjust droplist values if necessary
       self->reset_projection
       self->reset_display
       
       widget_control,self.projBas,/update,/map
       widget_control,self.dispBas,/update,/map
    endelse
    
    dm_center_kid,self.tlb
    widget_control,self.tlb,/realize,/update
    geom  = widget_info(self.tlb,/geometry)
    if float(!version.release) ge 5.5 then $
       self.geom = [geom.xsize,geom.ysize] $
    else begin
       self.geom = [geom.scr_xsize,geom.scr_ysize]
       widget_control,self.tlb,scr_xsize=self.geom[0],scr_ysize=self.geom[1]
    endelse
    
    widget_control,self.tlb,set_uvalue=self,/map
    
    mesg = ''
    if self.eint_yn then mesg = [mesg,'E-integrated intensity is enabled.','Go to "Option->View Intensity As" to disable it.','']
    if strlen(self.userf_str) gt 0 then mesg = [mesg,usertmp+' is specified.','Go to "Option->'+usertmp+'" to disable it.','']
    if total(self.dcsdetmask) ne 0 then begin
       if self.dcsdetmask[0] then tmpstr = 'Negative angle'
       if self.dcsdetmask[4] then begin
          if n_elements(tmpstr) ne 0 then tmpstr = tmpstr+' and upper and lower bank' else tmpstr = 'Upper and lower bank' 
       endif else begin
          if self.dcsdetmask[1] then begin
             if n_elements(tmpstr) ne 0 then tmpstr = tmpstr+' and lower bank' else tmpstr = 'Lower bank' 
          endif
          if self.dcsdetmask[3] then begin
             if n_elements(tmpstr) ne 0 then tmpstr = tmpstr+' and upper bank' else tmpstr = 'Upper bank' 
          endif
       endelse
       if self.dcsdetmask[2] then begin
          if n_elements(tmpstr) ne 0 then tmpstr = tmpstr+' and central bank' else tmpstr = 'Central bank' 
       endif
       mesg = [mesg,tmpstr+' detectors are masked.','Go to "Mask->Clear Mask" to clear it.','']
    endif
    if n_elements(mesg) ge 2 then ok = dialog_message(mesg[1:(n_elements(mesg)-2)],dialog_parent=self.tlb)
    xmanager,registerName,self.tlb,event_handler='dcs_mslice_event',cleanup='dcs_mslice_Exit',/no_block
    return,1
end

;definition of dcs_mslice class
pro dcs_mslice__define
    void={dcs_mslice,               $
         group_leader:     0L,      $ ;group leader
         tlb:              0L,      $ ;container base of most of the widgets
         xsize:     lonarr(5),      $ ;to record the xsize
         geom:        [0L,0L],      $ ;xsize,ysize
         ssfac:          1.0e,      $ ;self shielding factor
         temperature:    300e,      $ ;sample temperature for detailed balance calculation
         instrgeom:        0s,      $ ;instrument geometry  0:direct    1:inverse
         ftype:            0s,      $ ;file types: 0-DCS 1-SPE ...
         ftypename: ptr_new(),      $ ;store file type names: 0-"DCS" 1-"SPE" ...
         ftypeext:  ptr_new(),      $ ;store file type extensions
         bgratePtr: ptr_new(),      $ ;a pointer to the overall background rate or detector dark count array [[bgrate],[dbgrate]]
         bgeran:      [0e,0e],      $ ;background energy range
         bgtchnPtr: ptr_new(),      $ ;pointer which points to {tchan:[start,end] qty:[end-start+1]} background information
         bgdetsPtr: ptr_new(),      $ ;pointer which points to an [ndet] array of background
         bgdata:    ptr_new(),      $ ;pointer which points to a data structure which contains the background information
                                    $ ;-single crystal,powder
                                    $ ; {qty:[ntchan,ndet], err:[ntchan,ndet]}
                                    $ ;-diffuse scattering
                                    $ ; {qty:[ndet],err:[ndet]} or {qty:[nen,ndet],err[nen,ndet]}
         bgdata2:   ptr_new(),      $ ;same as bgdata, used when extraempcan_yn = 1
         intn_typ:         0s,      $ ;intensity type
                                    $ ;0-S(Q,omega) 1-Symmetrized S(Q,omega) 2-d2sigma/dOmegadEf 3-Chi(Q,omega) 4-GDOS 5-VDOS 6-raw data
         intn_typ_str:strarr(7),    $ ;save the intensity type names
         lamcorrnxspe:     0s,      $ ;flag for ki/kf correction in NXSPE files 0: no (const t bins) 1: yes (const t bins)
                                    $ ; 2 no (const E bins) 3 yes (const E bins)
         mon_sumfit:       0s,      $ ;flag for monitor count calculation method 0:sum 1:gauss fit area
         dcs_mon:          1s,      $ ;dcs monitor choice: 0-fc0 1-bm1 2-duration; if 2, overall background rate is disabled
         bin_avgsum:       0b,      $ ;flag for binning method 0:average 1:sum
         binthreshold:     0e,      $ ;treshold level for binning
         bin_conststep:    0b,      $ ;flag for constant binning intervals 0:no  1:yes
         bin_extlib:       0b,      $ ;flag for using external binning libraries 0:no 1:yes
         view_ortharbi:    0s,      $ ;flag for ortho or arbitrary viewing axis 0:orthogonal 1:arbitrary
         powd_typ:         0s,      $ ;0-none 1-Al 2-Cu 3-Stainless Steel 4-specify
         powd_qvalue:ptr_new(),     $ ;save the q values for the powder line
         eint_yn:          0b,      $ ;flag for E-Integrated intensity
                                    $ ;if set, then in Q-cut, I=sum(I[i,j].dE[i,j])/sum(dE[i,j])*E_width
         negative2th:      0b,      $ ;flag for perserving negative 2th in powder mode
         extraempcan_yn:   0b,      $ ;flag for allowing 2 sets of empty can files
         empcanfactor:[0e,0e],      $ ;factors to be applied to the empty can files, used only when two sets are available
         extravaxis_yn:    0b,      $ ;flag for allowing extra viewing axis in single crystal modes
         foldright:        0b,      $ ;flag for folding to the right
         pbgcol_bw:        0b,      $ ;0:black 1:white for the plotting window background color
         eadjust:          0s,      $ ;flag for elastic peak adjustment 0:no 1:auto, 2:specify
         eadjust_spec:     0e,      $ ;specified elastic peak position
         eadjust_checkfirst:0b,     $ ;flag for using the first file shift for all, default is no(check all files)
         userf_str:        '',      $ ;save user function strings:[macro string, user function string, user procedure string] 
         userfwarning:     0b,      $ ;warning for userfunction
         errbar:           1b,      $ ;flag for showing error bar
         useitool:         0b,      $ ;flag for using itools
         dcor_yn:          0b,      $ ;flag for dark angle correction in diffuse scattering modes
         dcor_eqran:fltarr(4),      $ ;E and 2theta range for the dark angle correction or tchan background
         dcor_info: ptr_new(),      $ ;a pointer points to {factor:[n_A2], a2:[n_A2]}
         edep_yn:          0b,      $ ;flag for calculating energy dep. det. effectioncy 0:no 1:yes
         bitflip:          0b,      $ ;flag for checking bitflip errors 0:no 1:yes for DCS type only
         dcs2thcorr:       0b,      $ ;flag for correcting dcs 2theta   0:no 1:yes for DCS type only
         mask_filler:      0e,      $ ;a filler to indicate empty data in grid data such as slices
         dcsdetmask:bytarr(5),      $ ;flag for dcs detector bank mask [negative, lower bank, central bank, upper bank,upper and lower]
         mask:      ptr_new(),      $ ;detector mask pointer
         eff:       ptr_new(),      $ ;detector efficiency pointer
         fileSel:   obj_new(),      $ ;data file selection
         fileBas:          0L,      $ ;file selection base
         intn_lab:         '',      $ ;intensity label
         legd_lab:         '',      $ ;legend label for 1d plot
         titl_lab:         '',      $ ;title label
         monrate:          0e,      $ ;monitor rate from dcs data file
         kfactor:   ptr_new(),      $ ;(ki/kf)
                                    $ ;-powder,single crystal,diffuse scattering multiple E
                                    $ ; [ntchan] not rebinned
                                    $ ;-diffuse scattering sinegle E
                                    $ ; 1 for elastic and all time channels ki/kf_ave for the others
         samp_typ:         1s,      $ ;0-powder, 1-single crystal, 2-diffuse scattering single E, 3-diffuse multiple E
         rebinsize: ptr_new(),      $ ;allowed rebin size
         e_bin:     intarr(5),      $ ;ebin size/erange choice
                                    $ ; [0:1] e_bin size [2:3] erange choice [4] e_bin size for diffuse scattering mult. E
         e_range:     [0e,0e],      $ ;diffuse scattering energy range
         projBas:          0L,      $ ;projection base
         dispBas:          0L,      $ ;display base
         latt_parm: dblarr(6),      $ ;lattice parameters [a,b,c,alpha,beta,gamma]
         latt_recparm:dblarr(6),    $ ;reciprocal lattice parameters
         orie_u:   [1e,0e,0e],      $ ;[ux,uy,uz]
         orie_v:   [0e,1e,0e],      $ ;[vx,vy,vz]
         orie_psi:         0e,      $ ;psi
         orie_offset:      0e,      $ ;save the offset angle of psi
         view_type: strarr(5),      $ ;view  type ['Energy','|Q|','2Theta','Det Group Number','Time Channel']
         view_label:strarr(5),      $ ;labels     ['E','|Q|','2Theta','Det Group', 'T']
         view_unit: strarr(5),      $ ;units      ['(meV)','(A^-1)','(^o)','','(usec)']
         view_u1typ:       1s,      $ ;powder u1 type 0-E, 1-Q, 2-2theta, 3-Det Group
         view_u1:   fltarr(3),      $ ;[u1h,u1k,u1l]
         view_u1lab:strarr(2),      $ ;['|Q|','Qh'] -u1 label
         view_u1fold:      0b,      $ ;flag for folding x axis, single crystal only
         view_u1foldcenter:0e,      $ ;folding center for u1, default is 0
         view_u2typ:       0s,      $ ;powder u2 type 0-E, 1-Q, 2-2theta, 3-Det Group
         view_u2:   fltarr(3),      $ ;[u2h,u2k,u2l]
         view_u2lab:strarr(2),      $ ;['E','Qk']   -u2 label
         view_u2fold:      0b,      $ ;flag for folding x axis, single crystal only
         view_u2foldcenter:0e,      $ ;folding center for u2, default is 0
         view_u3:   fltarr(3),      $ ;[u3h,u3k,u3l]
         view_u3lab:       '',      $ ;u3 label
         view_u3fold:      0b,      $ ;flag for folding x axis, single crystal only
         view_u3foldcenter:0e,      $ ;folding center for u3, default is 0
         volsel_x:         0s,      $ ;x-axis selection for volume plot
         volsel_y:         0s,      $ ;y-axis selection for volume plot
         volsel_z:         0s,      $ ;z-axis selection for volume plot
         volsel_t:  ptr_new(),      $ ;thickness selection for volume plot, for samp_typ=-1
         volsel_tft:ptr_new(),      $ ;thickness range for volume plot,for samp_typ=-1
         volsel_t1:        0s,      $ ;thickness selection for volume tab
         volsel_tft1:fltarr(2),     $ ;thickness range for volume tab
         vol_xran:  fltarr(2),      $ ;x-ran
         vol_xstep:        0e,      $ ;x-step
         vol_yran:  fltarr(2),      $ ;y-ran
         vol_ystep:        0e,      $ ;y-step
         vol_zran:  fltarr(2),      $ ;z-ran
         vol_zstep:        0e,      $ ;z-step
         vol_iran:  fltarr(2),      $ ;i-ran
         vol_smooth:       1s,      $ ;smoothing level
         vol_interp:       0s,      $ ;interpolation 0: nearest neighbor 1:trilinear
         vol_color:        4s,      $ ;color map number 4-Blue/Green/Red/Yellow
         dispsel_h:        0s,      $ ;x-axis selection of contour plot for volume data
         dispsel_v:        0s,      $ ;y-axis selection of contour plot for volume data
         dispsel_t: ptr_new(),      $ ;thickness selection of the slice for volume data
         dispsel_tft:ptr_new(),     $ ;thickness range of the slice for volume data
         disp_perp:        2s,      $ ;perp to axis number
         disp_pran: fltarr(2),      $ ;perp thickness [from,to]
         disp_hran: fltarr(2),      $ ;hor. ran. [from,to]
         disp_hstep:       0e,      $ ;horizontal step
         disp_vran: fltarr(2),      $ ;ver. ran. [from,to]
         disp_vstep:       0e,      $ ;vertical step
         disp_tran: fltarr(2),      $ ;extra viewing axis thickness [from,to]
         disp_iran: fltarr(2),      $ ;intensity range[from,to]
         disp_smooth:      1s,      $ ;smooth level
         disp_sm_kernel:fltarr(3,3),$ ;default smooth kernel
         disp_shading:     1s,      $ ;shading 0-flat 1-Gouraud
         dispcmap: strarr(41),      $ ;color map names
         disp_color:      13s,      $ ;color map number 13-RAINBOW
         cut_x:     intarr(4),      $ ;cut along axis number
         cut_ran:   fltarr(2),      $ ;cut range
         cut_step:         0e,      $ ;cut step
         cutsel_t:  ptr_new(),      $ ;thickness selection of the cut for volume data
         cutsel_tft:ptr_new(),      $ ;thickness range of the cut for volume data
         cut_t1ran: fltarr(2),      $ ;thickness 1 range
         cut_t2ran: fltarr(2),      $ ;thickness 2 range
         cut_t3ran: fltarr(2),      $ ;thickness 3 range
         cut_y:     intarr(4),      $ ;y axis number
         cut_yran:  fltarr(2),      $ ;y range
         cut_symbol:       0s,      $ ;plot symbol number
         symbols:  strarr(17),      $ ;symbol names
         cut_color:        0s,      $ ;color number
         colors:   strarr(16),      $ ;color names
         coldefine: bytarr(3),      $ ;user defined color
         cut_line:         0s,      $ ;linestyle number,default solid line
         lines:     strarr(7),      $ ;linestyle names
         traj_type: strarr(5),      $ ;traj type names
         traj_x:    intarr(4),      $ ;trajectory x type
         traj_xran: fltarr(2),      $ ;[xfrom,xto]
         traj_y:    intarr(4),      $ ;trajectory y type
         traj_yran: fltarr(2),      $ ;[yfrom,yto]
         traj_z1:   intarr(4),      $ ;trajectory z type
         traj_z1ran:fltarr(2),      $ ;[zfrom,zto]
         traj_z2:   intarr(4),      $ ;trajectory z type
         traj_z2ran:fltarr(2),      $ ;[zfrom,zto]
         traj_layer:       0s,      $ ;for trajectory overplot coloring
         detPosPtr: ptr_new(),      $ ;pointer which points to a structure
                                    $ ;{two_theta:[ndet], dtwo_theta:[ndet], psi:[ndet], dpsi:[ndet]}
         dataStrPtr:ptr_new(),      $ ;pointer which points to a data structure
                                    $ ;-single crystal,powder
                                    $ ; {qty:[ntchan,ndet], err:[ntchan,ndet], e:[ntchan],weight:weight,eief:eief,ewid:ewid[ntchan],$
                                    $ ;  info:{ch_wl:ch_wl,ch_ms:ch_ms,ch_srdenom:ch_srdenom,ch_res:ch_res,tsdmin:tsdmin,$
                                    $ ;  temperature:temperature,time:[ntchan]}}    time is only for dcs data
                                    $ ;-diffuse scattering
                                    $ ; {qty:[ndet,nfiles],err:[ndet,nfiles],rotation:[nfiles],weight:[nfiles],eief:eief,ewid:ewid[ntchan],$
                                    $ ;  info:{ch_wl:ch_wl,ch_ms:ch_ms,ch_srdenom:ch_srdenom,ch_res:ch_res,tsdmin:tsdmin,$
                                    $ ;  temperature:temperature},filename:open_file,filepath:open_path}
         projStrPtr:ptr_new(),      $ ;pointer which points to a structure
                                    $ ;-single crystal
                                    $ ; {V:[3,nen,ndet],en:[nen],two_theta:[ndet],$
                                    $ ;  qty:[nen,ndet],err:[nen,ndet],ewid:[nen]}
                                    $ ;-powder
                                    $ ; {Q:[nen,ndet],en:[nen],two_theta:[ndet],qty:[nen,ndet],err:[nen,ndet],ewid:[nen],time:[nen]}
                                    $ ;-diffuse scattering
                                    $ ; single E
                                    $ ; {V:[3,ndet,nfile],en:[1],two_theta:[ndet],$
                                    $ ;  qty:[ndet,nfile],err:[ndet,nfile]}
                                    $ ; multiple E
                                    $ ; {V:[3,nen,ndet,nfile],en:[nen],two_theta:[ndet],$
                                    $ ;  qty:[nen,ndet,nfile],err:[nen,ndet,nfile]}
         dataPtr:   ptr_new(),      $ ;general volume data pointer
                                    $ ;{data:[ndat,nitem],label:[nitem],unit:[nitem],is_uniq:[nitem],instrument:''} ; the first two item needs to be I and dI
         powd_line: ptr_new(),      $ ;powder line for diffuse scattering mode
         absp_info: ptr_new(),      $ ;absorption information [type,interoplate,l0,l1,Rin/width,Rout/thickness,angle]
                                    $ ; type: 0-cylinder 1-annulus 2-sphere 3-flat plate
         absp_npts: ptrarr(4),      $ ;saving npts choice option numbers
         absp_nptst:ptrarr(4),      $ ;saving npts choice option strings
         file_hist:ptrarr(10),      $ ;save the latest 10 file selections
         n_history:        0s,      $ ;number of histories
         gr_list:   obj_new(),      $ ;container of all the graphic objects
         current:   obj_new(),      $ ;current plot object
         error:            0b,      $ ;error flag
         helpfile:    ['',''],      $ ;help file
         dirs:        ['',''],      $ ;[data directory, work directory]
         pathsep:          '',      $ ;path separator
         symfont:          '',      $ ;symbol font name
         eief:             0e,      $ ;save fixed energy, Ei for direct geometry, Ef for inverse geometry
         estartend: fltarr(3),      $ ;save [e_start, e_end, n_en] for projection calculation, only used when user function is specified 
         phxfile:          '',      $ ;save spe detector info phx file name
         a2:        fltarr(4),      $ ;save a2 start,end,step info, a2[3] saves the first file angle
         info:      ptr_new(),      $ ;self information
         DAVETool:  obj_new()       $ ;from DAVE
    }
end

pro dcs_mslice,group_leader=group_leader, workDir=workDir, dataDir=dataDir, DAVETool=DAVETool,dataPtr=dataPtr,obj_mslice=obj_mslice,_extra=extra
    obj_mslice = obj_new('dcs_mslice',group_leader=group_leader,workDir=workDir,dataDir=dataDir,DAVETool=DAVETool,dataPtr=dataPtr)
end